/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rubeam.cmps.ComponentElement;
import rubeam.cmps.ComponentElementLoad;
import rubeam.cmps.ComponentNodalLoad;
import rubeam.cmps.ComponentNode;
import rubeam.gui.RuBeamInspector;
import rubeam.gui.RuBeamMenu;
import rubeam.model.Element;
import rubeam.model.Node;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractElementLoad;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.loads.ElementConLoadAng;
import rubeam.model.loads.ElementUniDistribLoadNor;
import rubeam.model.loads.NodalLoadAng;
import cib.cad.db.comp.Component;
import cib.cad.kernel.Kernel;
import cib.cad.kernel.cmds.Copy;
import cib.cad.kernel.cmds.Remove;
import cib.util.NameSpace;
import cib.util.coll.ClassFilter;
import cib.util.coll.CollUtilities;
import cib.util.coll.IteratorFilter;
import cib.util.coll.ObservableSet;

/**
 * @author Christian Koch
 */
public class RuBeam {

    public static final String VERSION = "10/30/2009";
    
    public static final String STRUCTURE = "RuBeamModel";
    public static final String ELEMENT_PREFIX = "e";
    public static final String NODE_PREFIX = "n";
    public static final String NODALLOADANG_PREFIX = "P";
//    public static final String NODEMOMENT_PREFIX = "nM";
    public static final String UNIDISTRIBLOAD_PREFIX = "q";
//    public static final String UNIDISTRIBSEC_PREFIX = "dq";
    public static final String CONLOADANG_PREFIX = "F";
    public static final String GENSECTION_PREFIX = "gen";
    public static final String RECSECTION_PREFIX = "rec";
    public static final String IPROFSECTION_PREFIX = "iprof";
    public static final String SANDSECTION_PREFIX = "sand";

    private static Structure s_structure = null;
    private static RuBeamInspector s_inpector = null;
    private static ObservableSet.Listener<Component> s_compSetListener;
    private static Remove.Listener s_removeListener;
    private static Copy.Listener s_copyListener;
    private static boolean AUTO_ANALYSING = false;
    private static boolean AUTO_UPDATING = true;

    // Remove
    private static Map<Remove, Set<ComponentNode>> m_removeNodeMap = new HashMap<Remove, Set<ComponentNode>>();
    private static Map<Remove, Set<ComponentElement>> m_removeElementMap = new HashMap<Remove, Set<ComponentElement>>();
    private static Map<Remove, Set<ComponentElement>> m_removeSelectElementMap = new HashMap<Remove, Set<ComponentElement>>();
    private static Map<Remove, Set<Component>> m_removeNodeLoadMap = new HashMap<Remove, Set<Component>>();
    private static Map<Remove, Set<Component>> m_removeElementLoadMap = new HashMap<Remove, Set<Component>>();
    private static Map<Remove, Map<ComponentNode, List<AbstractNodeLoad>>> m_removeAssignedNodeLoadMap 
        = new HashMap<Remove, Map<ComponentNode, List<AbstractNodeLoad>>>();
    private static Map<Remove, Map<ComponentElement, List<AbstractElementLoad>>> m_removeAssignedElementLoadMap 
        = new HashMap<Remove, Map<ComponentElement, List<AbstractElementLoad>>>();
    private static Map<Component, Node> m_removeLoadNodeMap = new HashMap<Component, Node>();
    private static Map<Component, Element> m_removeLoadElementMap = new HashMap<Component, Element>();
    // Copy
    private static Map<Copy, Set<ComponentNode>> m_copySelectNodeMap = new HashMap<Copy, Set<ComponentNode>>();

    public static void setAutoAnalysing(boolean auto) {
        AUTO_ANALYSING = auto;
        if (AUTO_ANALYSING) {
            validateStructure();
        }
        else {
            invalidateStructure();
        }
    }

    public static boolean isAutoAnalysing() {
        return AUTO_ANALYSING;
    }

    public static void setAutoUpdating(boolean auto) {
        AUTO_UPDATING = auto;
    }
    public static boolean isAutoUpdating() {
        return AUTO_UPDATING;
    }

    
    public static Structure getStructure() {
        Kernel krnl = Kernel.getInstance();
        NameSpace ns = krnl.getDatabase().getNameSpace();
        Object o = ns.getObject(STRUCTURE);
        if (o == null || !(o instanceof Structure)) {
            Structure g = new Structure();
            ns.put(g, STRUCTURE);
            return g;
        }
        return (Structure) o;
    }
    
    public static void setInspector(RuBeamInspector inspector) {
        s_inpector = inspector;
    }
    public static RuBeamInspector getInspector() {
        return s_inpector;
    }
    
    public static void invalidateStructure() {
        s_structure.setValid(false);
        AUTO_ANALYSING = false;
        RuBeamMenu.s_cbItem.setSelected(false);
        notifyCompsChanged();
        Kernel.getInstance().getUserInterface().updateUserInterface(); // repaint???
    }

    public static void validateStructure() {
        if (!s_structure.isValid()) {
            try {
                s_structure.solve();
            }
            catch (IllegalStateException e) {
                s_structure.setValid(false);
                AUTO_ANALYSING = false;
                RuBeamMenu.s_cbItem.setSelected(false);
                System.err.println(e.getMessage());
            }
            notifyCompsChanged();
        }
        Kernel.getInstance().getDatabase().updateDatabase();
        Kernel.getInstance().getUserInterface().updateUserInterface(); // repaint???
    }

    public static void initialize() {
        s_structure = getStructure();

        s_compSetListener = new ObservableSet.Listener<Component>() {
            public void wasChanged(final Component o) {
                if (o instanceof ComponentNode) {
                    ObservableSet<Component> compSet = Kernel.getInstance()
                            .getDatabase().getComponentSet();
                    ComponentNode cNode = (ComponentNode) o;

                    // get neighbor elements and set changed
                    Set<Element> elements = getStructure().getElements(cNode.getNode());
                    Iterator<Component> eit = CollUtilities.filterableIterator(
                            compSet.iterator(), new ClassFilter(
                                    ComponentElement.class));
                    while (eit.hasNext()) {
                        ComponentElement cElement = (ComponentElement) eit.next();
                        if (elements.contains(cElement.getElement())) {
                            compSet.wasChanged(cElement);
                         }
                    }
                    
                    // get element loads and set changed
                    eit = CollUtilities.filterableIterator(
                                compSet.iterator(), new ClassFilter(
                                        ComponentElementLoad.class));
                    while (eit.hasNext()) {
                        ComponentElementLoad cLoad = (ComponentElementLoad) eit.next();
                        if (elements.contains(cLoad.getLoad().getElement())) {
                           compSet.wasChanged(cLoad);
                        }
                    }
                    
                    // get node loads and set changed
                    List<AbstractNodeLoad> nodeLoads = 
                        s_structure.getNodeLoads(cNode.getNode()); //cNode.getNode().getLoadSet();
                    if (nodeLoads != null) {
                        eit = CollUtilities.filterableIterator(compSet
                                .iterator(), new ClassFilter(
                                ComponentNodalLoad.class));
                        while (eit.hasNext()) {
                            ComponentNodalLoad cLoad = (ComponentNodalLoad) eit
                                    .next();
                            for (AbstractNodeLoad load : nodeLoads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
            }
        };
        Kernel.getInstance().getDatabase().getComponentSet().addListener(
                s_compSetListener);

        // Remove listener
        s_removeListener = new Remove.Listener() {
            public void preDoRemove(Remove cmd) {
                Set<Component> compSet = Kernel.getInstance().getDatabase()
                        .getComponentSet();
                Set<Component> selSet = Kernel.getInstance().getDatabase()
                        .getSelectSet();
                // Nodes
                Set<ComponentNode> nodeSet = new HashSet<ComponentNode>();
                m_removeNodeMap.put(cmd, nodeSet);

                Set<ComponentElement> elementSet = new HashSet<ComponentElement>();
                
                Map<ComponentNode, List<AbstractNodeLoad>> nodeLoadMap = 
                    new HashMap<ComponentNode, List<AbstractNodeLoad>>();
                Iterator<Component> it = CollUtilities
                        .filterableIterator(selSet.iterator(), new ClassFilter(
                                ComponentNode.class));
                while (it.hasNext()) {
                    ComponentNode cNode = (ComponentNode) it.next();
                    nodeSet.add(cNode);
                    
                    // get neighbor elements and select
                    Set<Element> elements = s_structure.getElements(cNode.getNode());
                    Iterator<Component> eit = CollUtilities.filterableIterator(
                            compSet.iterator(), new ClassFilter(
                                    ComponentElement.class));
                    while (eit.hasNext()) {
                        ComponentElement cElement = (ComponentElement) eit.next();
                        if (elements.contains(cElement.getElement()))
                            elementSet.add(cElement);
                    }
                    // get assigned loads and save
                    List<AbstractNodeLoad> nodeLoads = s_structure.getNodeLoads(cNode
                            .getNode());// cNode.getNode().getLoadSet();
                    if (nodeLoads != null) {
                        nodeLoadMap.put(cNode, nodeLoads);
                    }
                }
                m_removeSelectElementMap.put(cmd, elementSet);
                m_removeAssignedNodeLoadMap.put(cmd, nodeLoadMap);
                selSet.addAll(elementSet);

                // Elements
                Map<ComponentElement, List<AbstractElementLoad>> elementLoadMap = new HashMap<ComponentElement, List<AbstractElementLoad>>();
                elementSet = new HashSet<ComponentElement>();
                m_removeElementMap.put(cmd, elementSet);

                it = CollUtilities.filterableIterator(selSet.iterator(),
                        new ClassFilter(ComponentElement.class));
                while (it.hasNext()) {
                    ComponentElement cElement = (ComponentElement) it.next();
                    elementSet.add(cElement);
                 
                    // get assigned loads and save
                    List<AbstractElementLoad> elementLoads = s_structure.getElementLoads(cElement.getElement());//cNode.getNode().getLoadSet();
                    if (elementLoads != null) {
                        elementLoadMap.put(cElement, elementLoads);
                    }
                }
                m_removeAssignedElementLoadMap.put(cmd, elementLoadMap);
                
                // Loads
                Set<Component> loadSet = new HashSet<Component>();
                m_removeNodeLoadMap.put(cmd, loadSet);
                it = CollUtilities.filterableIterator(selSet.iterator(),
                        new ClassFilter(ComponentNodalLoad.class));
                while (it.hasNext()) {
                    ComponentNodalLoad cLoad = (ComponentNodalLoad) it.next();
                    loadSet.add(cLoad);
                }
                loadSet = new HashSet<Component>();
                m_removeElementLoadMap.put(cmd, loadSet);
                it = CollUtilities.filterableIterator(selSet.iterator(),
                        new ClassFilter(ComponentElementLoad.class));
                while (it.hasNext()) {
                    ComponentElementLoad cLoad = (ComponentElementLoad) it.next();
                    loadSet.add(cLoad);
                }
            }

            public void postDoRemove(Remove cmd) {
                ObservableSet<Component> compSet = Kernel.getInstance().getDatabase()
                    .getComponentSet();
            
                // Nodes
                for (ComponentNode cNode : m_removeNodeMap.get(cmd)) {
                    s_structure.removeNode(cNode.getNode());
                    // get node loads and set changed
                    List<AbstractNodeLoad> loads = m_removeAssignedNodeLoadMap.get(cmd)
                    .get(cNode);
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities
                                .filterableIterator(
                                        compSet.iterator(),
                                        new ClassFilter(ComponentNodalLoad.class));
                        while (eit.hasNext()) {
                            ComponentNodalLoad cLoad = (ComponentNodalLoad) eit
                                    .next();
                            for (AbstractNodeLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Elements
                for (ComponentElement cElement : m_removeElementMap.get(cmd)) {
                    s_structure.removeElement(cElement.getElement());
                    // get element loads and set changed
                    List<AbstractElementLoad> loads = m_removeAssignedElementLoadMap.get(cmd).get(cElement);
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities.filterableIterator(compSet
                                .iterator(), new ClassFilter(
                                ComponentElementLoad.class));
                        while (eit.hasNext()) {
                            ComponentElementLoad cLoad = (ComponentElementLoad) eit
                                    .next();
                            for (AbstractElementLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Loads
                for (Component cLoad : m_removeNodeLoadMap.get(cmd)) {
                    AbstractNodeLoad load = ((ComponentNodalLoad)cLoad).getLoad();
                    m_removeLoadNodeMap.put(cLoad, load.getNode());
                    s_structure.removeNodeLoad(load);
                }
                for (Component cLoad : m_removeElementLoadMap.get(cmd)) {
                    AbstractElementLoad load = ((ComponentElementLoad)cLoad).getLoad();
                    m_removeLoadElementMap.put(cLoad, load.getElement());
                    s_structure.removeElementLoad(load);
                }
//                Kernel.getInstance().getUserInterface().repaintUserInterface();
                
            }

            public void postUndoRemove(Remove cmd) {
                Set<Component> selSet = Kernel.getInstance().getDatabase()
                        .getSelectSet();
                // Nodes
                for (ComponentNode cNode : m_removeNodeMap.get(cmd)) {
                    ObservableSet<Component> compSet = Kernel.getInstance()
                    .getDatabase().getComponentSet();
                    
                    s_structure.addNode(cNode.getNode());
                    List<AbstractNodeLoad> loads = m_removeAssignedNodeLoadMap.get(cmd)
                            .get(cNode);
                    if (loads != null) {
                        for (AbstractNodeLoad load : loads) {
                            s_structure.putNodeLoad(cNode.getNode(), load);
                        }
                    }
                    
                    // get node loads and set changed
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities.filterableIterator(compSet
                                .iterator(), new ClassFilter(
                                ComponentNodalLoad.class));
                        while (eit.hasNext()) {
                            ComponentNodalLoad cLoad = (ComponentNodalLoad) eit
                                    .next();
                            for (AbstractNodeLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Elements
                for (ComponentElement cElement : m_removeElementMap.get(cmd)) {
                    ObservableSet<Component> compSet = Kernel.getInstance()
                    .getDatabase().getComponentSet();
                    
                    s_structure.addElement(cElement.getElement());
                    List<AbstractElementLoad> loads = m_removeAssignedElementLoadMap.get(cmd).get(cElement);
                    if (loads != null) {
                        for (AbstractElementLoad load : loads) {
                            s_structure.putElementLoad(cElement.getElement(), load);
                        }
                    }
                    
                    // get element loads and set changed
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities.filterableIterator(compSet
                                .iterator(), new ClassFilter(
                                ComponentElementLoad.class));
                        while (eit.hasNext()) {
                            ComponentElementLoad cLoad = (ComponentElementLoad) eit
                                    .next();
                            for (AbstractElementLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Loads
                for (Component cLoad : m_removeNodeLoadMap.get(cmd)) {
                    AbstractNodeLoad nodeForce = ((ComponentNodalLoad)cLoad).getLoad();
                    Node node = m_removeLoadNodeMap.get(cLoad);
                    s_structure.addNodeLoad(nodeForce); 
                    s_structure.putNodeLoad(node, nodeForce);
                }
                for (Component cLoad : m_removeElementLoadMap.get(cmd)) {
                    AbstractElementLoad load = ((ComponentElementLoad)cLoad).getLoad();
                    Element element = m_removeLoadElementMap.get(cLoad);
                    s_structure.addElementLoad(load);
                    s_structure.putElementLoad(element, load);
                }
                selSet.removeAll(m_removeSelectElementMap.get(cmd));
            }

            public void postRedoRemove(Remove cmd) {
                ObservableSet<Component> compSet = Kernel.getInstance().getDatabase()
                    .getComponentSet();
                
                // Node
                for (ComponentNode cNode : m_removeNodeMap.get(cmd)) {
                    s_structure.removeNode(cNode.getNode());
                    List<AbstractNodeLoad> loads = m_removeAssignedNodeLoadMap.get(cmd)
                    .get(cNode);
            
                    // get node loads and set changed
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities
                                .filterableIterator(
                                        compSet.iterator(),
                                        new ClassFilter(ComponentNodalLoad.class));
                        while (eit.hasNext()) {
                            ComponentNodalLoad cLoad = (ComponentNodalLoad) eit
                                    .next();
                            for (AbstractNodeLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Elements
                for (ComponentElement cElement : m_removeElementMap.get(cmd)) {
                    s_structure.removeElement(cElement.getElement());
                    List<AbstractElementLoad> loads = m_removeAssignedElementLoadMap.get(cmd).get(cElement);
                    
                    // get element loads and set changed
                    if (loads != null) {
                        Iterator<Component> eit = CollUtilities.filterableIterator(compSet
                                .iterator(), new ClassFilter(
                                ComponentElementLoad.class));
                        while (eit.hasNext()) {
                            ComponentElementLoad cLoad = (ComponentElementLoad) eit
                                    .next();
                            for (AbstractElementLoad load : loads) {
                                if (load == cLoad.getLoad())
                                    compSet.wasChanged(cLoad);
                            }
                        }
                    }
                }
                // Loads
                for (Component cLoad : m_removeNodeLoadMap.get(cmd)) {
                    s_structure.removeNodeLoad(((ComponentNodalLoad)cLoad).getLoad());
                }
                for (Component cLoad : m_removeElementLoadMap.get(cmd)) {
                    s_structure.removeElementLoad(((ComponentElementLoad)cLoad).getLoad());
                }
//                Kernel.getInstance().getUserInterface().repaintUserInterface();
                
            }
        };
        Remove.addListener(s_removeListener);

        // Copy listener
        s_copyListener = new Copy.Listener() {
            public void preDoCopy(Copy cmd) {
                Set<Component> compSet = Kernel.getInstance().getDatabase()
                        .getComponentSet();
                Set<Component> selSet = Kernel.getInstance().getDatabase()
                        .getSelectSet();
                
                // Modify selection set (copy nodes if elment selected)
                Set<ComponentNode> addSelNodes = new HashSet<ComponentNode>();
                m_copySelectNodeMap.put(cmd, addSelNodes);
                Iterator<Component> it = CollUtilities
                        .filterableIterator(selSet.iterator(), new ClassFilter(
                                ComponentElement.class));
                Set<ComponentElement> selElements = new HashSet<ComponentElement>();
                while (it.hasNext()) {
                    selElements.add((ComponentElement) it.next());
                }
                Set<ComponentNode> setNodes = new HashSet<ComponentNode>();
                it = CollUtilities.filterableIterator(selSet.iterator(),
                        new ClassFilter(ComponentNode.class));
                while (it.hasNext()) {
                    setNodes.add((ComponentNode) it.next());
                }
                for (ComponentElement cElement : selElements) {
                    Node ni = cElement.getElement().getNodeI();
                    Node nk = cElement.getElement().getNodeK();
                    ComponentNode ca = null;
                    ComponentNode cb = null;
                    it = CollUtilities.filterableIterator(compSet.iterator(),
                            new ClassFilter(ComponentNode.class));
                    while (it.hasNext()) {
                        ComponentNode n = (ComponentNode) it.next();
                        if (n.getNode() == ni)
                            ca = n;
                        else if (n.getNode() == nk)
                            cb = n;
                    }
                    if (!selSet.contains(ca) && !selSet.contains(cb)) {
                        addSelNodes.add(ca);
                        addSelNodes.add(cb);
                    }
                }
                selSet.addAll(addSelNodes);
            }

            public void postDoCopy(Copy cmd, Map<Component, List<Component>> map) {
                Set<Component> selSet = Kernel.getInstance().getDatabase()
                        .getSelectSet();
                NameSpace ns = Kernel.getInstance().getDatabase()
                        .getNameSpace();
                Set<Component> src = map.keySet();
                List<Node> nodes = new ArrayList<Node>();
                List<Element> elements = new ArrayList<Element>();
                List<ComponentNode> srcNodes = new ArrayList<ComponentNode>();
                List<ComponentElement> srcElements = new ArrayList<ComponentElement>();
                
                // Nodes
                Iterator<Component> it = CollUtilities.filterableIterator(src
                        .iterator(), new ClassFilter(ComponentNode.class));
                while (it.hasNext()) {
                    ComponentNode srcNode = (ComponentNode) it.next();
                    srcNodes.add(srcNode);
                    nodes.add(srcNode.getNode());
                    for (Component c : map.get(srcNode)) {
                        ComponentNode cNode = (ComponentNode) c;
                        cNode.getNode().setName(
                                ns.generateName(RuBeam.NODE_PREFIX));
                        s_structure.addNode(cNode.getNode());
                    }
                }
                // Elements
                it = CollUtilities.filterableIterator(src.iterator(),
                        new ClassFilter(ComponentElement.class));
                while (it.hasNext()) {
                    ComponentElement srcElement = (ComponentElement) it.next(); // key
                    srcElements.add(srcElement);
                    elements.add(srcElement.getElement());
                    List<Component> dstElements = map.get(srcElement); // clone list
                    for (int i = 0; i < dstElements.size(); i++) {
                        ComponentElement dstElement = (ComponentElement) dstElements.get(i);
                        Element element = dstElement.getElement();
                        element.setName(ns.generateName(RuBeam.ELEMENT_PREFIX));
                        
                        // add to structure and assign section
                        s_structure.addElement(element);
                        s_structure.putSection(element, srcElement.getElement().getSection());
                        
                        // set reference
                        for (ComponentNode cNode : srcNodes) {
                            if (cNode.getNode() == element.getNodeI()) {
                                ComponentNode dstNode = (ComponentNode) map
                                        .get(cNode).get(i);
                                element.setNodeI(dstNode.getNode());
                            }
                            if (cNode.getNode() == element.getNodeK()) {
                                ComponentNode dstNode = (ComponentNode) map
                                        .get(cNode).get(i);
                                element.setNodeK(dstNode.getNode());
                            }
                        }
                    }
                }
                // Assign loads...
                _assignLoads(map, srcNodes, nodes, srcElements, elements);
            
                // Modify selection
                selSet.removeAll(m_copySelectNodeMap.get(cmd));
            }

            public void postUndoCopy(Copy cmd,
                    Map<Component, List<Component>> map) {
                Set<Component> src = map.keySet();

                // Nodes
                Iterator<Component> it = CollUtilities.filterableIterator(
                        src.iterator(), new ClassFilter(ComponentNode.class));
                while (it.hasNext()) {
                    for (Component c : map.get(it.next())) {
                        ComponentNode node = (ComponentNode) c;
                        s_structure.removeNode(node.getNode());
                    }
                }
                // Elements
                it = CollUtilities.filterableIterator(
                        src.iterator(), new ClassFilter(ComponentElement.class));
                while (it.hasNext()) {
                    for (Component c : map.get(it.next())) {
                        ComponentElement element = (ComponentElement) c;
                        s_structure.removeElement(element.getElement());
                    }
                }
                // Node loads
                it = CollUtilities.filterableIterator(
                        src.iterator(), new ClassFilter(ComponentNodalLoad.class));
                while (it.hasNext()) {
                    for (Component c : map.get(it.next())) {
                        ComponentNodalLoad load = (ComponentNodalLoad) c;
                        s_structure.removeNodeLoad(load.getLoad());
                    }
                }
                // Element loads
                it = CollUtilities.filterableIterator(
                        src.iterator(), new ClassFilter(ComponentElementLoad.class));
                while (it.hasNext()) {
                    for (Component c : map.get(it.next())) {
                        ComponentElementLoad load = (ComponentElementLoad) c;
                        s_structure.removeElementLoad(load.getLoad());
                    }
                }
            }

            public void postRedoCopy(Copy cmd,
                    Map<Component, List<Component>> map) {
                
                List<Node> nodes = new ArrayList<Node>();
                List<Element> elements = new ArrayList<Element>();
                List<ComponentNode> srcNodes = new ArrayList<ComponentNode>();
                List<ComponentElement> srcElements = new ArrayList<ComponentElement>();
                
                // Nodes
                Set<Component> src = map.keySet();
                Iterator<Component> it = CollUtilities.filterableIterator(
                        src.iterator(), new ClassFilter(ComponentNode.class));
                while (it.hasNext()) {
                    ComponentNode srcNode = (ComponentNode) it.next();
                    srcNodes.add(srcNode);
                    nodes.add(srcNode.getNode());
                    for (Component c : map.get(srcNode)) {
                        ComponentNode node = (ComponentNode) c;
                        s_structure.addNode(node.getNode());
                    }
                }
                // Elements (already has references to its nodes...)
                it = CollUtilities.filterableIterator(src.iterator(),
                        new ClassFilter(ComponentElement.class));
                while (it.hasNext()) {
                    ComponentElement srcElement = (ComponentElement) it.next(); // key
                    srcElements.add(srcElement);
                    elements.add(srcElement.getElement());
                    for (Component c : map.get(srcElement)) {
                        ComponentElement element = (ComponentElement) c;
                        s_structure.addElement(element.getElement());
                    }
                }
                // Assign loads...
                _assignLoads(map, srcNodes, nodes, srcElements, elements);
            }
            
            private void _assignLoads(Map<Component, List<Component>> map, 
                    List<ComponentNode> srcNodes, List<Node> nodes,
                    List<ComponentElement> srcElemets, List<Element> elements) {
                
                NameSpace ns = Kernel.getInstance().getDatabase().getNameSpace();
                Set<Component> src = map.keySet();
                
                // NodeLoads (assign if copied together with node)
                Iterator<Component> it = CollUtilities.filterableIterator(src.iterator(),
                        new ClassFilter(ComponentNodalLoad.class));
                while (it.hasNext()) {
                    ComponentNodalLoad srcLoad = (ComponentNodalLoad) it.next(); // key
                    AbstractNodeLoad load = srcLoad.getLoad();
                    List<Component> dstLoads = map.get(srcLoad); // clone list
                                                                 // load
                    List<Component> dstNodes = null;
                    if (nodes.contains(s_structure.getAssignedNode(load))) {
                        int idx = nodes.indexOf(load.getNode());
                        ComponentNode srcNode = srcNodes.get(idx);
                        dstNodes = map.get(srcNode); // clone list
                    }    
                    for (int i = 0; i < dstLoads.size(); i++) {
                        ComponentNodalLoad dstLoad = (ComponentNodalLoad) dstLoads
                                .get(i);
                        AbstractNodeLoad nodeLoad = dstLoad.getLoad();
                        // set new name
                        String prefix = null;
                        if (nodeLoad instanceof NodalLoadAng)
                            prefix = NODALLOADANG_PREFIX;
                        if (prefix != null)
                            nodeLoad.setName(ns.generateName(prefix));
                        
                        s_structure.addNodeLoad(nodeLoad);
                        if (dstNodes != null) {
                            ComponentNode dstNode = (ComponentNode) dstNodes.get(i);
                            Node node = dstNode.getNode();
                            s_structure.putNodeLoad(node, nodeLoad);
                        }
                    }
                }
                
                // ElementLoads (assign if copied together with element)
                it = CollUtilities.filterableIterator(src.iterator(),
                        new ClassFilter(ComponentElementLoad.class));
                while (it.hasNext()) {
                    ComponentElementLoad srcLoad = (ComponentElementLoad) it.next(); // key
                    AbstractElementLoad load = srcLoad.getLoad();
                    List<Component> dstLoads = map.get(srcLoad); // clone list
                                                                 // load

                    List<Component> dstElements = null;
                    if (elements.contains(s_structure.getAssignedElement(load))) {
                        int idx = elements.indexOf(load.getElement());
                        ComponentElement srcElement = srcElemets.get(idx);
                        dstElements = map.get(srcElement); // clone list
                    }
                    for (int i = 0; i < dstLoads.size(); i++) {
                        ComponentElementLoad dstLoad = (ComponentElementLoad) dstLoads
                                .get(i);
                        AbstractElementLoad elementLoad = dstLoad.getLoad();
                        
                        // set new name
                        String prefix = null;
                        if (elementLoad instanceof ElementUniDistribLoadNor)
                            prefix = UNIDISTRIBLOAD_PREFIX;
                        else if (elementLoad instanceof ElementConLoadAng)
                            prefix = CONLOADANG_PREFIX;
                        if (prefix != null)
                            elementLoad.setName(ns.generateName(prefix));
                        
                        s_structure.addElementLoad(elementLoad);
                        if (dstElements != null) {
                            ComponentElement dstElement = (ComponentElement) dstElements.get(i);
                            Element element = dstElement.getElement();
                            s_structure.putElementLoad(element, elementLoad);
                        }
                    }
                }
            }
        };
        Copy.addListener(s_copyListener);

        // Kernel Listener
        Kernel.getInstance().addListener(new Kernel.Listener() {
            public void storeStarts() {
                Kernel.getInstance().getDatabase().getComponentSet()
                        .removeListener(s_compSetListener);
                Remove.removeListener(s_removeListener);
                Copy.removeListener(s_copyListener);
            }

            public void storeEnds() {
                Kernel.getInstance().getDatabase().getComponentSet()
                        .addListener(s_compSetListener);
                Remove.addListener(s_removeListener);
                Copy.addListener(s_copyListener);
            }

            public void loadStarts() {
                Kernel.getInstance().getDatabase().getComponentSet()
                        .removeListener(s_compSetListener);
                Remove.removeListener(s_removeListener);
                Copy.removeListener(s_copyListener);
            }

            public void loadEnds() {
                // Remove
                m_removeNodeMap.clear();
                m_removeElementMap.clear();
                m_removeSelectElementMap.clear();
                m_removeNodeLoadMap.clear();
                m_removeElementLoadMap.clear();
                m_removeAssignedNodeLoadMap.clear();
                m_removeAssignedElementLoadMap.clear();
                // Copy
                m_copySelectNodeMap.clear();

                // Graph
                s_structure = getStructure();

                // Listeners
                Kernel.getInstance().getDatabase().getComponentSet()
                        .addListener(s_compSetListener);
                Remove.addListener(s_removeListener);
                Copy.addListener(s_copyListener);
                s_inpector.updateDialog();
            }

            public void cmdStarts(String arg0) {
            }

            public void cmdEnds(String cmd) {
                if (cmd.equalsIgnoreCase("new")) {
                    s_structure = getStructure();
                    invalidateStructure();
                    return;
                }
                if (cmd.equalsIgnoreCase("setagain")
                        || cmd.startsWith("win")
                        || cmd.equalsIgnoreCase("contextmenu")
                        || cmd.equalsIgnoreCase("setresultsteps")) {
                    return;
                }
                if (cmd.equalsIgnoreCase("analyse")) {
                    if (AUTO_UPDATING)
                        s_inpector.updateDialog();
                    return;
                }
                if (AUTO_ANALYSING) {
                    validateStructure();
                    if (AUTO_UPDATING)
                        s_inpector.updateDialog();
                    return;
                }
                if (s_structure.isValid()) {
                    return;
                }

                if (!s_structure.isValid()) {
                    invalidateStructure();
                    if (AUTO_UPDATING)
                        s_inpector.updateDialog();
                }
            }
        });

    }

    static void notifyCompsChanged() {
        ObservableSet<Component> compSet = Kernel.getInstance().getDatabase()
                .getComponentSet();
        compSet.removeListener(s_compSetListener);
        Iterator<Component> it = CollUtilities.filterableIterator(compSet
                .iterator(), new IteratorFilter() {
            public boolean matches(Object o) {
                return (o instanceof ComponentNode)
                        || (o instanceof ComponentElement)
                        || (o instanceof ComponentNodalLoad)
                        || (o instanceof ComponentElementLoad);
            }
        });
      while (it.hasNext()) {
            compSet.wasChanged(it.next());
        }
        compSet.addListener(s_compSetListener);
    }
}
