/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmds.elmload;

import java.awt.Window;
import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rubeam.RuBeam;
import rubeam.cmps.ComponentElement;
import rubeam.cmps.ComponentElementConLoadAng;
import rubeam.model.Element;
import rubeam.model.Structure;
import rubeam.model.loads.ElementConLoadAng;
import cib.cad.db.Database;
import cib.cad.db.comp.Component;
import cib.cad.kernel.Kernel;
import cib.cad.kernel.cmds.util.LineRubberBand;
import cib.util.NameSpace;
import cib.util.cmd.Cmd;
import cib.util.cmd.CmdAbortedException;
import cib.util.coll.ClassFilter;
import cib.util.coll.CollUtilities;
import cib.util.view.ViewUtilities;

public class AddElementConLoadAng implements Cmd {

    List<ComponentElement> m_elements = new ArrayList<ComponentElement>();
    List<ComponentElementConLoadAng> m_cLoads = new ArrayList<ComponentElementConLoadAng>();
    List<ElementConLoadAng> m_loads = new ArrayList<ElementConLoadAng>();
    Structure m_structure = null;

    public void doCmd(Object context) throws CmdAbortedException {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        Set<Component> compSet = db.getComponentSet();
        Set<Component> selSet = db.getSelectSet();
        NameSpace ns = db.getNameSpace();
        m_structure = RuBeam.getStructure();
        ViewUtilities.requestFocusAsynchronously(
                (Window)krnl.getUserInterface());
        
        if (krnl.hasArgument()) { // macro mode
            String name = krnl.readString("Enter name");
            Iterator<Component> eit = CollUtilities.filterableIterator(compSet
                    .iterator(), new ClassFilter(ComponentElement.class));
            while (eit.hasNext()) {
                ComponentElement cBeam = (ComponentElement) eit.next();
                Element beam = cBeam.getElement();
                if (beam.getName().equalsIgnoreCase(name)) {
                    m_elements.add(cBeam);
                    break;
                }
            }
        }

        Point2D pI = null;
        Point2D pK = null;
        if (m_elements.isEmpty() && selSet.isEmpty()) {
            pI = krnl.readPoint("Enter first point");
            LineRubberBand rb = null;
            try {
                // Create rubber band
                rb = new LineRubberBand(krnl, pI);
                pK = krnl.readPoint("Enter second point");
            } finally {
                if (rb != null)
                    rb.close();
            }
        }
        Iterator<Component> it = null;
        if (!selSet.isEmpty()) {
            it = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentElement.class));
            if (pI == null && pK == null && !it.hasNext()) {
                pI = krnl.readPoint("Enter first point");
                LineRubberBand rb = null;
                try {
                    // Create rubber band
                    rb = new LineRubberBand(krnl, pI);
                    pK = krnl.readPoint("Enter second point");
                } finally {
                    if (rb != null)
                        rb.close();
                }
            }
            else {
                while (it.hasNext())
                    m_elements.add((ComponentElement) it.next());
            }
        }

        double F = krnl.readDouble("Enter load value", 10000);
        double angle = Math.toRadians(krnl.readDouble("Enter load angle", 0.0));
        double li = 0;

        if (m_elements.isEmpty()) { // No element selected
            li = krnl.readDouble("Enter distance from point ni", pI
                    .distance(pK) / 2);

            while (li > pI.distance(pK) || li < 0) {
                if (li < 0) {
                    li = krnl.readDouble("li must be greater than 0", pI
                            .distance(pK));
                }
                else {
                    li = krnl.readDouble("li must be smaller than "
                            + pI.distance(pK), pI.distance(pK));
                }
            }
        }

        else { // One or more elements selected
            double min = Double.MAX_VALUE;
            for (ComponentElement cElement : m_elements) {
                if (cElement.getElement().getLength() < min)
                    min = cElement.getElement().getLength();
            }
            li = krnl.readDouble("Enter distance from point ni", min / 2);
            while (li > min || li < 0) {
                if (li < 0) {
                    li = krnl.readDouble("li must be greater than 0", min);
                }
                else {
                    li = krnl.readDouble("li must be smaller than " + min, min);
                }
            }
        }
        String name = null;
        if (krnl.hasArgument()) {
            name = krnl.readString("Enter name");
        } 

        if (pI != null && pK != null && m_elements.isEmpty()) { // No element selected
            if (name == null)
                name = ns.generateName(RuBeam.CONLOADANG_PREFIX);
            ElementConLoadAng load = new ElementConLoadAng(F, angle, li, name);
            m_structure.addElementLoad(load);
            m_loads.add(load);
            ComponentElementConLoadAng cLoad = new ComponentElementConLoadAng(
                    load, pI, pK);
            m_cLoads.add(cLoad);
            compSet.add(cLoad);
        }
        else {
            for (ComponentElement cElement : m_elements) {
                Element beam = cElement.getElement();
                if (li > beam.getLength()) {
                    li = beam.getLength();
                }
               
                name = ns.generateName(RuBeam.CONLOADANG_PREFIX);
                ElementConLoadAng load = new ElementConLoadAng(F, angle,
                        li, name);
                m_structure.addElementLoad(load);
                m_structure.putElementLoad(beam, load);
                m_loads.add(load);

                ComponentElementConLoadAng cLoad = new ComponentElementConLoadAng(
                        load, null, null);
                m_cLoads.add(cLoad);
                compSet.add(cLoad);
            }
        }

        System.out.println(this);
    }

    public void undoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        Set<Component> compSet = db.getComponentSet();

        for (int i = 0; i < m_loads.size(); i++) {
            ComponentElementConLoadAng cLoad = m_cLoads.get(i);
            ElementConLoadAng load = m_loads.get(i);
            m_structure.removeElementLoad(load);
            compSet.remove(cLoad);
        }
    }

    public void redoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        Set<Component> compSet = db.getComponentSet();

        for (int i = 0; i < m_elements.size(); i++) {
            ComponentElement cElement= m_elements.get(i);
            ComponentElementConLoadAng cLoad = m_cLoads.get(i);
            ElementConLoadAng load = m_loads.get(i);
            m_structure.addElementLoad(load);
            if (cElement != null)
                m_structure.putElementLoad(cElement.getElement(), load);
            compSet.add(cLoad);
        }
    }

    public boolean changesState() {
        return true;
    }

    public boolean isUndoable() {
        return true;
    }

    public String toString() {
        String ret = "Load set for beam(s): ";
        for (ComponentElement cBeam : m_elements)
            ret += cBeam.getElement().getName() + ", ";
        return ret;

    }
}
