/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmds.elmload;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rubeam.RuBeam;
import rubeam.cmps.ComponentElement;
import rubeam.cmps.ComponentElementLoad;
import rubeam.model.Element;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractElementLoad;
import cib.cad.db.Database;
import cib.cad.db.comp.Component;
import cib.cad.kernel.Kernel;
import cib.util.cmd.Cmd;
import cib.util.cmd.CmdAbortedException;
import cib.util.coll.ClassFilter;
import cib.util.coll.CollUtilities;
import cib.util.coll.ObservableSet;

public class AssignElementLoads implements Cmd {

    private List<ComponentElementLoad> m_cLoads = new ArrayList<ComponentElementLoad>();
    private List<Element> m_elements = new ArrayList<Element>();
    private List<Point2D[]> m_points = new ArrayList<Point2D[]>();
    private Structure m_structure = null;

    public void doCmd(Object context) throws CmdAbortedException {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        Set<Component> selSet = db.getSelectSet();
        m_structure = RuBeam.getStructure();
    
        // Iterators
        Iterator<Component> nit;
        Iterator<Component> lit;
        ComponentElement cElement = null;
        
        // macro mode
        if (krnl.hasArgument()) {
            String loadName = krnl.readString("Enter load name");
            lit = CollUtilities.filterableIterator(compSet
                    .iterator(), new ClassFilter(ComponentElementLoad.class));
            if (! lit.hasNext()) {
                System.err.println("Load '" + loadName + "' not found!");
                return;
            }
            while (lit.hasNext()) {
                ComponentElementLoad cLoad = (ComponentElementLoad) lit.next();
                if (cLoad.getLoad().getName().equals(loadName)) {
                    m_cLoads.add(cLoad);
                    break;
                }
            }
            if (m_cLoads.isEmpty()) {
                System.err.println("Load '" + loadName + "' not found!");
                return;
            }
            if (krnl.hasArgument()) {
                String elmName = krnl.readString("Enter element name");
                nit = CollUtilities.filterableIterator(compSet.iterator(),
                        new ClassFilter(ComponentElement.class));
                if (!nit.hasNext()) {
                    System.err.println("Element '" + elmName + "' not found!");
                    return;
                }
                while (nit.hasNext()) {
                    ComponentElement ce = (ComponentElement) nit.next();
                    if (ce.getElement().getName().equals(elmName)) {
                        cElement = ce;
                        break;
                    }
                }
                if (cElement == null) {
                    System.err.println("Element '" + elmName + "' not found!");
                    return;
                }
                ComponentElementLoad cLoad = (ComponentElementLoad) m_cLoads.get(0);
                AbstractElementLoad load = cLoad.getLoad();
                Element element = cElement.getElement();
                Element assignedElm = m_structure.getAssignedElement(load);
                if (assignedElm == null) {
                    // save old point
                    m_points.add(new Point2D[]{cLoad.getPointI(), cLoad.getPointK()});
                    m_elements.add(null);
                    m_structure.putElementLoad(element, load);
                    compSet.wasChanged(cLoad);
                }
                else if (m_structure.getAssignedElement(load) != element) {
                    // save old element
                    m_points.add(null);
                    m_elements.add(assignedElm);
                    m_structure.putElementLoad(element, load);
                    compSet.wasChanged(cLoad);
                }
                else {
                    System.out.println("Load already assigned");
                }
            }
            else {
                System.err.println("No element specified!");
                return;
            }
            
        }
        // interactive mode
        else {
            if (selSet.isEmpty()) {
                System.err.println("No components selected!");
                return;
            }

            nit = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentElement.class));
            if (!nit.hasNext()) {
                System.err.println("No element selected!");
                return;
            }
            lit = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentElementLoad.class));
            if (!lit.hasNext()) {
                System.err.println("No loads selected!");
                return;
            }

            cElement = (ComponentElement) nit.next();
            if (nit.hasNext()) {
                System.err.println("More than one element selected!");
                return;
            }
            while (lit.hasNext()) {
                ComponentElementLoad cLoad = (ComponentElementLoad) lit
                        .next();
                AbstractElementLoad load = cLoad.getLoad();
                Element element = cElement.getElement();
                Element assignedElm = m_structure.getAssignedElement(load);
                if (assignedElm == null) {
                    // save old point
                    m_points.add(new Point2D[]{cLoad.getPointI(), cLoad.getPointK()});
                    m_elements.add(null);
                    m_cLoads.add(cLoad);
                    m_structure.putElementLoad(element, load);
                    compSet.wasChanged(cLoad);
                }
                else if (assignedElm != element) {
                    // save old element
                    m_elements.add(assignedElm);
                    m_points.add(null);
                    m_cLoads.add(cLoad);
                    m_structure.putElementLoad(element, load);
                    compSet.wasChanged(cLoad);
                }
                else {
                    System.out.println("Load already assigned");
                }
            }
        }
        System.out.println(this);
                    
    }

    public void undoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        
        List<Element> tmpElements = new ArrayList<Element>();
        List<Point2D[]> tmpPoints = new ArrayList<Point2D[]>();
        
        for (int i=0; i < m_cLoads.size(); i++) {
            ComponentElementLoad cLoad = m_cLoads.get(i);
            AbstractElementLoad load = cLoad.getLoad();
            Element assignedElm = m_structure.getAssignedElement(load);
            // restore point and save element
            if (m_elements.get(i) == null) {
                tmpPoints.add(null);
                tmpElements.add(assignedElm);
                m_structure.removeElementLoad(load);
                m_structure.addElementLoad(load);
                Point2D[] pts = m_points.get(i);
                cLoad.setPointI(pts[0].getX(), pts[0].getY());
                cLoad.setPointK(pts[1].getX(), pts[1].getY());

            }
            // restore node and save point/node
            else {
                if (assignedElm == null) {
                    tmpElements.add(null);
                    tmpPoints.add(new Point2D[]{
                            cLoad.getPointI(), cLoad.getPointK()});
                }
                else {
                    tmpElements.add(m_structure.getAssignedElement(load));
                    tmpPoints.add(null);
                }
                m_structure.putElementLoad(m_elements.get(i), load);
            }
            compSet.wasChanged(cLoad);
        }
        m_elements = tmpElements;
        m_points = tmpPoints;
    }

    public void redoCmd(Object context) {
        undoCmd(context);
    }

    public boolean changesState() {
        return ! m_cLoads.isEmpty();
    }

    public boolean isUndoable() {
        return true;
    }

    public String toString() {
        return (m_cLoads.isEmpty())? "No loads assigned." 
                : m_cLoads.size() + " loads assigned.";
    }
}
