/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmds.nodeload;

import java.awt.geom.Point2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import rubeam.RuBeam;
import rubeam.cmps.ComponentNodalLoad;
import rubeam.cmps.ComponentNode;
import rubeam.model.Node;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractNodeLoad;
import cib.cad.db.Database;
import cib.cad.db.comp.Component;
import cib.cad.kernel.Kernel;
import cib.util.cmd.Cmd;
import cib.util.cmd.CmdAbortedException;
import cib.util.coll.ClassFilter;
import cib.util.coll.CollUtilities;
import cib.util.coll.ObservableSet;

public class AssignNodeLoads implements Cmd {

    private List<ComponentNodalLoad> m_cLoads = new ArrayList<ComponentNodalLoad>();
    private List<Node> m_nodes = new ArrayList<Node>();
    private List<Point2D> m_points = new ArrayList<Point2D>();
    private Structure m_structure = null;

    public void doCmd(Object context) throws CmdAbortedException {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        Set<Component> selSet = db.getSelectSet();
        m_structure = RuBeam.getStructure();
    
        // Iterators
        Iterator<Component> nit;
        Iterator<Component> lit;
        ComponentNode cNode = null;
        
        // macro mode
        if (krnl.hasArgument()) {
            String loadName = krnl.readString("Enter load name");
            lit = CollUtilities.filterableIterator(compSet
                    .iterator(), new ClassFilter(ComponentNodalLoad.class));
            if (! lit.hasNext()) {
                System.err.println("Load '" + loadName + "' not found!");
                return;
            }
            while (lit.hasNext()) {
                ComponentNodalLoad cLoad = (ComponentNodalLoad) lit.next();
                if (cLoad.getLoad().getName().equals(loadName)) {
                    m_cLoads.add(cLoad);
                    break;
                }
            }
            if (m_cLoads.isEmpty()) {
                System.err.println("Load '" + loadName + "' not found!");
                return;
            }
            if (krnl.hasArgument()) {
                String nodeName = krnl.readString("Enter node name");
                nit = CollUtilities.filterableIterator(compSet.iterator(),
                        new ClassFilter(ComponentNode.class));
                if (!nit.hasNext()) {
                    System.err.println("Node '" + nodeName + "' not found!");
                    return;
                }
                while (nit.hasNext()) {
                    ComponentNode cn = (ComponentNode) nit.next();
                    if (cn.getNode().getName().equals(nodeName)) {
                        cNode = cn;
                        break;
                    }
                }
                if (cNode == null) {
                    System.err.println("Node '" + nodeName + "' not found!");
                    return;
                }
                ComponentNodalLoad cLoad = (ComponentNodalLoad) m_cLoads.get(0);
                AbstractNodeLoad load = cLoad.getLoad();
                Node node = cNode.getNode();
                Node assignedNode = m_structure.getAssignedNode(load);
                if (assignedNode == null) {
                    // save old point
                    m_points.add(cLoad.getPoint());
                    m_nodes.add(null);
                    m_structure.putNodeLoad(node, load);
                    compSet.wasChanged(cLoad);
                }
                else if (m_structure.getAssignedNode(load) != node) {
                    // save old node
                    m_points.add(null);
                    m_nodes.add(assignedNode);
                    m_structure.putNodeLoad(node, load);
                    compSet.wasChanged(cLoad);
                }
                else {
                    System.out.println("Load already assigned");
                }
            }
            else {
                System.err.println("No node specified!");
                return;
            }
            
        }
        // interactive mode
        else {
            if (selSet.isEmpty()) {
                System.err.println("No components selected!");
                return;
            }

            nit = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentNode.class));
            if (!nit.hasNext()) {
                System.err.println("No node selected!");
                return;
            }
            lit = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentNodalLoad.class));
            if (!lit.hasNext()) {
                System.err.println("No loads selected!");
                return;
            }

            cNode = (ComponentNode) nit.next();
            if (nit.hasNext()) {
                System.err.println("More than one node selected!");
                return;
            }
            while (lit.hasNext()) {
                ComponentNodalLoad cLoad = (ComponentNodalLoad) lit
                        .next();
                AbstractNodeLoad load = cLoad.getLoad();
                Node node = cNode.getNode();
                Node assignedNode = m_structure.getAssignedNode(load);
                if (assignedNode == null) {
                    // save old point
                    m_points.add(cLoad.getPoint());
                    m_nodes.add(null);
                    m_cLoads.add(cLoad);
                    m_structure.putNodeLoad(node, load);
                    compSet.wasChanged(cLoad);
                }
                else if (assignedNode != node) {
                    // save old node
                    m_nodes.add(assignedNode);
                    m_points.add(null);
                    m_cLoads.add(cLoad);
                    m_structure.putNodeLoad(node, load);
                    compSet.wasChanged(cLoad);
                }
                else {
                    System.out.println("Load already assigned");
                }
            }
        }
        System.out.println(this);
                    
    }

    public void undoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        
        List<Node> tmpNodes = new ArrayList<Node>();
        List<Point2D> tmpPoints = new ArrayList<Point2D>();
        
        for (int i=0; i < m_cLoads.size(); i++) {
            ComponentNodalLoad cLoad = m_cLoads.get(i);
            AbstractNodeLoad load = cLoad.getLoad();
            Node assignedNode = m_structure.getAssignedNode(load);
            // restore point and save node
            if (m_nodes.get(i) == null) {
                tmpPoints.add(null);
                tmpNodes.add(assignedNode);
                m_structure.removeNodeLoad(load);
                m_structure.addNodeLoad(load);
                cLoad.setNode(m_points.get(i));

            }
            // restore node and save point/node
            else {
                if (assignedNode == null) {
                    tmpNodes.add(null);
                    tmpPoints.add(cLoad.getPoint());
                }
                else {
                    tmpNodes.add(m_structure.getAssignedNode(load));
                    tmpPoints.add(null);
                }
                m_structure.putNodeLoad(m_nodes.get(i), load);
            }
            compSet.wasChanged(cLoad);
        }
        m_nodes = tmpNodes;
        m_points = tmpPoints;
    }

    public void redoCmd(Object context) {
        undoCmd(context);
    }

    public boolean changesState() {
        return ! m_cLoads.isEmpty();
    }

    public boolean isUndoable() {
        return true;
    }

    public String toString() {
        return (m_cLoads.isEmpty())? "No loads assigned." 
                : m_cLoads.size() + " loads assigned.";
    }
}
