/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmds.sec;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import rubeam.RuBeam;
import rubeam.cmps.ComponentElement;
import rubeam.model.Structure;
import rubeam.model.sections.AbstractSection;
import cib.cad.db.Database;
import cib.cad.db.comp.Component;
import cib.cad.kernel.CmdLineDialog;
import cib.cad.kernel.Kernel;
import cib.cad.kernel.UserInterface;
import cib.util.cmd.Cmd;
import cib.util.cmd.CmdAbortedException;
import cib.util.coll.ClassFilter;
import cib.util.coll.CollUtilities;
import cib.util.coll.ObservableSet;

public class AssignSection implements Cmd {

    private Map<ComponentElement, AbstractSection> m_map = new HashMap<ComponentElement, AbstractSection>();
    private Structure m_structure = null;
    private AbstractSection m_section = null;
    private Object m_input;

    public void doCmd(Object context) throws CmdAbortedException {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        Set<Component> selSet = db.getSelectSet();
        m_structure = RuBeam.getStructure();
    
        // Iterators
        Iterator<Component> eit;
        ComponentElement cElement = null;
        
        // macro mode
        if (krnl.getArguments().size() == 2/* krnl.hasArgument() */) {
            String secName = krnl.readString("Enter section name");
            // find section
            for (AbstractSection sec : m_structure.readSections()) {
                if (sec.getName().equals(secName)) {
                    m_section = sec;
                    break;
                }
            }
            if (m_section == null) {
                System.err.println("Section '" + secName + "' not found!");
                return;
            }
            String elmName = krnl.readString("Enter element name");
            // find element
            eit = CollUtilities.filterableIterator(compSet.iterator(),
                    new ClassFilter(ComponentElement.class));
            if (!eit.hasNext()) {
                System.err.println("Element '" + elmName + "' not found!");
                return;
            }
            while (eit.hasNext()) {
                ComponentElement cElm = (ComponentElement) eit.next();
                if (cElm.getElement().getName().equals(elmName)) {
                    cElement = cElm;
                    break;
                }
            }
            if (cElement == null) {
                System.err.println("Element '" + elmName + "' not found!");
                return;
            }

            // save old section
            m_map.put(cElement, cElement.getElement().getSection());

            // set new section
            m_structure.putSection(cElement.getElement(), m_section);
            compSet.wasChanged(cElement);
            
        }
        // interactive mode
        else {
            if (selSet.isEmpty()) {
                System.err.println("No components selected!");
                return;
            }
            final UserInterface ui = krnl.getUserInterface();
            eit = CollUtilities.filterableIterator(selSet.iterator(),
                    new ClassFilter(ComponentElement.class));
            if (!eit.hasNext()) {
                System.err.println("No element selected!");
                return;
            }
            String secName = null;
            if (krnl.hasArgument()) {
                secName = krnl.readString("Enter section name", "Default");
                // find section
                for (AbstractSection sec : m_structure.readSections()) {
                    if (sec.getName().equals(secName)) {
                        m_section = sec;
                        break;
                    }
                }
            }
            else {
                Object input = null;
                CmdLineDialog dlg = ui.getCmdLineDialog();
                dlg = ui.getCmdLineDialog();
                dlg.setPrompt("Enter section name");

                dlg.setDefaultText(AbstractSection.getDefaultSection()
                        .getName());
                List<AbstractSection> secList = m_structure.readSections();
                for (AbstractSection sec : secList) {
                    dlg.addChoice(sec.getName());
                }
                dlg.addListener(new CmdLineDialog.Listener() {
                    public void textEntered(String text) {
                        m_input = text;
                        ui.disposeCmdLineDialog();
                    }

                    public void choiceSelected(int iChoice) {
                        m_input = iChoice;
                        ui.disposeCmdLineDialog();
                    }

                    public void dialogAborted() {
                        m_input = null;
                        ui.disposeCmdLineDialog();
                    }
                });
                ui.showCmdLineDialog(dlg);
                
                if (m_input instanceof String) {
                    secName = m_input.toString();
                    // find section
                    for (AbstractSection sec : m_structure.readSections()) {
                        if (sec.getName().equals(secName)) {
                            m_section = sec;
                            break;
                        }
                    }

                }
                else if (m_input instanceof Integer) {
                    m_section = secList.get((Integer) m_input);
                    secName = m_section.getName();
                }
                else {
                    throw new CmdAbortedException();
                }
            }
            
            if (m_section == null) {
                System.err.println("Section '" + secName + "' not found!");
                return;
            }
            while (eit.hasNext()) {
                cElement = (ComponentElement) eit.next();
                
                // save old section
                m_map.put(cElement, cElement.getElement().getSection());
                
                // set new section
                m_structure.putSection(cElement.getElement(), m_section);
                compSet.wasChanged(cElement);
             }
        }
        System.out.println(this);
                    
    }

    public void undoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        
        for (Map.Entry<ComponentElement, AbstractSection> entry : m_map.entrySet()) {
            m_structure.putSection(entry.getKey().getElement(), entry.getValue());
            compSet.wasChanged(entry.getKey());
        }
    }

    public void redoCmd(Object context) {
        Kernel krnl = (Kernel) context;
        Database db = krnl.getDatabase();
        ObservableSet<Component> compSet = db.getComponentSet();
        
        for (Map.Entry<ComponentElement, AbstractSection> entry : m_map.entrySet()) {
            m_structure.putSection(entry.getKey().getElement(), m_section);
            compSet.wasChanged(entry.getKey());
        }
    }

    public boolean changesState() {
        return (m_section != null);
    }

    public boolean isUndoable() {
        return true;
    }

    public String toString() {
        return "Set section '" + m_section.getName()+ "' for " + m_map.size() + " element(s).";
    }
}
