/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmps;

import java.awt.Color;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.NoSuchElementException;

import rubeam.model.Element;
import rubeam.model.loads.ElementConLoadAng;
import cib.cad.db.comp.Component;
import cib.cad.db.comp.ComponentAdapter;
import cib.cad.db.feature.AngleFeature;
import cib.cad.db.feature.DistanceFeature;
import cib.cad.db.feature.DoubleFeature;
import cib.cad.db.feature.Feature;
import cib.util.AttributedShape;
import cib.util.CoordSpace;
import cib.util.coll.NamedListIterator;
import cib.util.coll.NamedListIteratorAdapter;
import cib.util.geo.Vector2D;

/**
 * @author Christian Koch, Emanuel Bombasaro
 */
public class ComponentElementConLoadAng extends ComponentAdapter implements ComponentElementLoad {
    private static final long serialVersionUID = 0L;

    protected static final double TEXT_DIST = 2.5; // [mm]
    public static final double LOAD_DISTANCE = 1.0;
    private ElementConLoadAng m_load;
    private double m_xi, m_yi, m_xk, m_yk, m_x, m_y;

    public ComponentElementConLoadAng(ElementConLoadAng load, Point2D pi, Point2D pk) {
        m_load = load;
        Element beam = load.getElement();
        if (beam != null) {
            setPointI(beam.getNodeI().getPoint());
            setPointK(beam.getNodeK().getPoint());
        }
        else {
            setPointI(pi);
            setPointK(pk);
        }
    }

    public Object clone() {
        ComponentElementConLoadAng clone = (ComponentElementConLoadAng) super.clone();
        clone.m_load = (ElementConLoadAng) this.m_load.clone();
        return clone;
    }

    public void assign(Component rhs) {
        ComponentElementConLoadAng rhsComp = (ComponentElementConLoadAng) rhs;
        ElementConLoadAng rhsLoad = rhsComp.m_load;
        m_load.setF(rhsLoad.getF());
        m_load.setAngle(rhsLoad.getAngle());
        m_load.setLi(rhsLoad.getLi());       
        super.assign(rhs);
    }

    public ElementConLoadAng getLoad() {
        return m_load;
    }

    public Point2D getPointI() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        if (m_load.getElement() == null) {
            ptU = new Point2D.Double(m_xi, m_yi);
        }
        else {
            ptU = m_load.getElement().getNodeI().getPoint();
            m_xi = ptU.getX();
            m_yi = ptU.getY();
        }
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public Point2D getPointK() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        if (m_load.getElement() == null)
            ptU = new Point2D.Double(m_xk, m_yk);
        else {
            ptU = m_load.getElement().getNodeK().getPoint();
            m_xk = ptU.getX();
            m_yk = ptU.getY();
        }
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public Point2D getPointM() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        
        m_x = m_xi+m_load.getLi()/getPointI().distance(getPointK())*(m_xk-m_xi);
        m_y = m_yi+m_load.getLi()/getPointI().distance(getPointK())*(m_yk-m_yi);
        ptU = new Point2D.Double(m_x, m_y);
        
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }
    
    public void setPointI(double x, double y) {
        setPointI(new Point2D.Double(x, y));
    }
    public void setPointI(Point2D pt) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D _pt = ucrd.transform(pt, null);
        m_xi = _pt.getX();
        m_yi = _pt.getY();

        // Notification
        _notifyWasChanged();
    }
    public void setPointK(double x, double y) {
        setPointK(new Point2D.Double(x, y));
    }
    public void setPointK(Point2D pt) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D _pt = ucrd.transform(pt, null);
        m_xk = _pt.getX();
        m_yk = _pt.getY();

        // Notification
        _notifyWasChanged();
    }
    public double getLi() {
        return m_load.getLi();
        
    }
    public void setLi(double nLi) {
        m_load.setLi(nLi);
        // Notification
        _notifyWasChanged();
    }

    private AttributedShape getLoadShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

        double _delta = LOAD_DISTANCE * scale;

        Path2D path = new Path2D.Double();

        AttributedShape as = _attributeShape(new AttributedShape(path));
        if (m_load.getElement() != null) {
            if (m_load.getElement().getLength() < m_load.getLi())
                as.setDrawPaint(Color.RED);
            else
                as.setDrawPaint(ComponentElementLoad.ELEMENTLOAD_COLOR);
        }
        else {
            as.setDrawPaint(Color.LIGHT_GRAY);
        }

        double F = m_load.getF();
        double ang = m_load.getAngle();
        while (ang >= 2 * Math.PI){
            ang -= 2 * Math.PI;
        }
        while (ang < 0){
            ang += 2 * Math.PI;
        }
        Vector2D edir = new Vector2D(getPointI(), getPointK());
        double angE = Vector2D.X_UNIT.getAngleCCW(edir);
        edir.normalize();
        edir.left();
        edir.scaleBy(_delta/2);

        // position line
        Vector2D cv = new Vector2D(getPointM());
        Vector2D cv2 = new Vector2D(cv);
        cv2.add(edir);
        Vector2D cv3 = new Vector2D(cv);
        edir.right().right();
        cv3.add(edir);
        Line2D s = new Line2D.Double(cv2, cv3);
        path.append(s, false);
        // offset and force line
        edir.normalize();
        edir.left().left();
        edir.scaleBy(_delta + _delta);

        if (ang <= Math.PI) {
            Vector2D startP = new Vector2D(cv);
            startP.add(edir);
            Vector2D endP = new Vector2D(startP);
            Vector2D vec = new Vector2D(ang+angE);
            vec.scaleBy(_delta + _delta);
            endP.add(vec).add(vec);
            Line2D line = new Line2D.Double(startP, endP);
            path.append(line, false);
            if (F > 0) {
                Vector2D ar1 = new Vector2D(ang + angE + Math.PI - Math.toRadians(30));
                ar1.scaleBy(_delta);
                Vector2D p1 = new Vector2D(endP);
                p1.add(ar1);
                Line2D l1 = new Line2D.Double(endP, p1);
                path.append(l1, false);
                Vector2D ar2 = new Vector2D(ang + angE + Math.PI + Math.toRadians(30));
                ar2.scaleBy(_delta);
                Vector2D p2 = new Vector2D(endP);
                p2.add(ar2);
                Line2D l2 = new Line2D.Double(endP, p2);
                path.append(l2, false);
            }
            else {
                Vector2D ar1 = new Vector2D(ang + angE + Math.toRadians(30));
                ar1.scaleBy(_delta);
                Vector2D p1 = new Vector2D(startP);
                p1.add(ar1);
                Line2D l1 = new Line2D.Double(startP, p1);
                path.append(l1, false);
                Vector2D ar2 = new Vector2D(ang + angE - Math.toRadians(30));
                ar2.scaleBy(_delta);
                Vector2D p2 = new Vector2D(startP);
                p2.add(ar2);
                Line2D l2 = new Line2D.Double(startP, p2);
                path.append(l2, false);
            }
        }
        else {
            Vector2D startP = new Vector2D(cv);
            startP.add(edir);
            Vector2D endP = new Vector2D(startP);
            Vector2D vec = new Vector2D(ang+angE);
            vec.scaleBy(- _delta - _delta);
            endP.add(vec).add(vec);
            Line2D line = new Line2D.Double(startP, endP);
            path.append(line, false);  
            
            if (F > 0) {
                Vector2D ar1 = new Vector2D(ang + angE + Math.PI - Math.toRadians(30));
                ar1.scaleBy(_delta);
                Vector2D p1 = new Vector2D(startP);
                p1.add(ar1);
                Line2D l1 = new Line2D.Double(startP, p1);
                path.append(l1, false);
                Vector2D ar2 = new Vector2D(ang + angE + Math.PI + Math.toRadians(30));
                ar2.scaleBy(_delta);
                Vector2D p2 = new Vector2D(startP);
                p2.add(ar2);
                Line2D l2 = new Line2D.Double(startP, p2);
                path.append(l2, false);
            }
            else {
                Vector2D ar1 = new Vector2D(ang + angE + Math.toRadians(30));
                ar1.scaleBy(_delta);
                Vector2D p1 = new Vector2D(endP);
                p1.add(ar1);
                Line2D l1 = new Line2D.Double(endP, p1);
                path.append(l1, false);
                Vector2D ar2 = new Vector2D(ang + angE - Math.toRadians(30));
                ar2.scaleBy(_delta);
                Vector2D p2 = new Vector2D(endP);
                p2.add(ar2);
                Line2D l2 = new Line2D.Double(endP, p2);
                path.append(l2, false);
            }
        }
        return as;
    }

    public NamedListIterator<AttributedShape> shapeIterator() {
        return NamedListIteratorAdapter
        .singletonNamedListIterator(getLoadShape());
    }

    /*private AttributedText getText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        Vector2D vec = new Vector2D(getStart(), getEnd());
        Vector2D ext = new Vector2D(vec);
        ext.normalize();
        ext.left();
        ext.scaleBy(TEXT_DIST * scale);
        vec.scaleBy(0.5);
        Vector2D vCtr = new Vector2D(getStart());
        vCtr.add(vec);
        vCtr.add(ext);
        Point2D ptW = u2w.transform(vCtr, null);

        AffineTransform trf = AffineTransform.getTranslateInstance(ptW.getX(),
                ptW.getY());

        AttributedString string = new AttributedString(m_beam.getName());
        AttributedText text = new AttributedText(string, trf);
        text.setJustification(AttributedText.JUST_CENTER_CENTER);
        text.transformBy(w2u);
        return _attributeText(text);
    }

    public NamedListIterator<AttributedText> textIterator() {
        return NamedListIteratorAdapter.singletonNamedListIterator(getText());
    }*/

    public Point2D getControlPoint(int name) throws IllegalArgumentException {
        switch (name) {
        case 0:
            return getPointI();
        case 1:
            return getPointK();
        default:
            throw new IllegalArgumentException();
        }
    }

    public void setControlPoint(Point2D pnt, int name)
    throws UnsupportedOperationException {
        Point2D p = pnt;
        switch (name) {
        case 0:
            if (m_load.getElement() == null) {
                setPointI(p);
            }
            break;
        case 1:
            if (m_load.getElement() == null) {
                setPointK(p);
            }
            break;
        default:
            throw new IllegalArgumentException();
        }
    }

    public NamedListIterator<Point2D> controlPointIterator() {
        return new NamedListIteratorAdapter<Point2D>() {
            protected int _size() {
                return 2;
            }

            protected Point2D _get(int index) {
                switch (index) {
                case 0:
                    return getPointI();
                case 1:
                    return getPointK();
                default:
                    throw new IllegalArgumentException();
                }
            }

            protected void _set(int index, Point2D value) {
                Point2D p = (Point2D) value;
                switch (index) {
                case 0:
                    if (m_load.getElement() == null) {
                        setPointI(p);
                    }
                    break;
                case 1:
                    if (m_load.getElement() == null) {
                        setPointK(p);
                    }
                    break;
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }
    public void transformBy(AffineTransform mat) {
        if (m_load.getElement() == null) {
            // Transform point
            Point2D pt = mat.transform(getPointI(), null);
            setPointI(pt);
            pt = mat.transform(getPointK(), null);
            setPointK(pt);
        }
    }

    // Features
    private static final String[] FEATURE_NAMES = { 
        "F", // 0
        "\u03b1", // 1
        "li", // 2
    };

    public boolean hasFeature(String name) {
        for (int i = 0; i < FEATURE_NAMES.length; i++) {
            if (name.equals(FEATURE_NAMES[i]))
                return true;
        }
        return super.hasFeature(name);
    }

    public Feature getFeature(String name) {

        if (name.equals(FEATURE_NAMES[0])) {
            return new DoubleFeature(FEATURE_NAMES[0], m_load.getF());
        }
        else if (name.equals(FEATURE_NAMES[1])) {
            return new AngleFeature(FEATURE_NAMES[1], m_load.getAngle());
        }
        else if (name.equals(FEATURE_NAMES[2])) {
            return new DistanceFeature(FEATURE_NAMES[2], m_load.getLi());
        }
        else
            return super.getFeature(name);
    }

    public void setFeature(Feature feature) {
        if (feature.getName().equals(FEATURE_NAMES[0])) {
            m_load.setF(((DoubleFeature)feature).getDoubleValue());
        }
        else if (feature.getName().equals(FEATURE_NAMES[1])) {
            m_load.setAngle(((AngleFeature)feature).getAngle());
        }
        else if (feature.getName().equals(FEATURE_NAMES[2])) {
            m_load.setLi(((DistanceFeature)feature).getDistance());
        }
        else {
            super.setFeature(feature);
            return;
        }
        _notifyWasChanged();
    }

    public void addFeature(Feature feature) {
        throw new UnsupportedOperationException();
    }

    public Iterator<Feature> featureIterator() {

        final Iterator<Feature> itThis = new Iterator<Feature>() {
            private int m_index = 0;

            public boolean hasNext() {
                return m_index < FEATURE_NAMES.length;
            }

            public Feature next() {
                if (hasNext())
                    return getFeature(FEATURE_NAMES[m_index++]);
                throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

        final Iterator<Feature> itBase = super.featureIterator();

        return new Iterator<Feature>() {
            public boolean hasNext() {
                return itBase.hasNext() | itThis.hasNext();
            }

            public Feature next() {
                if (itBase.hasNext())
                    return itBase.next();
                else if (itThis.hasNext())
                    return itThis.next();
                else
                    throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

}
