/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.cmps;

import java.awt.Color;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Path2D;
import java.awt.geom.Point2D;
import java.util.Iterator;
import java.util.NoSuchElementException;

import rubeam.model.Element;
import rubeam.model.loads.ElementUniDistribLoadNor;
import cib.cad.db.comp.Component;
import cib.cad.db.comp.ComponentAdapter;
import cib.cad.db.feature.DoubleFeature;
import cib.cad.db.feature.Feature;
import cib.util.AttributedShape;
import cib.util.CoordSpace;
import cib.util.coll.NamedListIterator;
import cib.util.coll.NamedListIteratorAdapter;
import cib.util.geo.Vector2D;

/**
 * @author Christian Koch
 */
public class ComponentElementUniDistribLoadNor extends ComponentAdapter implements ComponentElementLoad {

    private static final long serialVersionUID = 0L;
    protected static final double TEXT_DIST = 2.5; // [mm]
    public static final double LOAD_DISTANCE = 1.0;
    private ElementUniDistribLoadNor m_load;
    private double m_xi, m_yi, m_xk, m_yk;
    
    public ComponentElementUniDistribLoadNor(ElementUniDistribLoadNor load, Point2D pi, Point2D pk) {
        m_load = load;
        Element beam = load.getElement();
        if (beam != null) {
            setPointI(beam.getNodeI().getPoint());
            setPointK(beam.getNodeK().getPoint());
        }
        else {
            setPointI(pi);
            setPointK(pk);
        }
    }

    public Object clone() {
        ComponentElementUniDistribLoadNor clone = (ComponentElementUniDistribLoadNor) super.clone();
        clone.m_load = (ElementUniDistribLoadNor) this.m_load.clone();
        /*List<BeamLoad> loadSet = clone.m_beam.getLoadSet();
        if (loadSet != null) {
            for (BeamLoad bl : loadSet) {
                bl.setBeam(clone.m_beam);
            }
        }*/
        return clone;
    }

    public void assign(Component rhs) {
        ComponentElementUniDistribLoadNor rhsComp = (ComponentElementUniDistribLoadNor) rhs;
        ElementUniDistribLoadNor rhsLoad = rhsComp.m_load;
        m_load.setQ0(rhsLoad.getQ0());
        super.assign(rhs);
    }

    public ElementUniDistribLoadNor getLoad() {
        return m_load;
    }

    public Point2D getPointI() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        if (m_load.getElement() == null) {
            ptU = new Point2D.Double(m_xi, m_yi);
        }
        else {
            ptU = m_load.getElement().getNodeI().getPoint();
            m_xi = ptU.getX();
            m_yi = ptU.getY();
        }
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public Point2D getPointK() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D ptU = null;
        if (m_load.getElement() == null)
            ptU = new Point2D.Double(m_xk, m_yk);
        else {
            ptU = m_load.getElement().getNodeK().getPoint();
            m_xk = ptU.getX();
            m_yk = ptU.getY();
        }
        try {
            ptU = ucrd.inverseTransform(ptU, null);
        } catch (NoninvertibleTransformException e) {
            // Should never happen
        }
        return ptU;
    }

    public void setPointI(double x, double y) {
        setPointI(new Point2D.Double(x, y));
    }
    public void setPointI(Point2D pt) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D _pt = ucrd.transform(pt, null);
        m_xi = _pt.getX();
        m_yi = _pt.getY();

        // Notification
        _notifyWasChanged();
    }
    public void setPointK(double x, double y) {
        setPointK(new Point2D.Double(x, y));
    }
    public void setPointK(Point2D pt) {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        AffineTransform ucrd = cs.getUserCrdTransform();
        Point2D _pt = ucrd.transform(pt, null);
        m_xk = _pt.getX();
        m_yk = _pt.getY();

        // Notification
        _notifyWasChanged();
    }
    
    private AttributedShape getLoadShape() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();

        Point2D sp = getPointI();
        Point2D ep = getPointK();
        double _delta = LOAD_DISTANCE * scale;

        Path2D path = new Path2D.Double();

        AttributedShape as = _attributeShape(new AttributedShape(path));
        if (m_load.getElement() != null) {
            as.setDrawPaint(ComponentElementLoad.ELEMENTLOAD_COLOR);
            as.setFillPaint(ComponentElementLoad.ELEMENTLOAD_COLOR);
        }
        else {
            as.setDrawPaint(Color.LIGHT_GRAY);
            as.setFillPaint(Color.LIGHT_GRAY);
        }

        ElementUniDistribLoadNor udl = m_load;// (UniDistribLoad) load;
        double q = udl.getQ0();
        Vector2D vdir = new Vector2D(sp, ep);
        vdir.normalize();
        vdir.left();
        vdir.scaleBy(_delta + _delta);
        Vector2D vsp1 = new Vector2D(sp);
        vsp1.add(vdir).add(vdir);
        Vector2D vsp2 = new Vector2D(sp);
        vsp2.add(vdir).add(vdir).add(vdir);
        Vector2D vep1 = new Vector2D(ep);
        vep1.add(vdir).add(vdir);
        Vector2D vep2 = new Vector2D(ep);
        vep2.add(vdir).add(vdir).add(vdir);

        path.moveTo(vsp1.x, vsp1.y);
        path.lineTo(vsp2.x, vsp2.y);
        path.lineTo(vep2.x, vep2.y);
        path.lineTo(vep1.x, vep1.y);
        path.closePath();

        double ang = Vector2D.X_UNIT.getAngleCCW(vdir);
        if (q > 0) {
            Vector2D ar1 = new Vector2D(ang - Math.toRadians(30));
            ar1.scaleBy(_delta);
            Vector2D sp1 = new Vector2D(vsp1);
            sp1.add(ar1);
            Line2D l1 = new Line2D.Double(vsp1, sp1);
            path.append(l1, false);
            Vector2D ar2 = new Vector2D(ang + Math.toRadians(30));
            ar2.scaleBy(_delta);
            Vector2D sp2 = new Vector2D(vsp1);
            sp2.add(ar2);
            Line2D l2 = new Line2D.Double(vsp1, sp2);
            path.append(l2, false);

            Vector2D ep1 = new Vector2D(vep1);
            ep1.add(ar1);
            l1 = new Line2D.Double(vep1, ep1);
            path.append(l1, false);
            Vector2D ep2 = new Vector2D(vep1);
            ep2.add(ar2);
            l2 = new Line2D.Double(vep1, ep2);
            path.append(l2, false);
        }
        else {
            Vector2D ar1 = new Vector2D(ang - Math.toRadians(150));
            ar1.scaleBy(_delta);
            Vector2D sp1 = new Vector2D(vsp2);
            sp1.add(ar1);
            Line2D l1 = new Line2D.Double(vsp2, sp1);
            path.append(l1, false);
            Vector2D ar2 = new Vector2D(ang + Math.toRadians(150));
            ar2.scaleBy(_delta);
            Vector2D sp2 = new Vector2D(vsp2);
            sp2.add(ar2);
            Line2D l2 = new Line2D.Double(vsp2, sp2);
            path.append(l2, false);

            Vector2D ep1 = new Vector2D(vep2);
            ep1.add(ar1);
            l1 = new Line2D.Double(vep2, ep1);
            path.append(l1, false);
            Vector2D ep2 = new Vector2D(vep2);
            ep2.add(ar2);
            l2 = new Line2D.Double(vep2, ep2);
            path.append(l2, false);
        }
        return as;
    }

    public NamedListIterator<AttributedShape> shapeIterator() {
        return NamedListIteratorAdapter
                .singletonNamedListIterator(getLoadShape());
    }

    /*private AttributedText getText() {
        final CoordSpace cs = CoordSpace.getCoordSpace();
        double scale = cs.getScale() / cs.getMillimetersPerNaturalUnit();
        AffineTransform w2u = cs.getWorldToUserTransform();
        AffineTransform u2w = cs.getUserToWorldTransform();
        Vector2D vec = new Vector2D(getStart(), getEnd());
        Vector2D ext = new Vector2D(vec);
        ext.normalize();
        ext.left();
        ext.scaleBy(TEXT_DIST * scale);
        vec.scaleBy(0.5);
        Vector2D vCtr = new Vector2D(getStart());
        vCtr.add(vec);
        vCtr.add(ext);
        Point2D ptW = u2w.transform(vCtr, null);

        AffineTransform trf = AffineTransform.getTranslateInstance(ptW.getX(),
                ptW.getY());

        AttributedString string = new AttributedString(m_beam.getName());
        AttributedText text = new AttributedText(string, trf);
        text.setJustification(AttributedText.JUST_CENTER_CENTER);
        text.transformBy(w2u);
        return _attributeText(text);
    }

    public NamedListIterator<AttributedText> textIterator() {
        return NamedListIteratorAdapter.singletonNamedListIterator(getText());
    }*/

    public Point2D getControlPoint(int name) throws IllegalArgumentException {
        switch (name) {
        case 0:
            return getPointI();
        case 1:
            return getPointK();
        default:
            throw new IllegalArgumentException();
        }
    }

    public void setControlPoint(Point2D pnt, int name)
            throws UnsupportedOperationException {
        Point2D p = pnt;
        switch (name) {
        case 0:
            if (m_load.getElement() == null) {
                setPointI(p);
            }
            break;
        case 1:
            if (m_load.getElement() == null) {
                setPointK(p);
            }
            break;
        default:
            throw new IllegalArgumentException();
        }
    }

    public NamedListIterator<Point2D> controlPointIterator() {
        return new NamedListIteratorAdapter<Point2D>() {
            protected int _size() {
                return 2;
            }

            protected Point2D _get(int index) {
                switch (index) {
                case 0:
                    return getPointI();
                case 1:
                    return getPointK();
                default:
                    throw new IllegalArgumentException();
                }
            }

            protected void _set(int index, Point2D value) {
                Point2D p = (Point2D) value;
                switch (index) {
                case 0:
                    if (m_load.getElement() == null) {
                        setPointI(p);
                    }
                    break;
                case 1:
                    if (m_load.getElement() == null) {
                        setPointK(p);
                    }
                    break;
                default:
                    throw new IllegalArgumentException();
                }
            }
        };
    }
    public void transformBy(AffineTransform mat) {
        if (m_load.getElement() == null) {
            // Transform point
            Point2D pt = mat.transform(getPointI(), null);
            setPointI(pt);
            pt = mat.transform(getPointK(), null);
            setPointK(pt);
        }
    }

    // Features
    private static final String[] FEATURE_NAMES = { 
        "F", // 0
    };

    public boolean hasFeature(String name) {
        for (int i = 0; i < FEATURE_NAMES.length; i++) {
            if (name.equals(FEATURE_NAMES[i]))
                return true;
        }
        return super.hasFeature(name);
    }

    public Feature getFeature(String name) {

        if (name.equals(FEATURE_NAMES[0])) {
            return new DoubleFeature(FEATURE_NAMES[0], m_load.getQ0());
        }
        else
            return super.getFeature(name);
    }

    public void setFeature(Feature feature) {
        if (feature.getName().equals(FEATURE_NAMES[0])) {
            m_load.setQ0(((DoubleFeature)feature).getDoubleValue());
        }
        else {
            super.setFeature(feature);
            return;
        }
        _notifyWasChanged();
    }

    public void addFeature(Feature feature) {
        throw new UnsupportedOperationException();
    }

    public Iterator<Feature> featureIterator() {

        final Iterator<Feature> itThis = new Iterator<Feature>() {
            private int m_index = 0;

            public boolean hasNext() {
                return m_index < FEATURE_NAMES.length;
            }

            public Feature next() {
                if (hasNext())
                    return getFeature(FEATURE_NAMES[m_index++]);
                throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };

        final Iterator<Feature> itBase = super.featureIterator();

        return new Iterator<Feature>() {
            public boolean hasNext() {
                return itBase.hasNext() | itThis.hasNext();
            }

            public Feature next() {
                if (itBase.hasNext())
                    return itBase.next();
                else if (itThis.hasNext())
                    return itThis.next();
                else
                    throw new NoSuchElementException();
            }

            public void remove() {
                throw new UnsupportedOperationException();
            }
        };
    }

}
