/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.gui;

import java.awt.Color;
import java.util.List;

import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.ListSelectionModel;
import javax.swing.table.DefaultTableModel;
import javax.swing.table.TableColumn;

import rubeam.RuBeam;
import rubeam.model.Node;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.loads.NodalLoadAng;
import cib.cad.kernel.InputDeviceAdapter;
import cib.cad.kernel.TextToken;

public class NodalLoadAngTablePanel extends JScrollPane {

    private static final long serialVersionUID = 1L;
    private static final NumberCellRenderer numberRenderer = new NumberCellRenderer();
    private static final String[] colNames = new String[] {
        "Name", "Node", RuBeam.NODALLOADANG_PREFIX, "\u03b1"
    };
    // GUI
    private JTable table = new JTable();
    private TableModel model = new TableModel();
    
    // Input device
    private InputDeviceAdapter inputDevice = new InputDeviceAdapter();
    
    public NodalLoadAngTablePanel(InputDeviceAdapter inputDevice) {
        super();
        this.inputDevice = inputDevice;
        initialize();
    }
    
    private void initialize() {
        this.setViewportView(table);
        this.setVerticalScrollBarPolicy(VERTICAL_SCROLLBAR_AS_NEEDED);
        this.setHorizontalScrollBarPolicy(HORIZONTAL_SCROLLBAR_AS_NEEDED);
        table.setModel(model);
        table.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);
        table.setGridColor(Color.LIGHT_GRAY);
        TableColumn tc = table.getColumnModel().getColumn(2);
        tc.setCellRenderer(numberRenderer);
        tc = table.getColumnModel().getColumn(3);
        tc.setCellRenderer(numberRenderer);
        setCellEditors();        
    }
    
    public void updatePanel() {
        model.fireTableDataChanged();
    }
    public AbstractNodeLoad getSelectedNodeLoad() {
        int row = table.getSelectedRow();
        if (row != -1)
            return RuBeam.getStructure().readNodeLoads(NodalLoadAng.class).get(row);
        else 
            return null;
    }
    
    private void setCellEditors() {
        TableColumn tc = table.getColumnModel().getColumn(1);
        tc.setCellEditor(new NodeListCellEditor());
    }
    
    class TableModel extends DefaultTableModel {

        private static final long serialVersionUID = 1L;
        
        public int getColumnCount() {
            return colNames.length;
        }
        public String getColumnName(int column) {
            return colNames[column];
        }
        
        public int getRowCount() {
            int size = RuBeam.getStructure().readNodeLoads(
                    NodalLoadAng.class).size();
            return size;
        }
        
        public Class<?> getColumnClass(int columnIndex) {
            switch (columnIndex) {
            case 0:
                return String.class;
            case 1:
                return String.class;
            case 2:
                return Double.class;
            case 3:
                return Double.class;
            default:
                throw new IllegalStateException();
            }
        }
        
        public boolean isCellEditable(int row, int column) {
            return true;
        }
        
        public Object getValueAt(int rowIndex, int columnIndex) {
            Structure structure = RuBeam.getStructure();
            List<AbstractNodeLoad> nodeLoadList = structure.readNodeLoads(
                    NodalLoadAng.class);
            if (rowIndex >= nodeLoadList.size())
                return null;
            NodalLoadAng nodeLoad = (NodalLoadAng)nodeLoadList.get(rowIndex);
            switch (columnIndex) {
            case 0:
                return nodeLoad.getName();
            case 1:
                Node node = structure.getAssignedNode(nodeLoad); 
                if (node == null)
                    return "";
                return node.getName();
            case 2:
                return Double.valueOf(nodeLoad.getP());
            case 3:
                return Double.valueOf(Math.toDegrees(nodeLoad.getAngle()));
            default:
                throw new IllegalStateException();
            }
        }
        public void setValueAt(Object value, int rowIndex, int columnIndex) {
            switch (columnIndex) {
            case 0: // set name
                try {
                    String newName = value.toString();
                    List<AbstractNodeLoad> loadList = RuBeam.getStructure()
                            .readNodeLoads(NodalLoadAng.class);
                    String loadName = loadList.get(rowIndex).getName();
                    String fire = "setnodalloadname " 
                        + TextToken.stringToLiteral(loadName) + " " 
                        + TextToken.stringToLiteral(newName);
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 1: // set node
                try {
                    if (value != null) {
                        String newNode = value.toString();
                        List<AbstractNodeLoad> loadList = RuBeam.getStructure()
                                .readNodeLoads(NodalLoadAng.class);
                        String loadName = loadList.get(rowIndex).getName();
                        String fire = "assignnodalload "
                                + TextToken.stringToLiteral(loadName) + " "
                                + TextToken.stringToLiteral(newNode);
                        fireInput(fire);
                    }
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 2: // set value P
                try {
                    double newP = Double.parseDouble(value.toString());
                    List<AbstractNodeLoad> loadList = RuBeam.getStructure()
                            .readNodeLoads(NodalLoadAng.class);
                    String loadName = loadList.get(rowIndex).getName();
                    String fire = "setnodalloadvalue " 
                        + TextToken.stringToLiteral(loadName) + " " + newP;
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            case 3:  // set angle
                try {
                    double newAlpha = Double.parseDouble(value.toString());
                    List<AbstractNodeLoad> loadList = RuBeam.getStructure()
                            .readNodeLoads(NodalLoadAng.class);
                    String loadName = loadList.get(rowIndex).getName();
                    String fire = "setnodalloadangle " 
                        + TextToken.stringToLiteral(loadName) + " " + newAlpha;
                    fireInput(fire);
                }
                catch (NumberFormatException e) {
                    // do nothing
                }
                model.fireTableCellUpdated(rowIndex, columnIndex);
                break;
            default: 
                throw new IllegalStateException();
            }
        }
    }
    
    private void fireInput(String input) {
        if (inputDevice.isEnabled()) {
            inputDevice.fireInterruptingInput(input);
        }
            
    }
}
