/*
 * Copyright (C) 2009 Emanuel Bombasaro, Christian Koch
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT 
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with 
 * this program; if not, look at 
 * http://www.gnu.org/licenses/old-licenses/gpl-2.0.html or write to the Free 
 * Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, 
 * MA  02110-1301, USA.
 * 
 * Contact address: emanuel.bombasaro@cademia.org, christian.koch@cademia.org
 * Homepage       : http://www.rubeam.cademia.org/
 */

package rubeam.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.Image;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.File;
import java.io.IOException;
import java.io.PrintWriter;
import java.net.URL;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JDialog;
import javax.swing.JFileChooser;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JPopupMenu;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;
import javax.swing.filechooser.FileNameExtensionFilter;

import rubeam.RuBeam;
import rubeam.Util;
import rubeam.model.Element;
import rubeam.model.Node;
import rubeam.model.Structure;
import rubeam.model.loads.AbstractElementLoad;
import rubeam.model.loads.AbstractNodeLoad;
import rubeam.model.sections.AbstractSection;
import cib.cad.kernel.InputDeviceAdapter;
import cib.cad.view.GraphicalUserInterface;

public class RuBeamInspector extends JDialog {

    private static final long serialVersionUID = 1L;

    // Panels
    private JPanel nodePanel = new JPanel();
    private NodeTablePanel nodeTablePanel;
    private JPanel elementPanel = new JPanel();
    private ElementTablePanel elementTablePanel;
    private JPanel nodeLoadPanel = new JPanel();
    private NodalLoadAngTablePanel nodalLoadAngTablePanel;
    private JPanel elementLoadPanel = new JPanel();
    private ElementUniDistribLoadNorTablePanel uniDistribLoadNorTablePanel;
    private ElementConLoadAngTablePanel conLoadAngTablePanel;
    private JPanel sectionPanel = new JPanel();
    private GeneralSectionTablePanel linearGeneralSectionTablePanel;
    private LinearRectangleSectionTablePanel linearRectangleSectionTablePanel;
    private LinearIProfileSectionTablePanel linearIProfileSectionTablePanel;
    private LinearSandwichSectionTablePanel linearSandwichSectionTablePanel; 
//    private JPanel settingsPanel = new JPanel();
    private JPanel structurePanel = new JPanel();
    private JTextField visualizeTextField = new JTextField(3);
    private JTextField resultTextField = new JTextField(3);
    private HtmlPanel structureOutputPanel;
    private JPanel resultsPanel = new JPanel();
    private HtmlPanel resultsOutputPanel;
    private JTabbedPane tabbedPane = new JTabbedPane();
    
    // Input Device
    private InputDeviceAdapter inputDevice = new InputDeviceAdapter();

    // Mouse Listener
    private MouseListener mouseListener = null;

    private static final String TITLE = "ruBeam Inspector";
    private static final int INPUT = 0;
    private static final int COMMAND = 1;
    private static final int INTERRUPTING_COMMAND = 2;
    private static Dimension m_buttonDim = new Dimension(30, 25);
    private static boolean MAC_OS = false;
    private static boolean WINDOWS = false;
    private static boolean LINUX = false;
    static {
        String os = System.getProperty("os.name").toLowerCase();
        if (os.startsWith("mac os x"))
            MAC_OS = true;
        else if (os.startsWith("windows"))
            WINDOWS = true;
        else if (os.startsWith("linux"))
            LINUX = true;
    }

    public RuBeamInspector(GraphicalUserInterface gui) {
        super(gui);
        setTitle(TITLE);
        setModal(false);
        setPreferredSize(new Dimension(400, 300));
        setDefaultCloseOperation(JDialog.DISPOSE_ON_CLOSE); // only for testing
        setResizable(true);

        gui.addInputDevice(inputDevice);

        mouseListener = new MouseAdapter() {
            public void mouseEntered(MouseEvent e) {
                JButton b = (JButton) e.getSource();
                if (b.isEnabled())
                    b.setBorderPainted(true);
            }

            public void mouseExited(MouseEvent e) {
                ((JButton) e.getSource()).setBorderPainted(false);
            }
        };

        initialize();
        pack();
    }

    private void initialize() {
        initNodePanel();
        tabbedPane.addTab("Nodes", null, nodePanel, "Edit nodes");
        initElementPanel();
        tabbedPane.addTab("Elements", null, elementPanel, "Edit elements");
        initSectionPanel();
        tabbedPane.addTab("Sections", null, sectionPanel, "Edit sections");
        initNodalLoadPanel();
        tabbedPane.addTab("Node loads", null, nodeLoadPanel, "Edit node loads");
        initElementLoadPanel();
        tabbedPane.addTab("Element loads", null, elementLoadPanel,
                "Edit element loads");
        initStructurePanel();
        tabbedPane.addTab("Structure", null, structurePanel, "Dump structure");
        initResultsPanel();
        tabbedPane.addTab("Results", null, resultsPanel, "Dump results");
        /*initSettingsPanel();
        tabbedPane.addTab("Settings", null, settingsPanel, "Edit settings");*/

        tabbedPane.addChangeListener(new ChangeListener() {
           public void stateChanged(ChangeEvent e) {
               Component comp = tabbedPane.getSelectedComponent();
               if (comp == structurePanel)
                   structureOutputPanel.requestFocus();
               else if (comp == resultsPanel)
                   resultsOutputPanel.requestFocus();
           } 
        });
        
        add(tabbedPane);
    }

    private void initNodePanel() {
        nodePanel.setLayout(new BorderLayout());

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JButton button = _button("Add.png", mouseListener, new _CmdListener(
                "addnode"), "Add a new node to the structure.");
        buttonPanel.add(button);

        button = _button("Remove.png", mouseListener, null,
                "Remove the selected row.");
        buttonPanel.add(button);
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("removenode");
                }
                else if (inputDevice.isEnabled()) {
                    Node node = nodeTablePanel.getSelectedNode();
                    if (node != null) {
                        inputDevice.fireInterruptingInput("removenode "
                                + node.getName());
                    }
                }
            }
        });
        nodePanel.add(buttonPanel, BorderLayout.NORTH);

        // Center
        nodeTablePanel = new NodeTablePanel(inputDevice);
        nodePanel.add(nodeTablePanel, BorderLayout.CENTER);
    }

    private void initElementPanel() {
        elementPanel.setLayout(new BorderLayout());

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JButton button = _button("Add.png", mouseListener, new _CmdListener(
                "addelement"), "Add a new element to the structure.");
        buttonPanel.add(button);

        button = _button("Remove.png", mouseListener,
        /* new _CmdListener("remove") */null, "Remove the selected row.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("removeelement");
                }
                else if (inputDevice.isEnabled()) {
                    Element element = elementTablePanel.getSelectedElement();
                    if (element != null) {
                        inputDevice.fireInterruptingInput("removeelement "
                                + element.getName());
                    }
                }
            }
        });
        buttonPanel.add(button);

        elementPanel.add(buttonPanel, BorderLayout.NORTH);

        // Center
        elementTablePanel = new ElementTablePanel(inputDevice);
        elementPanel.add(elementTablePanel, BorderLayout.CENTER);
    }

    private void initNodalLoadPanel() {
        nodeLoadPanel.setLayout(new BorderLayout());

        // Center
        final JTabbedPane featurePane = new JTabbedPane();
        // featurePane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        nodalLoadAngTablePanel = new NodalLoadAngTablePanel(inputDevice);
        featurePane.addTab("Nodal angular load", null, nodalLoadAngTablePanel,
                "Edit nodal angular load.");

        nodeLoadPanel.add(featurePane, BorderLayout.CENTER);

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JButton button = _button("Add.png", mouseListener, null,
                "Add a new nodal load to the structure.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Component comp = featurePane.getSelectedComponent();
                if (comp == nodalLoadAngTablePanel)
                    inputDevice.fireInterruptingInput("addnodalloadang");
                // here add other loads
            } 
        });
        buttonPanel.add(button);

        /*final JPopupMenu popupAdd = new JPopupMenu();
        JMenuItem item = new JMenuItem("Nodal angular load");
        item.addActionListener(new _CmdListener("addnodalloadang"));
        popupAdd.add(item);

        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (!popupAdd.isVisible()) {
                    JButton button = (JButton) e.getSource();
                    popupAdd.show(button, 0, button.getHeight());
                }
                else
                    popupAdd.setVisible(false);
            }
        });*/

        button = _button("Remove.png", mouseListener, null,
                "Remove the selected row.");
        buttonPanel.add(button);
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("removenodalload");
                }
                else if (inputDevice.isEnabled()) {
                    java.awt.Component comp = featurePane
                            .getSelectedComponent();
                    AbstractNodeLoad load = null;
                    if (comp == nodalLoadAngTablePanel) {
                        load = nodalLoadAngTablePanel.getSelectedNodeLoad();
                    }
                    if (load != null) {
                        inputDevice.fireInterruptingInput("removenodalload "
                                + load.getName());
                    }
                }
            }
        });

        button = _button("Assign2.png", mouseListener, new _CmdListener(
                "assignnodalloads"), "Assign selected loads to selected node.");
        buttonPanel.add(button);

        nodeLoadPanel.add(buttonPanel, BorderLayout.NORTH);
    }

    private void initElementLoadPanel() {
        elementLoadPanel.setLayout(new BorderLayout());

        // Center
        final JTabbedPane featurePane = new JTabbedPane();
        // featurePane.setTabLayoutPolicy(JTabbedPane.SCROLL_TAB_LAYOUT);
        uniDistribLoadNorTablePanel = new ElementUniDistribLoadNorTablePanel(
                inputDevice);
        featurePane.addTab("Uni distributed normal load", null,
                uniDistribLoadNorTablePanel,
                "Edit uniformly distributed normal loads.");

        conLoadAngTablePanel = new ElementConLoadAngTablePanel(inputDevice);
        featurePane.addTab("Concentrated angular load", null,
                conLoadAngTablePanel, "Edit concentrated angular load.");

        elementLoadPanel.add(featurePane, BorderLayout.CENTER);

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JButton button = _button("Add.png", mouseListener, null,
                "Add a new element load to the structure.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                Component comp = featurePane.getSelectedComponent();
                if (comp == uniDistribLoadNorTablePanel)
                    inputDevice.fireInterruptingInput("addelementunidisloadnor");
                else if (comp == conLoadAngTablePanel)
                    inputDevice.fireInterruptingInput("addelementconload");
            } 
        });
        buttonPanel.add(button);

        /*final JPopupMenu popupAdd = new JPopupMenu();
        JMenuItem item = new JMenuItem("Uni distributed normal load");
        item.addActionListener(new _CmdListener("addelementunidisloadnor"));
        popupAdd.add(item);

        item = new JMenuItem("Concentrated angular load");
        item.addActionListener(new _CmdListener("addelementconload"));
        popupAdd.add(item);

        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (!popupAdd.isVisible()) {
                    JButton button = (JButton) e.getSource();
                    popupAdd.show(button, 0, button.getHeight());
                }
                else
                    popupAdd.setVisible(false);
            }
        });*/

        button = _button("Remove.png", mouseListener, null,
                "Remove the selected row.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("removeelementload");
                }
                else if (inputDevice.isEnabled()) {
                    java.awt.Component comp = featurePane
                            .getSelectedComponent();
                    AbstractElementLoad load = null;
                    if (comp == uniDistribLoadNorTablePanel) {
                        load = uniDistribLoadNorTablePanel
                                .getSelectedElementLoad();
                    }
                    else if (comp == conLoadAngTablePanel) {
                        load = conLoadAngTablePanel.getSelectedElementLoad();
                    }
                    if (load != null) {
                        inputDevice.fireInterruptingInput("removeelementload "
                                + load.getName());
                    }
                }
            }
        });
        buttonPanel.add(button);
        

        button = _button("Assign2.png", mouseListener, new _CmdListener(
                "assignelementloads"),
                "Assign selected loads to selected element.");
        buttonPanel.add(button);

        elementLoadPanel.add(buttonPanel, BorderLayout.NORTH);
    }

    private void initSectionPanel() {
        sectionPanel.setLayout(new BorderLayout());

        // Center
        final JTabbedPane featurePane = new JTabbedPane();
        linearGeneralSectionTablePanel = new GeneralSectionTablePanel(inputDevice);
        featurePane.addTab("General", null,
                linearGeneralSectionTablePanel, "Edit general sections.");

        linearRectangleSectionTablePanel = new LinearRectangleSectionTablePanel(inputDevice);
        featurePane.addTab("Linear rectangle", null,
                linearRectangleSectionTablePanel, "Edit linear rectangle sections.");
        
        linearIProfileSectionTablePanel = new LinearIProfileSectionTablePanel(inputDevice);
        featurePane.addTab("Linear I-profile", null,
                linearIProfileSectionTablePanel, "Edit linear I-profile sections.");
        
        linearSandwichSectionTablePanel = new LinearSandwichSectionTablePanel(inputDevice);
        featurePane.addTab("Linear sandwich", null,
                linearSandwichSectionTablePanel, "Edit linear sandwich sections.");

        sectionPanel.add(featurePane, BorderLayout.CENTER);

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        JButton button = _button("Add.png", mouseListener, null,
                "Add a new section to the structure.");
        button.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
               Component comp = featurePane.getSelectedComponent();
               if (comp == linearGeneralSectionTablePanel)
                   inputDevice.fireInterruptingInput("addgensection");
               else if (comp == linearRectangleSectionTablePanel)
                   inputDevice.fireInterruptingInput("addrecsection");
               else if (comp == linearIProfileSectionTablePanel)
                   inputDevice.fireInterruptingInput("addiprofsection");
               else if (comp == linearSandwichSectionTablePanel)
                   inputDevice.fireInterruptingInput("addsandsection");
           } 
        });
        buttonPanel.add(button);

        /*final JPopupMenu popupAdd = new JPopupMenu();
        JMenuItem item = new JMenuItem("General");
        item.addActionListener(new _CmdListener("addgensection"));
        popupAdd.add(item);

        item = new JMenuItem("Linear rectangle");
        item.addActionListener(new _CmdListener("addrecsection"));
        popupAdd.add(item);

        item = new JMenuItem("Linear I-profile");
        item.addActionListener(new _CmdListener("addiprofsection"));
        popupAdd.add(item);

        item = new JMenuItem("Linear sandwich");
        item.addActionListener(new _CmdListener("addsandsection"));
        popupAdd.add(item);
        
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (!popupAdd.isVisible()) {
                    JButton button = (JButton) e.getSource();
                    popupAdd.show(button, 0, button.getHeight());
                }
                else
                    popupAdd.setVisible(false);
            }
        });*/

        button = _button("Remove.png", mouseListener, null,
                "Remove the selected row.");
        buttonPanel.add(button);
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("removesection");
                }
                else if (inputDevice.isEnabled()) {
                    java.awt.Component comp = featurePane
                            .getSelectedComponent();
                    AbstractSection sec = null;
                    if (comp == linearGeneralSectionTablePanel) {
                        sec = linearGeneralSectionTablePanel
                                .getSelectedSection();
                    }
                    else if (comp == linearRectangleSectionTablePanel) {
                        sec = linearRectangleSectionTablePanel.getSelectedSection();
                    }
                    else if (comp == linearIProfileSectionTablePanel) {
                        sec = linearIProfileSectionTablePanel.getSelectedSection();
                    }
                    else if (comp == linearSandwichSectionTablePanel) {
                        sec = linearSandwichSectionTablePanel.getSelectedSection();
                    }
                    if (sec != null) {
                        inputDevice.fireInterruptingInput("removesection "
                                + sec.getName());
                    }
                }
            }
        });
        buttonPanel.add(button);
        
        button = _button("Assign2.png", mouseListener, null,
                "Assign selected section.");
        buttonPanel.add(button);
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (inputDevice == null) {
                    System.out.println("assignsection");
                }
                else if (inputDevice.isEnabled()) {
                    java.awt.Component comp = featurePane
                            .getSelectedComponent();
                    AbstractSection sec = null;
                    if (comp == linearGeneralSectionTablePanel) {
                        sec = linearGeneralSectionTablePanel
                                .getSelectedSection();
                    }
                    else if (comp == linearRectangleSectionTablePanel) {
                        sec = linearRectangleSectionTablePanel
                                .getSelectedSection();
                    }
                    else if (comp == linearIProfileSectionTablePanel) {
                        sec = linearIProfileSectionTablePanel
                                .getSelectedSection();
                    }
                    else if (comp == linearSandwichSectionTablePanel) {
                        sec = linearSandwichSectionTablePanel
                                .getSelectedSection();
                    }
                    if (sec != null) {
                        inputDevice.fireInterruptingInput("assignsection "
                                + sec.getName());
                    }
                    else {
                        inputDevice.fireInterruptingInput("assignsection");
                    }
                }
            }
        });
        buttonPanel.add(button);


        sectionPanel.add(buttonPanel, BorderLayout.NORTH);
        
    }
    /*private void initSettingsPanel() {

    }*/

    private void initResultsPanel() {
        resultsPanel.setLayout(new BorderLayout());

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        
        JLabel label = new JLabel("Result steps:");
        buttonPanel.add(label);
       
        resultTextField.setEditable(true);
        resultTextField.setHorizontalAlignment(JTextField.RIGHT);
        resultTextField.setText("" + RuBeam.getStructure().getResultSteps());
        resultTextField.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                    resultsPanel.requestFocus();
                    int resultSteps = 0;
                    try {
                        resultSteps = Integer.parseInt(resultTextField.getText());
                    }
                    catch (NumberFormatException ex) {
                        resultSteps = 0;
                    }
                    inputDevice.fireInterruptingInput("setresultsteps " + resultSteps);
                }
            } 
        });
        buttonPanel.add(resultTextField);
        
        JButton button = _button("Refresh.png", mouseListener, null, "Refresh results.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                resultsPanel.requestFocus();
                int resultSteps = 0;
                try {
                    resultSteps = Integer.parseInt(resultTextField.getText());
                }
                catch (NumberFormatException ex) {
                    resultSteps = 0;
                }
                inputDevice.fireInput("setresultsteps " + resultSteps);
            }
        });
        buttonPanel.add(button);
       
        button = _button(/*"Save.png"*/"Export HTML", mouseListener, null, "Export results as HTML file.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
              writeHtmlFile(resultsOutputPanel.getText(), "results.html");  
            } 
         });
        buttonPanel.add(button);
        
        resultsPanel.add(buttonPanel, BorderLayout.NORTH);

        // Center
        resultsOutputPanel = new HtmlPanel();
        resultsPanel.add(resultsOutputPanel, BorderLayout.CENTER);
    }

    private void initStructurePanel() {
        structurePanel.setLayout(new BorderLayout());

        // North
        JPanel buttonPanel = new JPanel(new FlowLayout(FlowLayout.LEFT));
        
        JLabel label = new JLabel("Visualize steps:");
        buttonPanel.add(label);
       
        visualizeTextField.setEditable(true);
        visualizeTextField.setHorizontalAlignment(JTextField.RIGHT);
        visualizeTextField.setText("" + RuBeam.getStructure().getVisualizeSteps());
        visualizeTextField.addKeyListener(new KeyAdapter() {
            public void keyPressed(KeyEvent e) {
                if (e.getKeyCode() == KeyEvent.VK_ENTER) {
                    structurePanel.requestFocus();
                    int steps = 0;
                    try {
                        steps = Integer.parseInt(visualizeTextField.getText());
                    }
                    catch (NumberFormatException ex) {
                        steps = 0;
                    }
                    inputDevice.fireInterruptingInput("setvisualizesteps " + steps);
                }
            } 
        });
        buttonPanel.add(visualizeTextField);
        
        /*JButton button = _button("Refresh.png", mouseListener, null, "Refresh results.");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                structurePanel.requestFocus();
                int steps = 0;
                try {
                    steps = Integer.parseInt(visualizeTextField.getText());
                }
                catch (NumberFormatException ex) {
                    steps = 0;
                }
                inputDevice.fireInput("setvisualizesteps " + steps);
            }
        });
        buttonPanel.add(button);*/
        
        JButton button = _button("Structure", mouseListener, null,
                "Import and export structure.");
        buttonPanel.add(button);

        final JPopupMenu popupAdd = new JPopupMenu();
        JMenuItem item = new JMenuItem("Import macro");
        item.addActionListener(new _CmdListener("importmacro"));
        popupAdd.add(item);
        item = new JMenuItem("Export macro");
        item.addActionListener(new _CmdListener("exportmacro"));
        popupAdd.add(item);
        item = new JMenuItem("Export HTML");
        item.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                writeHtmlFile(structureOutputPanel.getText(), "structure.html");
            }
        });
        popupAdd.add(item);
        popupAdd.addSeparator();
        item = new JMenuItem("Clear");
        item.addActionListener(new _CmdListener("clearstructure"));
        popupAdd.add(item);

        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (!popupAdd.isVisible()) {
                    JButton button = (JButton) e.getSource();
                    popupAdd.show(button, 0, button.getHeight());
                }
                else
                    popupAdd.setVisible(false);
            }
        });
        
        /*button = _button("Save.png""Save as HTML", mouseListener, null, "Save structure as HTML file.");
        button.addActionListener(new ActionListener() {
           public void actionPerformed(ActionEvent e) {
             writeHtmlFile(structureOutputPanel.getText());  
           } 
        });
        buttonPanel.add(button);
        
        button = _button("Import", mouseListener, new _CmdListener(
        "importmacro"), "Import structure from macro file.");
        buttonPanel.add(button);
        
        button = _button("Export", mouseListener, new _CmdListener(
        "exportmacro"), "Export structure to macro file.");
        buttonPanel.add(button);*/
        
        
        structurePanel.add(buttonPanel, BorderLayout.NORTH);

        // Center
        structureOutputPanel = new HtmlPanel();
        structurePanel.add(structureOutputPanel, BorderLayout.CENTER);
    }

    public void updateDialog() {
        if (isVisible()) {
            nodeTablePanel.updatePanel();
            elementTablePanel.updatePanel();
            linearGeneralSectionTablePanel.updatePanel();
            linearRectangleSectionTablePanel.updatePanel();
            linearIProfileSectionTablePanel.updatePanel();
            linearSandwichSectionTablePanel.updatePanel();
            nodalLoadAngTablePanel.updatePanel();
            uniDistribLoadNorTablePanel.updatePanel();
            conLoadAngTablePanel.updatePanel();
            
            Structure struct = RuBeam.getStructure();
            structureOutputPanel.setText(struct.toHTML());
            visualizeTextField.setText("" + struct.getVisualizeSteps());
            String results = null;
            if (!struct.isValid()) {
                results = struct.resultsToHTML();
            }
            else {
                StringBuffer resultsHTML = new StringBuffer(Util.getHtmlHead("ruBeam Results"));
                resultsHTML.append("<body> <h1>Results</h1>");
                resultsHTML.append("<body><b>Results available, they have to be updated.</b></body>");   
                results = resultsHTML.toString();
            }
            resultsOutputPanel.setText(results);
            resultTextField.setText("" + struct.getResultSteps());
        }
    }
    public void updateResultPanel() {
        if (isVisible()) {
            Cursor cursor = getOwner().getCursor();
            setCursor(Cursor.getPredefinedCursor(Cursor.WAIT_CURSOR));
            resultsOutputPanel.setText(RuBeam.getStructure().resultsToHTML());
            setCursor(cursor);
        }
    }

    public void setVisible(boolean b) {
        super.setVisible(b);
        updateDialog();
    }

    // Input Device
    private class _CmdListener implements ActionListener {
        private String m_cmd = null;
        private int m_inputType = INTERRUPTING_COMMAND;

        public _CmdListener(String cmd) {
            m_cmd = cmd;
        }

        public _CmdListener(String cmd, int inputType) {
            m_cmd = cmd;
            m_inputType = inputType;
        }

        public void actionPerformed(ActionEvent e) {
            if (inputDevice == null) {
                System.out.println(m_cmd);
            }
            else if (inputDevice.isEnabled()) {
                switch (m_inputType) {
                case COMMAND:
                    inputDevice.fireInput(m_cmd + ";setagain " + m_cmd);
                    break;
                case INTERRUPTING_COMMAND:
                    inputDevice.fireInterruptingInput(m_cmd + ";setagain "
                            + m_cmd);
                    break;
                case INPUT:
                default:
                    inputDevice.fireInput(m_cmd);
                }
            }
        }
    }

    private static URL getImageURL(String path) {
        return RuBeamInspector.class.getClassLoader().getResource(path);
    }

    private static JButton _button(String name, MouseListener ml,
            _CmdListener cl, String toolTip) {
        JButton button;
        if (MAC_OS) {
            Toolkit toolkit = Toolkit.getDefaultToolkit();
            String nsImage = "";
            if (name.startsWith("Add"))
                nsImage = "NSImage://NSAddTemplate";
            else if (name.startsWith("Remove"))
                nsImage = "NSImage://NSRemoveTemplate";
            else if (name.startsWith("Assign"))
                nsImage = "NSImage://NSExitFullScreenTemplate";
            else if (name.startsWith("Refresh"))
                nsImage = "NSImage://NSRefreshTemplate";
            
            if (! nsImage.isEmpty()) {
                Image image = toolkit.getImage(nsImage);
                button = new JButton(new ImageIcon(image));
                button.putClientProperty("JButton.buttonType", "textured");
                button.setHorizontalTextPosition(SwingConstants.LEFT);
            }
            else {
                button = new JButton();
                int idx = name.indexOf(".");
                if (idx != -1) {
                    name = name.substring(0, idx);
                };
                button.setText(name);
                button.putClientProperty( "JButton.buttonType", "textured" );
            }
        }
        else {
            URL url = getImageURL("rubeam/img/" + name);
            if (url != null) {
                button = new JButton(new ImageIcon(url));
                button.setPreferredSize(m_buttonDim);
                button.setBorderPainted(false);
                button.addMouseListener(ml);
            }
            else {
                button = new JButton();
                int idx = name.indexOf(".");
                if (idx != -1) {
                    name = name.substring(0, idx);
                };
                button.setText(name);
                button.setBorderPainted(true);
            }
        }
        button.setToolTipText(toolTip);
        button.setFocusable(false);
        if (cl != null)
            button.addActionListener(cl);
        return button;
    }
    
    private void writeHtmlFile(String html, String defaultFileName) {
        JFileChooser chooser = new JFileChooser();
        FileNameExtensionFilter filter = new FileNameExtensionFilter(
            "HTML-files", "html");
        chooser.setFileFilter(filter);
        chooser.setSelectedFile(new File(defaultFileName));
        chooser.addActionListener(new SaveFileListener(chooser, html));
        chooser.showSaveDialog(this);
   
    }
    
    class SaveFileListener implements ActionListener {
        private JFileChooser chooser;
        private String html;
        
        public SaveFileListener(JFileChooser chooser, String html) {
            this.chooser = chooser;
            this.html = html;
        }

        private int showSaveDisplayQuestion(File file) {
            String message = "File " + file.getAbsolutePath()
                    + "already exists." + " Replace?";
            return JOptionPane.showOptionDialog(RuBeamInspector.this, message,
                    "Replace existing file?", JOptionPane.YES_NO_OPTION,
                    JOptionPane.WARNING_MESSAGE, null, null, null);
        }

        public void actionPerformed(ActionEvent aev) {
            if (chooser == null)
                return;

            File file = chooser.getSelectedFile();
            if (file != null) {

                if (file.exists()) {
                    int answer = showSaveDisplayQuestion(file);
                    if (answer == JOptionPane.NO_OPTION) {
                        // User doesn't want to overwrite the file
                        return;
                    }
                }
                else {
                    if (! file.getName().endsWith(".html")) {
                        File newFile = new File(file.getAbsolutePath() + ".html"); 
                        file.renameTo(newFile);
                        file = newFile;
                    }
                    try {
                        file.createNewFile();
                    } catch (IOException e) {
                        System.err.println("Can't create file " + file.getAbsolutePath());
                        return;
                    }
                }
                if (!file.canWrite()) {
                    System.err.println("Can't write to " + file.getAbsolutePath());
                    return;
                }
                try {
                    PrintWriter pw = new PrintWriter(file);
                    pw.write(html);
                    pw.flush();
                    pw.close();
                } catch (IOException e) {
                    System.err.println(e);
                    return;
                }
            }
        }
    }

}
