/*
 * Copyright (C) 2009-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.image.BufferedImage;
import java.awt.geom.AffineTransform;
import java.awt.geom.Point2D;
import java.io.File;
import java.io.RandomAccessFile;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.imageio.ImageIO;

/** 
 * A placeable and serializable attributed image. The image is fitted in a
 * rectangular area defined by the width and height in model units and the lower
 * left point as the placement point. The rectangle can be further transformed
 * by an AffineTransform object.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.4 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class AttributedImage implements Cloneable, Serializable {
    
    // Serialization data
    private static final long serialVersionUID = 0L;
    private static Object s_stream = null;
    private static List<File> s_sharedFiles = null;
    private static List<BufferedImage> s_sharedImages = null;
    
    // Attributes
    private transient BufferedImage m_bi = null;
    private transient File m_originalFile = null;   // File set by the user
    private transient File m_physicalFile = null;   // Really existing file
    
    private double m_width = 1;
    private double m_height = 1;
    private AffineTransform m_trf = new AffineTransform();

    public AttributedImage() {
    }
    
    public AttributedImage(double width, double height, AffineTransform trf) {
        setImageSize(width, height);
        setTransform((AffineTransform)trf.clone());
        m_trf = trf;
    }

    // Object protocol
    public Object clone() {
        AttributedImage clonedObject;
        try {
            clonedObject = (AttributedImage)super.clone();
        }
        catch (CloneNotSupportedException e) {
            // This shouldn't happen, since we are Cloneable
            throw new InternalError();
        }

        // Extend to deep clone
        m_trf = (AffineTransform)m_trf.clone();
        
        return clonedObject;
    }

    private void writeObject(ObjectOutputStream out) throws IOException {
        // Non-transient attributes
        out.defaultWriteObject();

        // Save original file object (does not include the file contents)
        out.writeObject(m_originalFile);

        // Save file contents because it could be deleted when loading next time
        if (s_stream != out) {
            s_stream = out;
            s_sharedFiles = new ArrayList<File>();
        }
        int bufId = s_sharedFiles.indexOf(m_physicalFile);
        if (bufId != -1) {
            // Write the ID and an empty buffer
            out.writeInt(bufId);
            out.writeLong(0);
            return;
        }

        // Add File object to list and use position as ID
        s_sharedFiles.add(m_physicalFile);
        bufId = s_sharedFiles.size() - 1;
        out.writeInt(bufId);
        
        // Save file contents
        RandomAccessFile raf = new RandomAccessFile(m_physicalFile, "r");
        long bufLen = raf.length();
        out.writeLong(bufLen);

        final int LEN = 1024;
        byte[] buf = new byte[LEN]; 
        for (long i = 0; i < bufLen; i += LEN) {
            int len = i + LEN < bufLen ? LEN : (int)(bufLen - i);
            raf.read(buf, 0, len);
            out.write(buf, 0, len);
        }
        raf.close();
    }

    private void readObject(ObjectInputStream in) throws IOException,
    ClassNotFoundException {
        // Non-transient attributes
        in.defaultReadObject();

        // Read original file object (could be deleted or changed)
        m_originalFile = (File)in.readObject();
        
        // Init buffer administration
        if (s_stream != in) {
            s_stream = in;
            s_sharedFiles = new ArrayList<File>();
            s_sharedImages = new ArrayList<BufferedImage>();
        }

        // Read buffer ID and length
        int bufId = in.readInt();
        long bufLen = in.readLong();

        // Already deserialized?
        if (bufId < s_sharedFiles.size()) {
            m_physicalFile = s_sharedFiles.get(bufId);
            m_bi = s_sharedImages.get(bufId);
            return;
        }

        // Does the original file exist?
        if (m_originalFile.exists()) {
            m_physicalFile = m_originalFile;
            in.skip(bufLen);
        }
        else {
            // Create temporary file from buffer
            m_physicalFile = File.createTempFile("cademia", 
                    "_" + m_originalFile.getName());
            RandomAccessFile raf = new RandomAccessFile(m_physicalFile, "rw");
            for (long i = 0; i < bufLen; i++) {
                raf.write(in.read());
            }
            if (! true) { // Buffered version generates IllegalStateException?!
                final int LEN = 1024;
                byte[] buf = new byte[LEN]; 
                for (long i = 0; i < bufLen; i += LEN) {
                    int len = i + LEN < bufLen ? LEN : (int)(bufLen - i);
                    in.read(buf, 0, len);
                    raf.write(buf, 0, len);
                }
            }
            raf.close();
        }
        
        // Restore image from file
        m_bi = ImageIO.read(m_physicalFile);

        // Update administration
        s_sharedFiles.add(m_physicalFile);
        s_sharedImages.add(m_bi);
    }

    // Component protocol
    public AffineTransform getTransform() {
        return (AffineTransform)m_trf.clone();
    }
    public void setTransform(AffineTransform trf) {
//        m_trf = trf;
        m_trf = (AffineTransform)trf.clone();
    }

    /**
     * Gets the image object.
     * 
     * @return the image object
     */
    public BufferedImage getImage() {
        return m_bi;
    }
    
    /**
     * Sets the image file.
     * 
     * @param file the image file
     * @throws IOException if the file does not exist or is not readable.
     */
    public void setImageFile(File file) throws IOException {
        m_bi = ImageIO.read(file);
        m_originalFile = file;
        m_physicalFile = file;
    }
    
    /**
     * Gets the original image file. Note that the file could not exist anymore.
     * 
     * @return the original image file
     */
    public File getOriginalImageFile() {
        return m_originalFile;
    }
    
    /**
     * Gets the file with the image representation. If the original image
     * file still exists then this method behaves exactly behaves as
     * getImageFile(). Otherwise a temporary existing file is returned.
     * 
     * @return the physical image file
     */
    public File getPhysicalImageFile() {
        return m_physicalFile;
    }
    
    public void transformBy(AffineTransform at) {
        m_trf.preConcatenate(at);
    }
    
    /**
     * Sets the image size for subsequent render operations. Please note that
     * this operation does not change the dimension of the stored image.
     * 
     * Negative width or height values result in mirrored images.
     * 
     * @param width the width (if negative then the image is mirrored about the
     *     y axis)
     * @param height the height (if negative then the image is mirrored about
     *     the x axis)
     */
    public void setImageSize(double width, double height) {
        m_width = width;
        m_height = height;
    }
    
    /**
     * Returns the width of the rendered image.
     * 
     * @return the width (if negative then the image is mirrored about the
     *     y axis)
     */
    public double getWidth() {
        return m_width;
    }
    
    /**
     * Returns the height of the rendered image.
     * 
     * @return the width (if negative then the image is mirrored about the
     *     x axis)
     */
    public double getHeight() {
        return m_height;
    }
    
    /**
     * Calculates the four corner points of the image.
     * 
     * @param p1 the first point
     * @param p2 the second point
     * @param p3 the third point
     * @param p4 the fourth point
     */
    public void cornerPoints(Point2D p1, Point2D p2, Point2D p3, Point2D p4) {
        p1.setLocation(0, 0);
        p2.setLocation(m_width, 0);
        p3.setLocation(m_width, m_height);
        p4.setLocation(0, m_height);
        m_trf.transform(p1, p1);
        m_trf.transform(p2, p2);
        m_trf.transform(p3, p3);
        m_trf.transform(p4, p4);
    }
}