/* 
 * Copyright (C) 2009-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util;

import java.awt.event.MouseEvent;

import javax.swing.event.MouseInputAdapter;
import javax.swing.event.MouseInputListener;

/** 
 * Delivers extended mouse events like 'mouse holding' and 'double clicked'.
 * Obviously such events can only be delivered after a certain wait time. If the
 * user is not interested in such extended events the standard MouseListener and
 * MouseAdapter classes should be used.
 * 
 * For an optimal runtime behavior it is important to specify the event filter
 * properly. 
 * 
 * Although ExtendedMouseAdapter objects deliver the standard mouse events the
 * class does not implement the MouseListener interface. Like with the standard
 * MouseListener the ExtendedMouseAdapter must be extended to receive
 * notification.
 * 
 * This decision was made intentionally. Like standard MouseListeners
 * ExtendedMouseListeners must be added to a data source via
 * Component.addMouseListener(). Because ExtendedMouseListener is not of type
 * MouseListener objects of this class cannot be passed directly. Instead the
 * method ExtendedMouseListener.asMouseListener() must be used. 
 *   
 * @author Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class ExtendedMouseAdapter {

    private MouseInputListener m_mouseInputListener = new MouseInputAdapter() {

        /**
         * After a wait time this thread notifies about a 'mouse holding' event.
         * It the thread is set quiet then nothing happens.
         * 
         */
        class CheckHoldingThread extends Thread {
            private static final long WAIT = 300;
            private long m_startPressed = -1;
            private MouseEvent m_mouseEvent = null;

            public CheckHoldingThread(MouseEvent e) {
                m_startPressed = System.currentTimeMillis();
                m_mouseEvent = e;
            }
            public void setQuiet() {
                m_startPressed = -1;
            }
            public void run() {
                while (true) {
                    try {
                        Thread.sleep(WAIT / 5);
                    }
                    catch (InterruptedException x) {
                    }
                    if (m_startPressed < 0) {
                        break;
                    }
                    else if (System.currentTimeMillis() - m_startPressed > 
                    WAIT) {
                        ExtendedMouseAdapter.this.mouseHolding(m_mouseEvent);
                        m_clickCount = 0;   // Change click count
                        break;
                    }
                }
            }
        }

        /**
         * After a wait time this thread notifies about a 'mouse clicked' event.
         * It the thread is set quiet then nothing happens.
         * 
         */
        class CheckDoubleClickThread extends Thread {
            private static final long WAIT = 200;
            private long m_startClicked = -1;
            private MouseEvent m_mouseEvent = null;

            public CheckDoubleClickThread(MouseEvent e) {
                m_startClicked = System.currentTimeMillis();
                m_mouseEvent = e;
            }
            public void setQuiet() {
                m_startClicked = -1;
            }
            public void setMouseEvent(MouseEvent e) {
                m_mouseEvent = e;
            }
            public void run() {
                while (true) {
                    try {
                        Thread.sleep(WAIT / 5);
                    }
                    catch (InterruptedException x) {
                    }
                    if (m_startClicked < 0) {
                        break;
                    }
                    else if (System.currentTimeMillis() - m_startClicked >
                    WAIT) {
                        ExtendedMouseAdapter.this.mouseClicked(m_mouseEvent);
                        break;
                    }
                }
            }
        }
        
        private CheckHoldingThread m_holdThread = null;
        private CheckDoubleClickThread m_doubleClickThread = null;
        private int m_clickCount = -1;  // Click count unchanged
        
        public void mousePressed(final MouseEvent e) {
            // Fire the ordinary event
            ExtendedMouseAdapter.this.mousePressed(e);

            // Check for a hold event
            if ((ExtendedMouseAdapter.this.m_eventFilter & 
                    ExtendedMouseAdapter.HOLDING_EVENT) != 0) {
                m_holdThread = new CheckHoldingThread(e);
                m_holdThread.start();
            }
        }
        public void mouseReleased(MouseEvent e) {
            // Disable holding event
            if (m_holdThread != null) {
                m_holdThread.setQuiet();
                m_holdThread = null;
            }
            
            // Fire the ordinary event
            ExtendedMouseAdapter.this.mouseReleased(e);
        }
        
        public void mouseClicked(MouseEvent e) {
            // Click count changed?
            if (m_clickCount >= 0)
                e = new MouseEvent(e.getComponent(), e.getID(), e.getWhen(),
                        e.getModifiers(), e.getX(), e.getY(), e.getXOnScreen(),
                        e.getYOnScreen(), m_clickCount, e.isPopupTrigger(),
                        e.getButton()); 
            m_clickCount = -1;  // Reset click count
            
            if ((ExtendedMouseAdapter.this.m_eventFilter & 
                    ExtendedMouseAdapter.MULTI_CLICKED_EVENT) != 0) {
                if (e.getClickCount() == 0) {   // Mouse held
                    if (m_doubleClickThread != null) {
                        m_doubleClickThread.setQuiet();
                        m_doubleClickThread = null;
                    }
                    ExtendedMouseAdapter.this.mouseClicked(e);
                }
                else if (e.getClickCount() == 1) {
                    if (m_doubleClickThread != null)
                        m_doubleClickThread.setQuiet();
                    m_doubleClickThread = new CheckDoubleClickThread(e);
                    m_doubleClickThread.start();
                }
                else {
                    if (m_doubleClickThread != null)
                        m_doubleClickThread.setMouseEvent(e);
                }
            }
            else if ((ExtendedMouseAdapter.this.m_eventFilter & 
                    ExtendedMouseAdapter.HOLDING_EVENT) != 0) {
                ExtendedMouseAdapter.this.mouseClicked(e);
            }
            else {
                ExtendedMouseAdapter.this.mouseClicked(e);
            }
        }
        
        public void mouseEntered(MouseEvent e) {
            ExtendedMouseAdapter.this.mouseEntered(e);
        }

        public void mouseExited(MouseEvent e) {
            ExtendedMouseAdapter.this.mouseExited(e);
        }
        
        public void mouseDragged(MouseEvent e) {
            // Disable holding event
            if (m_holdThread != null) {
                m_holdThread.setQuiet();
                m_holdThread = null;
            }

            // Disable double click event
            if (m_doubleClickThread != null) {
                m_doubleClickThread.setQuiet();
                m_doubleClickThread = null;
            }

            // Fire the ordinary event
            ExtendedMouseAdapter.this.mouseDragged(e);
        }
        public void mouseMoved(MouseEvent e) {
            // Disable holding event
            if (m_holdThread != null) {
                m_holdThread.setQuiet();
                m_holdThread = null;
            }

            // Disable double click event
            if (m_doubleClickThread != null) {
                m_doubleClickThread.setQuiet();
                m_doubleClickThread = null;
            }

            // Fire the ordinary event
            ExtendedMouseAdapter.this.mouseMoved(e);
        }
    };

    // Events and filter
    public static final int HOLDING_EVENT = 1;
    public static final int MULTI_CLICKED_EVENT = 2;
    private int m_eventFilter = 0;
    
    ExtendedMouseAdapter() {
        // Disabled
    }
    public ExtendedMouseAdapter(int filter) {
        m_eventFilter = filter;
    }
    
    /**
     * This method allows to add an ExtendedMouseAdapter as a MouseListener to a
     * component. 
     * 
     * @return the reference to a MouseListener implementation
     */
    public MouseInputListener asMouseInputListener() {
        return m_mouseInputListener;
    }

    /**
     * Mouse event 'Clicked' occurred. For a 'long click' the
     * MouseEvent.getClickCount() method returns a value of 0.
     * 
     * If the filter contains the MOUSE_MULIPLE_CLICKED_EVENT then only ONE
     * SINGLE event will be delivered (as opposed to the standard where many
     * clicked events are generated).
     * 
     * @param e
     */
    public void mouseClicked(MouseEvent e) {
    }
    /**
     * Mouse event 'Entered' occurred.
     * 
     * @param e
     */
    public void mouseEntered(MouseEvent e) {
    }
    /**
     * Mouse event 'Exited' occurred.
     * 
     * @param e
     */
    public void mouseExited(MouseEvent e) {
    }
    /**
     * Mouse event 'Pressed' occurred.
     * 
     * @param e
     */
    public void mousePressed(MouseEvent e) {
    }
    /**
     * Mouse event 'Released' occurred.
     *
     * @param e
     */
    public void mouseReleased(MouseEvent e) {
    }
    /**
     * Extended event 'Holding' occurred. This means, that the mouse is 
     * currently held down. Event will only be delivered if the filter contains
     * the HOLDING_EVENT.
     * 
     * Subsequent 'pressed' and 'clicked' events will be delivered as well.
     * 
     * @param e
     */
    public void mouseHolding(MouseEvent e) {
    }
    
    /**
     * Mouse event 'Dragged' occurred.
     * 
     * @param e
     */
    public void mouseDragged(MouseEvent e) {
    }

    /**
     * Mouse event 'Moved' occurred.
     * 
     * @param e
     */
    public void mouseMoved(MouseEvent e) {
    }
}