/* 
 * Copyright (C) 2008-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.binding;

import java.util.Set;

/** 
 * Represents a Binding. A bound object is assigned exactly one Binding. The
 * Binding contains a set of binding objects that the bound object is dependent
 * on.
 * <p>
 * Bindings are stored in a {@link BindingModel} that ensures the consistency of
 * the Binding graph. This BindingModel (and only the BindingModel) shall use
 * the protected methods of a Binding object. The BindingModel contains the
 * methods to set the bound object and the binding objects of a Binding.
 * <p>
 * When one of the binding objects changes then the bound object must be 
 * updated. This task is delegated to the <code>Updater</code> object of the
 * Binding. A binding has exactly one Updater object.
 * 
 * @param <E> the type of the bound object and the binding objects
 * 
 * @author Albertus Olivier, Berthold Firmenich
 * @version $Revision: 1.4 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public abstract class Binding<E> implements Cloneable {
    /**
     * Makes <code>clone()</code> public.
     * 
     * @return the cloned object
     */
    public Object clone() {
        Object clone = null;
        try {
            clone = super.clone();
        }
        catch (CloneNotSupportedException e) {
            throw new InternalError();
        }
        return clone;
    }
    
    /**
     * Returns the bound object.
     * 
     * @return the bound object
     */
    public abstract E getBoundObject();
    
    /**
     * Sets the bound object. This method is protected and shall only be used by
     * {@link BindingModel} instances.
     * 
     * @param obj the bound object
     */
    protected abstract void _setBoundObject(E obj);
	
    /**
     * Gets the binding objects.
     * 
	 * @return the binding objects
	 */
	public abstract Set<E> getBindingObjects();
    
	/**
	 * Removes a binding object.
	 * 
     * @param obj the binding object to be removed
     */
    public abstract void removeBindingObject(E obj);
    
    /**
     * Adds a binding object. This method is protected and shall only be used by
     * {@link BindingModel} instances.
     * 
     * @param obj
     */
    protected abstract void _addBindingObject(E obj);
    
    /**
     * This method is called when at least one binding object was changed. The
     * context is passed as an argument. The method should not be called by
     * applications.
     * 
	 * @param context the Binding context
	 */
	public abstract void update(BindingContext<E> context);
}