/* 
 * Copyright (C) 2008-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.binding;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

/** 
 * Contains algorithms to sort and update the {@link BindingModel}.
 * 
 * @param <E> the type of the bound object and the binding objects
 * 
 * @author Albertus Olivier, Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class BindingUtils<E> {

    static class IEdge<E> {
        private E start;
        private E end;
        private IEdge(E start, E end) {
            this.start = start;
            this.end = end;
        }
        public String toString() {
            return "(" + start + " -> " + end + ")";
        }
    }
    
    private static <E> Set<IEdge<E>> _edges(E boundObject, Binding<E> b) {
        Set<IEdge<E>> edgeSet = new HashSet<IEdge<E>>();
        Set<E> bindingObjects = b.getBindingObjects();
        Iterator<E> it = bindingObjects.iterator();
        while (it.hasNext()) {
            E bindingObject = it.next();
            edgeSet.add(new IEdge<E>(bindingObject, boundObject));
        }
        return edgeSet;
    }
    
    /**
     * Updates changed objects in the correct order. The changed objects must be
     * contained in the change set. The Binding graph must be contained in the
     * BindingModel. Applications must pass a BindingContext instance.
     * 
     * @param <E> the type of bound and binding objects
     * @param bindingModel the binding model
     * @param changedObjects the changed objects
     * @param context an application specific context
     */
    public static <E> void updateBindings(BindingModel<E> bindingModel,
            Set<E> changedObjects, BindingContext<E> context) {

        // Anything to do?
        if (changedObjects.isEmpty())
            return;
        
 		Set<IEdge<E>> edgeSet = getEdges(bindingModel);
 		Set<IEdge<E>> changeSet = getChangedSet(edgeSet, changedObjects);
        Set<IEdge<E>> changeDomain = getChangeDomain(edgeSet, changeSet);
        List<List<E>> updateOrder = topologicalSort(changeDomain);

        int nUpdates = 0;
        if (context == null)
            context = new BindingContext<E>();
        for (List<E> list : updateOrder) {
            for (E boundObject : list) {
                if (bindingModel.containsKey(boundObject)) {
                    Binding<E> binding = bindingModel.get(boundObject);
                    binding.update(context);
                    nUpdates++;
                }
            }
        }
//        System.out.println("Updated " + nUpdates + " bindings");
	}

    /*
     * Outline of the algorithm for the three relevant cases.
     * 
     *  changed  change  change   update
     *  objects   set    domain    order
     * 
     *  +-----+                   +-----+
     *  | cng |                   | no  |
     *  +--+--+    |        |     +--+--+
     *     B       |        |        B to update
     *  +--v--+    v        v     +--v--+
     *  |     |                   | yes |
     *  +--+--+             |     +--+--+
     *     B                |        B to update
     *  +--v--+             v     +--v--+
     *  |     |                   | yes |
     *  +-----+                   +-----+
     * 
     ************************************************************* 
     * 
     *  +-----+                   +-----+
     *  |     |                   | no  |
     *  +--+--+    |        |     +--+--+
     *     B       |        |        B to update
     *  +--v--+    v        v     +--v--+
     *  | cng |                   | yes |
     *  +--+--+    |        |     +--+--+
     *     B       |        |        B to update
     *  +--v--+    v        v     +--v--+
     *  |     |                   | yes |
     *  +-----+                   +-----+
     * 
     ************************************************************* 
     * 
     *  +-----+
     *  |     |
     *  +--+--+
     *     B     
     *  +--v--+                   +-----+
     *  |     |                   | no  |
     *  +--+--+    |        |     +--+--+
     *     B       |        |        B to update
     *  +--v--+    v        v     +--v--+
     *  | cng |                   | yes |
     *  +-----+                   +-----+
     * 
     */
    
	private static <E> Set<IEdge<E>> getEdges(Map<E,Binding<E>> bindingMap) {
		Set<IEdge<E>> edgeSet = new HashSet<IEdge<E>>();
		Set<E> boundObjects = bindingMap.keySet();
		for (E boundObject : boundObjects) {
            Binding<E> b = bindingMap.get(boundObject);
            edgeSet.addAll(_edges(boundObject, b));
		}
		
//		System.out.println("Edges:");
//		for (IEdge<E> e:edgeSet) {
//            System.out.println(e);
//        }

		return edgeSet;
	}
	
	
	/**
	 * Every bound object that was changed has to be updated. This is important
	 * because the constraints of the Binding may be violated otherwise!
	 * 
	 * This method calculates the set of edges where the start OR end object was
	 * changed!
	 * 
	 * @return a set that contains all the edges where start or end object
	 * was changed
	 */
	private static <E> Set<IEdge<E>> getChangedSet(Set<IEdge<E>> edgeSet, 
	        Set<E> changedObjects) {
	    Set<IEdge<E>> changeSet = new HashSet<IEdge<E>>();
	    for (IEdge<E> edge : edgeSet) {
            if (changedObjects.contains(edge.start) ||
                    changedObjects.contains(edge.end))
                changeSet.add(edge);
	    }
        
//		System.out.println("Edges changed:");
//        for (IEdge<E> e:changeSet) {
//            System.out.println(e);
//        }

        return changeSet;
	}

	/**
	 * Calculates the set of affected edges. The affected edges are dependent
	 * on the set of changed edges. A changed edge is an edge where the start
	 * or end object was changed.
	 * 
	 * @param set contains all the edges of the graph
	 * @param changedSet contains all the changed edges of the graph
	 * @return a set that contains all the connected edges.
	 */
	private static <E> Set<IEdge<E>> getChangeDomain(Set<IEdge<E>> set, 
	        Set<IEdge<E>> changedSet) {
	    Set<IEdge<E>> allSet = new HashSet<IEdge<E>>(set);
	    Set<IEdge<E>> affectedSet = new HashSet<IEdge<E>>(changedSet);
	    if (affectedSet.isEmpty())
	        return affectedSet;

	    Set<IEdge<E>> tmp = new HashSet<IEdge<E>>();
	    List<IEdge<E>> list = new ArrayList<IEdge<E>>(changedSet);

	    while(list.size() > 0) {
	        tmp.clear();
	        IEdge<E> edge = list.remove(0); // use the list as a stack
	        Iterator<IEdge<E>> allSetIter = allSet.iterator();
	        while(allSetIter.hasNext()) {
	            IEdge<E> e = allSetIter.next();
	            if (e.start == edge.end) {
	                tmp.add(e);
	                list.add(e);
	                allSetIter.remove();
	            }
	        }
	        affectedSet.addAll(tmp);
	    }
        
//	    System.out.println("Affected edges:");
//        for (IEdge<E> e:affectedSet) {
//            System.out.println(e);
//        }

        return affectedSet;
	}   
	    
	/**
	 * @param set contains the edges of the graph that must be topologically
	 * sorted.
	 * @return a list with the length equal to the number of update steps, each
	 * entry in this list contains a list of objects that should be updated in
	 * the corresponding update step.
	 */
    private static <E> List<List<E>> topologicalSort(Set<IEdge<E>> set) {
		List<List<E>> list = new ArrayList<List<E>>();
		if (set.isEmpty())
			return list;
		int counter = 0;
		Set<E> start = new HashSet<E>();
		for (IEdge<E> e:set) {
			start.add(e.start);
			start.add(e.end);
		}
		Set<E> completed = new HashSet<E>();
		while (true) {
			Set<E> end = new HashSet<E>();
			Iterator<IEdge<E>> it = set.iterator();
			while (it.hasNext()) {
				IEdge<E> e = it.next();
				if (! completed.contains(e.start))
					end.add(e.end);
				else
					it.remove();
			}
			list.add(new ArrayList<E>());
			Iterator<E> iter = start.iterator();
			boolean cycle = true;
			while (iter.hasNext()) {
				E obj = iter.next();
				if (! end.contains(obj)) {
					cycle = false;
					list.get(counter).add(obj);
					completed.add(obj);
					iter.remove();
				}
			}
			if (cycle)
				throw new RuntimeException("Cycle Detected!!!\n"+list);
			counter++;
			if (end.isEmpty())
				break;
		}

		// The first list of objects must not be updated because the edge set
		// contained edges where the start OR end object were changed!
		list.remove(0);
		
//		System.out.println("Topologically sorted objects:");
//        for (List<E> _list : list) {
//            System.out.println(list.indexOf(_list) + ". Step");
//            for (E e : _list) {
//                System.out.println("  " + e);
//            }
//        }
		
		return list;
	}
}