/* 
 * Copyright (C) 2008-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.binding;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

/** 
 * Represents an implemented BindingModel.
 * 
 * @param <E> the type of the bound object and the binding objects
 * 
 * @author Albertus Olivier, Berthold Firmenich
 * @version $Revision: 1.4 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class DefaultBindingModel<E> implements BindingModel<E> {
    private static final long serialVersionUID = 1L;
    private Map<E,Binding<E>> m_map = new HashMap<E,Binding<E>>();

    /**
     * Constructs the BindingModel instance.
     * 
     */
    public DefaultBindingModel() {
    }

    /**
     * Adds a binding object to a Binding if the binding graph remains acyclic.
     * 
     * @param binding the Binding
     * @param obj the binding object
     * @exception IllegalArgumentException in case of a cycle.
     * @see BindingModel#addBindingObject(Binding, java.lang.Object)
     */
    public void addBindingObject(Binding<E> binding, E obj) {
        // Check for cycles
        E boundObject = binding.getBoundObject();
        Binding<E> _binding = m_map.get(obj);
        if (_binding != null)
            _checkForCycles(_binding, boundObject);
        
        // Add Binding
        binding._addBindingObject(obj);
    }
    
    /* (non-Javadoc)
     * @see java.util.Map#clear()
     */
    public void clear() {
        m_map.clear();
    }
    /* (non-Javadoc)
     * @see java.util.Map#containsKey(java.lang.Object)
     */
    public boolean containsKey(Object key) {
        return m_map.containsKey(key);
    }
    /* (non-Javadoc)
     * @see java.util.Map#containsValue(java.lang.Object)
     */
    public boolean containsValue(Object value) {
        return m_map.containsValue(value);
    }
    /* (non-Javadoc)
     * @see java.util.Map#entrySet()
     */
    public Set<Entry<E,Binding<E>>> entrySet() {
        return m_map.entrySet();
    }
    /* (non-Javadoc)
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object o) {
        return m_map.equals(o);
    }
    /* (non-Javadoc)
     * @see java.util.Map#get(java.lang.Object)
     */
    public Binding<E> get(Object key) {
        return m_map.get(key);
    }
    /* (non-Javadoc)
     * @see java.lang.Object#hashCode()
     */
    public int hashCode() {
        return m_map.hashCode();
    }
    /* (non-Javadoc)
     * @see java.util.Map#isEmpty()
     */
    public boolean isEmpty() {
        return m_map.isEmpty();
    }
    /* (non-Javadoc)
     * @see java.util.Map#keySet()
     */
    public Set<E> keySet() {
        return m_map.keySet();
    }
    /**
     * Adds a Binding to a bound object if the binding graph remains acyclic.
     * If the bound object was assigned another Binding then the Binding will
     * be replaced.
     * 
     * @param key the bound object
     * @param value the Binding
     * @exception IllegalArgumentException in case of a cycle
     * @see java.util.Map#put(java.lang.Object, java.lang.Object)
     */
    public Binding<E> put(E key, Binding<E> value) {
        _checkForCycles(value, key);
        value._setBoundObject(key);
        return m_map.put(key, value);
    }
    private void _checkForCycles(Binding<E> binder, E startOfCycle) {
        Set<E> bindingObjects = binder.getBindingObjects();
        for (E bindingObject : bindingObjects) {
            if (bindingObject == startOfCycle)
                throw new IllegalArgumentException("Cycle detected");
            if (m_map.containsKey(bindingObject)) {
                Binding<E> _binder = m_map.get(bindingObject);
                _checkForCycles(_binder, startOfCycle);
            }
        }
    }
    /**
     * Adds a set of Bindings to bound objects if the binding graph remains
     * acyclic. In case of a cycle this method will restore the initial object
     * state. Bound objects that already had Bindings will receive new Bindings.
     * 
     * @param m the map with (Object, Binding) pairs
     * @exception IllegalArgumentException in case of a cycle
     * @see java.util.Map#putAll(java.util.Map)
     */
    public void putAll(Map<? extends E,? extends Binding<E>> m) {
        // Save existing pairs
        Map<E,Binding<E>> undoMap = new HashMap<E,Binding<E>>();
        Set<? extends E> mKeySet = m.keySet();
        for (E key : mKeySet) {
            if (containsKey(key))
                undoMap.put(key, get(key));
        }
        
        // Put new pairs
        try {
            for (E key : mKeySet)
                put(key, m.get(key));
        }
        catch (IllegalArgumentException e) {
            for (E key : mKeySet) {
                if (undoMap.containsKey(key))
                    put(key, undoMap.get(key));
                else
                    remove(key);
            }
            throw e;
        }
    }
    /* (non-Javadoc)
     * @see java.util.Map#remove(java.lang.Object)
     */
    public Binding<E> remove(Object key) {
        return m_map.remove(key);
    }
    /* (non-Javadoc)
     * @see java.util.Map#size()
     */
    public int size() {
        return m_map.size();
    }
    /* (non-Javadoc)
     * @see java.util.Map#values()
     */
    public Collection<Binding<E>> values() {
        return m_map.values();
    }
}
