/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.cmd;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;

import cib.util.Dbg;

/** 
 * The Command Manager does, undoes und redoes Commands. The command
 * history is recorded.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.9 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class CmdMgr {

    public static final int CAN_UNDO_MORE = 1;
    public static final int CAN_REDO_MORE = 2;

    protected LinkedList<Cmd> m_cmdHistory = new LinkedList<Cmd>();
    protected int m_cmdHistoryIndex = -1;
    protected int m_maximumSize = 20;
    
    public CmdMgr() {
    }

   /** 
    * Sets the maximum number of stored commands
    */
    public void setMaximumSize(int n) {
        m_maximumSize = n <= 0 ? -1 : n;
    }
   /** 
    * Gets the maximum number of stored commands
    * @return maximum number of commands
    */
    public int getMaximumSize() {
        return m_maximumSize;
    }
    
   /** 
    * Clears the command history
    */
    public void clear() {
        m_cmdHistory.clear();
        m_cmdHistoryIndex = -1;
    }

   /** 
    * Checks whether any commands are stored
    * @return true if no commands are stored
    */
    public boolean isEmpty() {
        return m_cmdHistory.isEmpty();
    }

   /** 
    * Does a command that has been previously created.
    * @param cmd the command object
    * @param context an application specific context object
    * @return Flags indicating if more undos or redos are possible
    */
    public int doCmd(Cmd cmd, Object context) {

        // Start the command
        try {
            cmd.doCmd(context);
        }
        catch (CmdAbortedException e) {
            return getFlags();
        }

        if (cmd.changesState()) {
            if (cmd.isUndoable()) {
                // Add command object to linked list
                m_cmdHistory.add(++m_cmdHistoryIndex, cmd);

                // Graph may branch: remove all following commands until end of list
                Object lastCmd = null;
                while (cmd != (lastCmd = m_cmdHistory.getLast()))
                    m_cmdHistory.remove(lastCmd);
                
                // Reduce the size of the chain eventually
                if (m_maximumSize > 0) {
                    while (m_cmdHistory.size() > m_maximumSize) {
                        m_cmdHistory.removeFirst();
                        m_cmdHistoryIndex--;
                    }
                }
            }
            else {
                // Clear command history
                clear();
            }
        }
        if (Dbg.on())
            dumpCmdHistory();

        return getFlags();
    }

   /** 
    * Does a command that has been previously created.
    * @param cmd the command object
    * @return Flags indicating if more undos or redos are possible
    */
    public int doCmd(Cmd cmd) {
        return doCmd(cmd, null);
    }

   /** 
    * Undoes the next command.
    * @param context an application specific context object
    * @return Flags indicating if more undos or redos are possible
    */
    public int undoCmd(Object context) {
        if (m_cmdHistoryIndex < 0)
            return getFlags();
        Cmd cmd = m_cmdHistory.get(m_cmdHistoryIndex);

        m_cmdHistoryIndex--;
        cmd.undoCmd(context);
        if (Dbg.on())
            dumpCmdHistory();
        return getFlags();
    }

   /** 
    * Undoes the next command.
    * @return Flags indicating if more undos or redos are possible
    */
    public int undoCmd() {
        return undoCmd(null);
    }

   /** 
    * Redoes the next command.
    * @param context an application specific context object
    * @return Flags indicating if more undos or redos are possible
    */
    public int redoCmd(Object context) {
        int index = m_cmdHistoryIndex + 1;
        if (index < m_cmdHistory.size()) {
            Cmd cmd = m_cmdHistory.get(index);
            m_cmdHistoryIndex = index;
            cmd.redoCmd(context);
        }
        if (Dbg.on())
            dumpCmdHistory();
        return getFlags();
    }

   /** 
    * Redoes the next command.
    * @return Flags indicating if more undos or redos are possible
    */
    public int redoCmd() {
        return redoCmd(null);
    }

   /** 
    * Check if at least one more undo is possible
    * @return true if at least one more undo is possible
    */
    public boolean canUndoMore() {
        return m_cmdHistoryIndex >= 0 ? true : false;
    }

   /** 
    * Check if at least one more redo is possible
    * @return true if at least one more redo is possible
    */
    public boolean canRedoMore() {
        return m_cmdHistoryIndex + 1 < m_cmdHistory.size() ? true : false;
    }

   /** 
    * Returns an interator over the undoable commands
    * @return Iterator an iterator over the string representations
    */
    public Iterator<Cmd> undoableCommandsIterator() {
        return new UndoableCommandsIterator(m_cmdHistoryIndex);
    }
    private class UndoableCommandsIterator implements Iterator<Cmd> {
        private int m_ind;
        public UndoableCommandsIterator(int ind) {
            m_ind = ind;
        }
        public boolean hasNext() {
            return m_ind >= 0;
        }
        public Cmd next() {
            if (m_ind < 0)
                throw new NoSuchElementException();
            Cmd cmd = m_cmdHistory.get(m_ind--);
            return cmd;
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

   /** 
    * Returns an interator over the redoable commands
    * @return Iterator an iterator over the string representations
    */
    public Iterator<Cmd> redoableCommandsIterator() {
        return new RedoableCommandsIterator(m_cmdHistoryIndex + 1);
    }
    private class RedoableCommandsIterator implements Iterator<Cmd> {
        private int m_ind;
        public RedoableCommandsIterator(int ind) {
            m_ind = ind;
        }
        public boolean hasNext() {
            return m_ind < m_cmdHistory.size();
        }
        public Cmd next() {
            if (m_ind >= m_cmdHistory.size())
                throw new NoSuchElementException();
            Cmd cmd = m_cmdHistory.get(m_ind++);
            return cmd;
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

   /** 
    * Dumps the command history to the standard output stream
    */
    public void dumpCmdHistory() {
        System.out.println("Dump of command history");
        for (int index = 0; index < m_cmdHistory.size(); index++) {
            Object o = m_cmdHistory.get(index);
            String cursor = m_cmdHistoryIndex == index ? " <- current" : "";
            System.out.println(" " + index + " " + o.toString() + cursor);
        }
    }

    // Helper to calculate the undo and redo flags
    protected int getFlags() {
        int flags = 0;
        if (canUndoMore())
            flags |= CAN_UNDO_MORE;
        if (canRedoMore())
            flags |= CAN_REDO_MORE;
        return flags;
    }
}