/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.util.ListIterator;
import java.util.NoSuchElementException;

/** 
 * An adapter for ListIterators that has the behaviour of an empty List that
 * cannot be manipulated.
 * To obtain another behaviour the adapter must be extended by specific
 * classes. The protected methods _size() and _get() must be implemented in the
 * derived classes. To allow the List to be changed the protected methods 
 * _add(), _remove() and _set() must be implemented in the deriving class.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.6 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class ListIteratorAdapter<E> implements ListIterator<E> {
    
    private int m_index = 0;
    private int m_indexToSet = -1;

    // Creation of typical iterators
    public static <E> ListIterator<E> emptyListIterator() {
        return new ListIteratorAdapter<E>();
    }
    public static <E> ListIterator<E> singletonListIterator(E o) {
        return new _SingletonListIterator<E>(o);
    }
    private static class _SingletonListIterator<E> 
    extends ListIteratorAdapter<E> {
        private E m_obj = null;
        private _SingletonListIterator(E obj) {
            m_obj = obj;
        }
        protected int _size() {
            return 1;
        }
        protected E _get(int index) {
            return m_obj;
        }
    }
    
    // =========================================================================
    // Methods to be implemented by deriving classes

    /** 
     * Returns the list size. In this implementation the size 0 s returned.
     * Should be overridden by classes that want another behaviour.
     *      
     * @return the list size
     */
    protected int _size() {
        return 0;
    }

    /** 
     * Returns the list element at index <tt>index</tt>. Should be
     * overridden by deriving classes.
     *      
     * @param index the element index
     * @return the list element at index <tt>index</tt>
     * @throws IllegalArgumentException if there is no such element
     */
    protected E _get(int index) {
        throw new IllegalArgumentException();
    }

    /** 
     * Adds an element at index <tt>index</tt>. Should be overridden by deriving
     * classes that allow adding elements.
     *      
     * @param index the index where to insert the new element
     * @param o the element to add
     * @throws UnsupportedOperationException if this operation is not supported
     */
    protected void _add(int index, E o) {
        throw new UnsupportedOperationException();
    }

    /** 
     * Removes the element at index <tt>index</tt>. Should be overridden by
     * deriving classes that allow removing elements.
     *      
     * @param index the index where to insert the new element
     * @throws UnsupportedOperationException if this operation is not supported
     */
    protected void _remove(int index) {
        throw new UnsupportedOperationException();
    }

    /** 
     * Sets (changes) the element at index <tt>index</tt>. Should be overridden
     * by deriving classes that allow setting elements.
     *      
     * @param index the index of the element to change
     * @throws UnsupportedOperationException if this operation is not supported
     */
    protected void _set(int index, E value) {
        throw new UnsupportedOperationException();
    }

    // =========================================================================
    // ListIterator interface: methods already implemented

    public boolean hasNext() {
        return m_index < _size();
    }
    public boolean hasPrevious() {
        return m_index > 0;
    }
    public E next() {
        if (! hasNext())
            throw new NoSuchElementException();
        m_indexToSet = m_index;
        return _get(m_index++);
    }
    public E previous() {
        if (! hasPrevious())
            throw new NoSuchElementException();
        m_indexToSet = --m_index;
        return _get(m_index);
    }
    public int nextIndex() {
        return m_index;
    }
    public int previousIndex() {
        return m_index - 1;
    }
    public void add(E o) {
        _add(m_index, o);
        m_indexToSet = -1;
    }
    public void remove() {
        if (m_indexToSet == -1)
            throw new IllegalStateException();
        _remove(m_indexToSet);
        m_indexToSet = -1;
    }
    public void set(E o) {
        if (m_indexToSet == -1)
            throw new IllegalStateException(); 
        _set(m_indexToSet, o);
    }
}