/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.util.NoSuchElementException;

/** 
 * An adapter for ListIterators that has the behaviour of an empty List that
 * cannot be manipulated.
 * To obtain another behaviour the adapter must be extended by specific
 * classes. The protected methods _size() and _get() must be implemented in the
 * derived classes. To allow the List to be changed the protected methods 
 * _add(), _remove() and _set() must be implemented in the deriving class.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.6 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class NamedListIteratorAdapter<E> extends ListIteratorAdapter<E>
implements NamedListIterator<E> {

    // =========================================================================
    // Creation of typical iterators

    /** 
     * Returns a <tt>NamedListIterator</tt> for an empty list. The list may not
     * be changed.
     *      
     * @return the <tt>NamedListIterator</tt>.
     */
    public static <E> NamedListIterator<E> emptyNamedListIterator() {
        return new NamedListIteratorAdapter<E>();
    }

    /** 
     * Returns a <tt>NamedListIterator</tt> for a singleton list (i.e. a list
     * containing one single element). The list and the element may not be
     * changed.
     *      
     * @param o the single element of the list 
     * @return the <tt>NamedListIterator</tt>.
     */
    public static <E> NamedListIterator<E> singletonNamedListIterator(E o) {
        return new _SingletonNamedListIterator<E>(o);
    }

    private static class _SingletonNamedListIterator<E>
    extends NamedListIteratorAdapter<E> {
        private E m_obj = null;
        private _SingletonNamedListIterator(E obj) {
            m_obj = obj;
        }
        protected int _size() {
            return 1;
        }
        protected E _get(int index) {
            return m_obj;
        }
    }
    
    
    // =========================================================================
    // Methods to be implemented by deriving classes

    /** 
     * Returns the name of the element at index <tt>index</tt>. In this 
     * implementation the index is returned as name. Should be overridden by
     * classes that want another behaviour.
     *      
     * @param index the element index
     * @return the name of element at index <tt>index</tt>
     */
    protected int _name(int index) {
        return index;
    }
    
    // =========================================================================
    // NamedListIterator interface: methods already implemented

    /** 
     * Returns as name the index of the element that would be returned by a
     * subsequent call to <tt>previous()</tt>.
     *      
     * @return the name as index of the element that would be returned by a
     *  subsequent call to <tt>previous()</tt>.
     * @throws NoSuchElementException if the list iterator is at the beginning
     *  of the list
     */
    public int previousName() throws NoSuchElementException {
        int index = previousIndex();
        if (index == -1)
            throw new NoSuchElementException();
        return _name(index);
    }

    /** 
     * Returns as name the index of the element that would be returned by a
     * subsequent call to <tt>next()</tt>.
     *      
     * @return the name as index of the element that would be returned by a
     *  subsequent call to <tt>next()</tt>.
     * @throws NoSuchElementException if the list iterator is at the end of the
     *  list
     */
    public int nextName() throws NoSuchElementException {
        int index = nextIndex();
        if (index == _size())
            throw new NoSuchElementException();
        return _name(index);
    }
}