/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.NoSuchElementException;

/**
 * A filterable listIterator implementation.
 *
 * @author Bertie Olivier
 * @version $Revision: 1.7 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _FilterableListIterator<E> implements ListIterator<E> {

        protected IteratorFilter<E> m_filter;
        protected ListIterator<E> m_iter;
        protected E obj;
        protected boolean hasNext;
        protected boolean hasPrevious;
        protected boolean searchNext=true;
        protected int nextIndex = -1;
        protected int previousIndex = -1;

        protected _FilterableListIterator(ListIterator<E> iterator) {
            m_iter = iterator;
        }
        protected _FilterableListIterator(ListIterator<E> iterator, 
                IteratorFilter<E> filter) {
            m_filter = filter;
            m_iter = iterator;
            filterNext();
        }

        public boolean hasNext() {
            return hasNext;
        }
        public E next() {
            if (obj == null || !hasNext)
                throw new NoSuchElementException();
            if(!searchNext)
                filterNext();
            searchNext = true;
            E tmpObj = obj;
            previousIndex = nextIndex;
            filterNext();
            hasPrevious = true;
            return tmpObj;
        }

        public boolean hasPrevious() {
            return hasPrevious;
        }
        public E previous() {
            if (obj == null ||!hasPrevious)
                throw new NoSuchElementException();
            if(searchNext)
                filterPrevious();
            searchNext = false;
            E tmpObj = obj;
            nextIndex = previousIndex;
            filterPrevious();
            hasNext = true;
            return tmpObj;
        }
        public int nextIndex() {
            return nextIndex;
        }
        public int previousIndex() {
            return previousIndex;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
        public void set(Object o) {
            throw new UnsupportedOperationException();
        }
        public void add(Object o) {
            throw new UnsupportedOperationException();
        }


        // helper methods
        protected void filterNext() {
            while (m_iter.hasNext()) {
                obj = m_iter.next();
                if (m_filter.matches(obj)) {
                    hasNext = true;
                    nextIndex = m_iter.nextIndex()-1;
                    return;
                }
            }
            searchNext = true;
            nextIndex = m_iter.nextIndex();
            hasNext = false;
        }
        protected void filterPrevious() {
            while (m_iter.hasPrevious()) {
                obj = m_iter.previous();
                if (m_filter.matches(obj)) {
                    hasPrevious = true;
                    previousIndex = m_iter.previousIndex()+1;
                    return;
                }
            }
            previousIndex = -1;
            hasPrevious = false;
        }

        public static void main(String[] args) {
            List<Object> testList = new LinkedList<Object>();

            testList.add(new Integer(0));
            testList.add("A");
            testList.add("B");
            testList.add(new Integer(3));
            testList.add(new Integer(4));
            testList.add("E");
            testList.add("F");
            testList.add(new Integer(7));

            System.err.println(testList);

            IteratorFilter<Object> filter = new IteratorFilter<Object>() {
                public boolean matches(Object o) {
                    return o instanceof Integer;
                }
            };
            ListIterator<Object> i = new _FilterableListIterator<Object>(
                    testList.listIterator(), filter);

            System.out.print("["+i.previousIndex()+"|"+i.nextIndex()+"]->");
            while(i.hasNext())
                System.out.print(" "+i.next()+"<-["+i.previousIndex()+"|"+
                        i.nextIndex()+"]->");
            System.out.println("\n");


            System.out.print("["+i.previousIndex()+"|"+i.nextIndex()+"]->");
            while(i.hasPrevious())
                System.out.print(" "+i.previous()+"<-["+i.previousIndex()+"|"+
                        i.nextIndex()+"]->");
            System.out.println("\n");

        }
}