/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.io.IOException;
import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import cib.util.NameSpace;

/**
 * A wrapper for sets containing named elements only.
 *
 * If an unnamed element is added then a unique name is generated and assigned
 * to the element. It is guaranteed that if such an element is removed from the
 * set and added again in this session (the typical procedure in an undo - redo
 * sequence) then the previously generated name will be reassigned again.
 *
 * @author Berthold Firmenich, Torsten Richter
 * @version $Revision: 1.11 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _NamedElementsSet<E> implements Set<E>, Serializable {

    private static final long serialVersionUID = 0L;

    private Set<E> m_set;
    private NameSpace m_nameSpace;

    private transient Map<E, String> t_generatedNames;

    // Constructors
    protected _NamedElementsSet() {
        // Default constructor disabled
    }
    protected _NamedElementsSet(Set<E> set, NameSpace nameSpace) {
        // Init transient data
        t_generatedNames = new HashMap<E, String>();

        // Check parameters
        if (set == null || nameSpace == null)
            throw new IllegalArgumentException();

        // Store attributes
        m_set = set;
        m_nameSpace = nameSpace;

        // Assign a name if necessary
        for (E e : set)
            assignName(e);
    }

    // Serialization
    private void readObject(java.io.ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        // Init transient data
        t_generatedNames = new HashMap<E, String>();

        // Default deserialization
        in.defaultReadObject();
    }

    // Set implementation
    public boolean add(E e) {
        if (! m_set.add(e))
            return false;   // Set was not changed
        assignName(e);
        return true;
    }
    public boolean addAll(Collection<? extends E> c) {
        if (! m_set.addAll(c))
            return false;   // Set was not changed

        // Assign a name if necessary
        for (E e : c)
            assignName(e);
        return true;
    }
    public void clear() {
        // Remove names
        for (E e : m_set)
            m_nameSpace.remove(e);

        // Clear set
        m_set.clear();
    }
    public boolean contains(Object o) {
        return m_set.contains(o);
    }
    public boolean containsAll(Collection<?> c) {
        return m_set.containsAll(c);
    }
    public boolean equals(Object o) {
        return m_set.equals(o);
    }
    public int hashCode() {
        return m_set.hashCode();
    }
    public boolean isEmpty() {
        return m_set.isEmpty();
    }
    public Iterator<E> iterator() {
        return m_set.iterator();
    }
    public boolean remove(Object o) {
        if (! m_set.remove(o))
            return false;   // Set was not changed

        // Remove name from NameSpace
        m_nameSpace.remove(o);
        return true;
    }
    public boolean removeAll(Collection<?> c) {
        if (! m_set.removeAll(c))
            return false;  // Set was not changed

        // Remove all names from NameSpace
        for (Object object : c)
            m_nameSpace.remove(object);
        return true;
    }
    public boolean retainAll(Collection<?> c) {
        Set<E> old = new HashSet<E>(m_set);
        old.removeAll(c);

        // Remove names from NameSpace
        for (E e : old)
            m_nameSpace.remove(e);
        return m_set.retainAll(c);
    }
    public int size() {
        return m_set.size();
    }
    public Object[] toArray() {
        return m_set.toArray();
    }
    public <T> T[] toArray(T[] a) {
        return m_set.toArray(a);
    }

    // Helpers
    private void assignName(E e) {
        // Name already assigned?
        if (m_nameSpace.containsObject(e))
            return;

        // Re-use previous name if possible
        if (t_generatedNames.containsKey(e)) {
            String name = t_generatedNames.get(e);
            if (! m_nameSpace.containsName(name)) {
                m_nameSpace.put(e, name);
                return;
            }
        }

        // Generate a name and save it for later use
        String name = m_nameSpace.generateName();
        m_nameSpace.put(e, name);
        t_generatedNames.put(e, name);
    }

    public String toString() {
        return m_set.toString();
    }
}