/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import cib.util.ObservableObject;

/**
 * A wrapper for an observable set.
 *
 * @author Berthold Firmenich, Torsten Richter
 * @version $Revision: 1.17 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _ObservableSet<E> implements ObservableSet<E>, Serializable {

    private static final long serialVersionUID = 0L;

    protected Set<E> m_observedSet = null;
    protected Set<ObservableSet.Listener<E>> m_listeners =
        new HashSet<ObservableSet.Listener<E>>();
    protected ObservableObject.Listener m_objectListener = null;

    // Constructors
    protected _ObservableSet(Set<E> set) {
        if (set == null)
            throw new IllegalArgumentException();

        m_observedSet = set;

        m_objectListener = new ObservableObject.Listener() {
            private static final long serialVersionUID = 0L;

            public void wasChanged(ObservableObject oo) {
                for (ObservableSet.Listener<E> osl : m_listeners)
                    osl.wasChanged((E)oo);
            }

            public void wasCloned(ObservableObject src, ObservableObject dst) {
                for (ObservableSet.Listener<E> osl : m_listeners)
                    osl.wasCloned((E)src, (E)dst);
            }
        };
    }

    // ObservableSet implementation
    public boolean addListener(ObservableSet.Listener<E> l) {
        return m_listeners.add(l);
    }

    public boolean removeListener(ObservableSet.Listener<E> l) {
        return m_listeners.remove(l);
    }

    public void clearListeners() {
        m_listeners.clear();
    }

    public Iterator<ObservableSet.Listener<E>> listenerIterator() {
        return m_listeners.iterator();
    }

    public void wasChanged(E e) {
        for (ObservableSet.Listener<E> osl : m_listeners)
            osl.wasChanged(e);
    }

    public void wasCloned(E src, E dst) {
        for (ObservableSet.Listener<E> osl : m_listeners)
            osl.wasCloned(src, dst);
    }

    // Set implementation
    public boolean add(E e) {
        if (!m_observedSet.add(e))
            return false;

        // Notification
        for (ObservableSet.Listener<E> osl : m_listeners)
            osl.wasAdded(e);

        // Add object listener
        if (e instanceof ObservableObject) {
            ObservableObject oo = (ObservableObject) e;
            oo.addListener(m_objectListener);
        }

        return true;
    }

    public boolean addAll(Collection<? extends E> c) {
        HashSet<E> set = new HashSet<E>(c);
        set.removeAll(m_observedSet);   // will be added

        // Operation
        if (! m_observedSet.addAll(c))
            return false;

        // Notification
        for (ObservableSet.Listener<E> osl : m_listeners) {
            for (E e : set) {
                osl.wasAdded(e);
            }
        }

        // Add object listener
        for (E e : set) {
            if (e instanceof ObservableObject) {
                ObservableObject oo = (ObservableObject)e;
                oo.addListener(m_objectListener);
            }
        }

        return true;
    }
    public void clear() {
        removeAll(m_observedSet);
    }

    public boolean contains(Object o) {
        return m_observedSet.contains(o);
    }

    public boolean containsAll(Collection<?> c) {
        return m_observedSet.containsAll(c);
    }

    public boolean equals(Object o) {
        return m_observedSet.equals(o);
    }

    public int hashCode() {
        return m_observedSet.hashCode();
    }

    public boolean isEmpty() {
        return m_observedSet.isEmpty();
    }

    public Iterator<E> iterator() {
        final Iterator<E> it = m_observedSet.iterator();
        return new Iterator<E>() {
            private E m_currentElement = null;
            public boolean hasNext() {
                return it.hasNext();
            }
            public E next() {
                m_currentElement = it.next();
                return m_currentElement;
            }
            public void remove() {
                // Remove the element
                boolean containsObject = m_observedSet.contains(
                        m_currentElement);
                it.remove();

                // Notify
                if (containsObject) {
                    for (ObservableSet.Listener<E> osl : m_listeners)
                        osl.wasRemoved(m_currentElement);

                    // Remove object listener
                    if (m_currentElement instanceof ObservableObject) {
                        ObservableObject oo = 
                                (ObservableObject)m_currentElement;
                        oo.removeListener(m_objectListener);
                    }
                }
            }
        };
    }

    public boolean remove(Object o) {
        if (!m_observedSet.remove(o))
            return false;

        for (ObservableSet.Listener<E> osl : m_listeners)
            osl.wasRemoved((E)o);

        // Remove object listener
        if (o instanceof ObservableObject) {
            ObservableObject oo = (ObservableObject)o;
            oo.removeListener(m_objectListener);
        }

        return true;
    }

    public boolean removeAll(Collection<?> c) {
        HashSet<?> set = new HashSet(c);
        set.retainAll(m_observedSet);   // will be removed

        // Operation
        if (! m_observedSet.removeAll(c))
            return false;

        // Notification
        for (ObservableSet.Listener osl : m_listeners) {
            for (Object object : set) {
                osl.wasRemoved(object);
            }
        }

        // Remove object listeners
        for (Object object : set) {
            if (object instanceof ObservableObject) {
                ObservableObject oo = (ObservableObject)object;
                oo.removeListener(m_objectListener);
            }
        }

        return true;
    }
    public boolean retainAll(Collection<?> c) {
        // Calculate objects to be removed
        HashSet<E> set = new HashSet<E>(m_observedSet);
        set.removeAll(c);    // will be removed

        // Operation
        if (! m_observedSet.retainAll(c))
            return false;

        // Notification
        for (ObservableSet.Listener<E> osl : m_listeners) {
            for (E e : set) {
                osl.wasRemoved(e);
            }
        }

        // Remove object listeners
        for (E e : set) {
            if (e instanceof ObservableObject) {
                ObservableObject oo = (ObservableObject)e;
                oo.removeListener(m_objectListener);
            }
        }

        return true;
    }

    public int size() {
        return m_observedSet.size();
    }

    public Object[] toArray() {
        return m_observedSet.toArray();
    }

    public <T> T[] toArray(T[] a) {
        return m_observedSet.toArray(a);
    }

    public String toString() {
        return m_observedSet.toString();
    }
}