/*
 * Copyright (C) 2009-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.coll;

import java.io.Serializable;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.List;

/**
 * Instances of this class apply a total order to an observable set. The class
 * behaves as an unmodifiable List instance with swappable elements.
 * 
 * Changes of the observed set are reflected by instances of this class.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.2 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
class _OrderedSet<E> extends AbstractList<E> implements SwappableList<E>,
Serializable {

    private static final long serialVersionUID = 0L;

    protected ObservableSet<E> m_set = null;
    protected List<E> m_list = null;

    // Constructors
    protected _OrderedSet(ObservableSet<E> set) {
        m_set = set;
        m_list = new ArrayList<E>(set);
        
        // Sync set and list
        set.addListener(new ObservableSet.Listener<E>() {
            private static final long serialVersionUID = 0L;
            public void wasAdded(E o) {
                if (m_list.contains(o))
                    throw new IllegalArgumentException(
                            "Inconsistent add(): Object was already contained");
                m_list.add(o);
            }
            public void wasRemoved(E o) {
                if (! m_list.contains(o))
                    throw new IllegalArgumentException(
                            "Inconsistent remove(): Object was not contained");
                m_list.remove(o);
            }
        });
    }

    // AbstractList methods
    @Override
    public E get(int index) {
        return m_list.get(index);
    }

    @Override
    public int size() {
        return m_list.size();
    }

    // SwappableList methods
    public void swap(int i, int j) {
        // Prerequisites
        if (i < 0 || j < 0 || i >= size() || j >= size())
            throw new IndexOutOfBoundsException();

        // Swap
        E ei = m_list.get(i);
        E ej = m_list.get(j);
        m_list.set(i, ej);
        m_list.set(j, ei);
        
        // Set changed
        m_set.wasChanged(ei);
        m_set.wasChanged(ej);
    }
}