/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.featureLogic;

import java.io.Reader;
import java.sql.SQLException;
import java.util.Date;
import java.util.Map;
import java.util.Set;

/**
 * The FeatureLogic interface specifies the operations needed for the data
 * manipulation of Feature Logic models. The model can be queried by Feature
 * Logic terms that are interpreted during runtime. Access to the result set is
 * possible via an Iterator object.
 * <p>
 * Error handling is based upon exceptions ({@link FeatureLogicException
 * FeatureLogicException} and {@link java.sql.SQLException SQLException}).
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.15 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public interface FeatureLogic {

//  ------------------------------------------------------------------------
//  Calculation of element sets from feature terms
//  ------------------------------------------------------------------------

   /**
     * Returns a set iterator for a feature logic term.
     *
     * @param r Reader representing the feature term
     * @return iterator
     */
    public ResultSetIterator interpret(Reader r) throws SQLException,
            FeatureLogicException;

    /**
     * Returns a set iterator for a feature logic term.
     *
     * @param s String representing the feature term
     * @return iterator
     */
    public ResultSetIterator interpret(String s) throws SQLException,
            FeatureLogicException;

    /**
     * Returns the element of the singleton described by a feature term.
     *
     * @param r Reader representing the feature term
     * @return Name of the singleton element
     * @exception FeatureLogicException if term doesn't describe a singleton
     */
    public String singleton(Reader r) throws SQLException,
            FeatureLogicException;

    /**
     * Returns the element of the singleton described by a feature term.
     *
     * @param s String representing the feature term
     * @return Name of the singleton element
     * @exception FeatureLogicException if term doesn't describe a singleton
     */
    public String singleton(String s) throws SQLException,
            FeatureLogicException;

//  ------------------------------------------------------------------------
//  Data manipulation
//  ------------------------------------------------------------------------

    /**
     * @return If the batch mode is switched on.
     */
    public boolean isBatchMode();

    /**
     * Can switch on or off the batch mode. The batch mode is mainly used in the
     * database implementations of the FeatureLogic to improve the performance
     * greatly. The FeatureLogic will collect SQL statements and executes them
     * after calling {@link #executeBatch()}.
     *
     * Affected methods are:
     * {@link #addElement()},{@link #addElement(String)},
     * {@link #defineFeature(String)}
     * {@link #addFeature(String, String, String)},
     * {@link #changeFeature(String, String, String)} and
     * <code>findOrCreateAtom(...)</code>.
     *
     * @param batchMode
     */
    public void setBatchMode(boolean batchMode);

    /**
     * Executes collected SQL queries with one step.
     *
     * @throws SQLException
     */
    public void executeBatch() throws SQLException;


    /**
     * Clears the FeatureLogic.
     *
     * @param onlyTransient Clear only transient sets, maps, attributes.
     * @throws SQLException
     */
    public void clear(boolean onlyTransient) throws SQLException;

   /**
     * Adds a new element with a generated name.
     *
     * @return element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String addElement() throws SQLException, FeatureLogicException;

    /**
     * Adds a new element.
     *
     * @param name element name
     * @return <code>true</code> If it is a new element and was added to the domain.
     *         <code>false</code> if it was already in the domain.
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public boolean addElement(String name) throws SQLException, FeatureLogicException;

    /**
     * Removes an element from the domain. Removes all relations where the
     * element is involved.
     *
     * @param name the element ID
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public void removeElement(String name) throws SQLException, FeatureLogicException;

    /**
     * Defines a feature.
     *
     * @param f the feature to define
     * @throws SQLException
     * @throws FeatureLogicException
     */
    public void defineFeature(String f) throws SQLException, FeatureLogicException;

    /**
     * Adds the ordered pair (element, value) to the feature relation. element,
     * value and feature must exist in the data model.
     *
     * @param element 1st component of ordered pair
     * @param f the feature
     * @param value 2nd component of ordered pair
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public void addFeature(String element, String f, String value)
            throws SQLException, FeatureLogicException;

    /**
     * Changes the 2'nd component of the ordered pair (element, oldValue) of the
     * feature relation to a new value.
     *
     * @param element 1st component of ordered pair
     * @param f the feature
     * @param value 2nd component of ordered pair
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public void changeFeature(String element, String f, String value)
            throws SQLException, FeatureLogicException;

    /**
     * Removes an element's feature
     *
     * @param element element name
     * @param feature the feature
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public void removeFeature(String element, String feature)
            throws SQLException, FeatureLogicException;

    /**
     * Finds or creates the primitive element with int value 'value'.
     *
     * @param value the value
     * @return the element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String findOrCreateAtom(int value) throws SQLException,
            FeatureLogicException;

    /**
     * Finds or creates the primitive element with long value 'value'.
     *
     * @param value the value
     * @return the element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String findOrCreateAtom(long value) throws SQLException,
            FeatureLogicException;

    /**
     * Finds or creates the primitive element with double value 'value'.
     *
     * @param value the value
     * @return the element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String findOrCreateAtom(double value) throws SQLException,
            FeatureLogicException;

    /**
     * Finds or creates the primitive element with String value 'value'.
     * Take care to enclose text atoms with quotation marks: value = "\"text\""
     *
     * @param value the value
     * @return the element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String findOrCreateAtom(String value) throws SQLException,
            FeatureLogicException;

    /**
     * Finds or creates the primitive element from the long value stored in
     * the Date 'date'.
     *
     * @param date the date
     * @return the element name
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String findOrCreateAtom(Date date) throws SQLException,
            FeatureLogicException;

//  ------------------------------------------------------------------------
//  Miscellaneous
//  ------------------------------------------------------------------------

   /**
     * Checks whether an element exists in the Domain.
     *
     * @param element the element to be checked
     * @return true if element is in the Domain
     * @throws SQLException
     */
    public boolean isElement(String element) throws SQLException;

    /**
     * Checks whether an element is used as primitive Element.
     *
     * @param element the non primitive element to be checked
     * @return true if element is primitive
     * @throws SQLException
     */
    public boolean isUsedAsPrimitiveElement(String element) throws SQLException;

    /**
     * Checks whether an element is used as non-primitive Element.
     *
     * @param element the non primitive element to be checked
     * @return true if element is non primitive
     * @throws SQLException
     */
    public boolean isUsedAsNonPrimitiveElement(String element)
            throws SQLException;

    /**
     * Checks whether a feature is defined.
     *
     * @param f the feature to be checked
     * @return true if f is a feature
     * @throws SQLException
     */
    public boolean isFeature(String f) throws SQLException;

    /**
     * Extracts the value of a feature.
     *
     * @param element the element with the feature
     * @param f the feature name
     * @return the value of the feature
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public String getFeatureValue(String element, String f)
            throws SQLException, FeatureLogicException;

    /**
     * Adds all features and values of a given element to a Map. The map is
     * cleared before adding the features.
     *
     * @param element the 1st component
     * @param features the Map
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public void getFeatureValues(String element, Map<String,String> features)
            throws SQLException, FeatureLogicException;

    /**
     * Adds all features and values of for each element in a given Set.
     * It returns a Map where each element is mapped to another map containing
     * the features and values.
     *
     * @param elements
     * @return A Map with the features and values for each element in the set.
     * @throws SQLException
     * @throws FeatureLogicException
     */
    public Map<String,Map<String,String>> getFeatureValues(Set<String> elements)
            throws SQLException, FeatureLogicException;

    /**
     * Returns the atomic value of a primitive element.
     *
     * @param prmElement the primitive element
     * @return the value of the atom (Integer, Double or String)
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public Object getAtom(String prmElement) throws SQLException,
            FeatureLogicException;

    /**
     * Returns a map with all primitive elements and their atom objects.
     * Use this method if you have enough RAM and for saving time lost by to
     * many single calls over the network.
     *
     * @return A map.
     * @throws FeatureLogicException
     * @throws SQLException
     */
    public Map<String,Object> getAllAtoms() throws SQLException,
            FeatureLogicException;

//  ------------------------------------------------------------------------
//  Feature Logic algebra
//  ------------------------------------------------------------------------

   /**
     * Gets the set containing Top (set consisting of all elements)
     *
     * @return descriptor of the resulting set
     */
    public SetDsc top();

    /**
     * Gets the set containing Bottom (empty set)
     *
     * @return descriptor of the resulting set
     */
    public SetDsc bottom();

    /**
     * Gets the set containing a single element of the domain.
     *
     * @param x element name
     * @return descriptor of the resulting set
     */
    public SetDsc element(String x);

    /**
     * Gets the primitive element(s) mapped to an int atom
     *
     * @param atom the int constant
     * @return descriptor of the resulting set
     */
    public SetDsc primitiveElement(int atom);

    /**
     * Gets the primitive element(s) mapped to an long atom
     *
     * @param atom the int constant
     * @return descriptor of the resulting set
     */
    public SetDsc primitiveElement(long atom);

    /**
     * Gets the primitive element(s) mapped to a double atom
     *
     * @param atom the double constant
     * @return descriptor of the resulting set
     */
    public SetDsc primitiveElement(double atom);

    /**
     * Gets the primitive element(s) mapped to a string atom
     *
     * @param atom the string constant
     * @return descriptor of the resulting set
     */
    public SetDsc primitiveElement(String atom);

    /**
     * Extraction S.f -- All values of the feature f of objects in S
     *
     * @param f feature f
     * @param s set S
     * @return descriptor of the resulting set
     */
    public SetDsc extract(String f, SetDsc s);

    /**
     * Selection f:S -- All objects where the value of f is S
     *
     * @param f feature f
     * @param s set S
     * @return descriptor of the resulting set
     */
    public SetDsc selection(String f, SetDsc s);

    /**
     * Existence f:[] -- All objects where the feature f is defined
     *
     * @param f feature f
     * @return descriptor of the resulting set
     */
    public SetDsc existence(String f);

    /**
     * Divergence f^ -- All objects where the feature f is not defined
     *
     * @param f feature f
     * @return descriptor of the resulting set
     */
    public SetDsc divergence(String f);

    /**
     * Agreement f==g -- All objects where the features f and g have the same
     * values
     *
     * @param f feature f
     * @param g feature g
     * @return descriptor of the resulting set
     */
    public SetDsc agreement(String f, String g);

    /**
     * Disagreement f!=g -- All objects where the features f and g have
     * different values
     *
     * @param f feature f
     * @param g feature g
     * @return descriptor of the resulting set
     */
    public SetDsc disagreement(String f, String g);

    /**
     * Union {S, T} -- All objects contained in S or T
     *
     * @param s set S
     * @param t set T
     * @return descriptor of the resulting set
     */
    public SetDsc union(SetDsc s, SetDsc t);

    /**
     * Intersection [S, T] -- All objects contained in S and T
     *
     * @param s set S
     * @param t set T
     * @return descriptor of the resulting set
     */
    public SetDsc intersection(SetDsc s, SetDsc t);

    /**
     * Complement ~S -- All objects of the domain that are not contained in S
     *
     * @param s feature term
     * @return descriptor of the resulting set
     */
    public SetDsc complement(SetDsc s);

    /**
     * Difference S~~T -- Relative complement or set-theoretic difference of
     * S and T. The set of elements contains all elements of S that are not in T.
     *
     * @param s set S
     * @param t set T
     * @return descriptor of the resulting set
     */
    public SetDsc difference(SetDsc s, SetDsc t);

    /**
     * Implication S->T -- If an object is contained in S then it must be
     * contained in T
     *
     * @param s set S
     * @param t set T
     * @return descriptor of the resulting set
     */
    public SetDsc implication(SetDsc s, SetDsc t);

    /**
     * Equivalence S<->T -- S holds if and only if T holds
     *
     * @param s set S
     * @param t set T
     * @return descriptor of the resulting set
     */
    public SetDsc equivalence(SetDsc s, SetDsc t);

    /**
     * Gets the cardinality of the set S.
     *
     * @param s set S
     * @return descriptor of the resulting set
     */
    public int cardinality(SetDsc s) throws SQLException, FeatureLogicException;

//  ------------------------------------------------------------------------
//  Iteration of the result set
//  ------------------------------------------------------------------------

   /**
     * Returns the result set iterator of the set S.
     *
     * @param s set S
     * @return Result set iterator
     */
    public ResultSetIterator resultSetIterator(SetDsc s) throws SQLException;
}