/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.featureLogic.db.access;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import cib.util.featureLogic.FeatureLogic;
import cib.util.featureLogic.FeatureLogicAdapter;
import cib.util.featureLogic.FeatureLogicException;
import cib.util.featureLogic.ResultSetIterator;
import cib.util.featureLogic.ResultSetIteratorImp;
import cib.util.featureLogic.SetDsc;
import cib.util.featureLogic.interpreter.Interpreter;
import cib.util.featureLogic.interpreter.ParseException;
import cib.util.featureLogic.interpreter.TokenMgrError;

/**
 * Implementation of the FeatureLogic class for Access databases.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.13 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class FeatureLogicAccess extends FeatureLogicAdapter {

    private Connection m_con = null;
    private int m_iSet = 0;

    public static void main(String[] args)
    throws ClassNotFoundException, SQLException {
        // Commandline args
        if (args.length == 0) {
            System.out.println(
                "usage:  FeatureLogicAccess -dbURL <JDBC database URL>");
            System.out.println(
                "sample: FeatureLogicAccess -dbURL jdbc:odbc:FeatureLogic");
            System.exit(0);
        }

        // Get command line switsches
        String databaseURL = null;
        for (int i = 0; i < args.length; i++) {
            if (args[i].equals("-dbURL"))
                databaseURL = args[++i];
        }
        if (databaseURL == null) {
            System.out.println("Please enter the database URL");
            System.exit(1);
        }

        // Database constants
        final String driver         = "sun.jdbc.odbc.JdbcOdbcDriver";
        final String user         = "";
        final String password     = "";

        // Load JDBC driver
        System.out.println("Loading driver " + driver + " ...");
        Class.forName(driver);

        // Connect to database
        System.out.println("Connecting to database " + databaseURL + " ...");
        Connection con = DriverManager.getConnection(databaseURL, user,
            password);

        Interpreter interpreter = new Interpreter(System.in);
        FeatureLogic featureLogic = new FeatureLogicAccess(con);
        interpreter.setFeatureLogic(featureLogic);
        while (true) {
            interpreter.ReInit(System.in);
            System.out.print("Enter Feature Term: ");
            System.out.flush();
            try {
                SetDsc setDsc = interpreter.start();
                if (setDsc == null)
                    System.exit(0);
                   System.out.println(setDsc); // Temporary show string
                ResultSetIterator rsi = featureLogic.resultSetIterator(setDsc);
                while (rsi.hasNext()) {
                    System.out.println("\t" + rsi.next());
                }
            }
            catch (TokenMgrError e) {
                System.out.println(e);
            }
            catch (ParseException e) {
                System.out.println(e);
            }
            catch (SQLException e) {
                System.out.println(e);
            }
            catch (FeatureLogicException e) {
                System.out.println(e);
            }
        }
    }

   /**
    * @param con the JDBC database connection
    */
    public FeatureLogicAccess(Connection con) throws SQLException {
        m_con = con;
        clearResultSet();
        m_iSet = 0;
    }

//  ------------------------------------------------------------------------
//  Data manipulation
//  ------------------------------------------------------------------------

    public boolean isBatchMode() {
        return false;
    }

    public void setBatchMode(boolean batchMode) {
    }

    public void executeBatch() throws SQLException {
    }

    public void clear(boolean onlyTransient) throws SQLException {
        if (m_con == null)
            return;

        Statement stmt = null;
        stmt = m_con.createStatement();
        stmt.addBatch("DELETE FROM ATOM");
        stmt.addBatch("DELETE FROM DOMAIN");
        stmt.addBatch("DELETE FROM FEATURE");
        stmt.addBatch("DELETE FROM RELSLOT");
        stmt.addBatch("DELETE FROM NextAvailableId");
        stmt.addBatch("INSERT INTO NextAvailableId (ID) VALUES (1)");
        stmt.addBatch("DELETE * FROM ResultSet");
        stmt.executeBatch();
        stmt.close();
    }

    private long _generateUniqueId() throws SQLException {

        // Get current component ID
        String sql = "SELECT id FROM NextAvailableId;";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);
        rs.next();
        long idCmp = rs.getLong(1);
        rs.close();
        stmt.close();

        // Update component ID
        long idCmpNext = idCmp + 1;
        sql =
            "UPDATE NextAvailableId SET id=" + idCmpNext + " WHERE id=" +
            idCmp + ";";
        executeUpdate(sql);

        return idCmp;
    }

    private boolean _storeElement(String element) throws SQLException {
        if (! isElement(element)) {
            String sql =
                "INSERT INTO Domain (element) SELECT '" + element + "';";
            executeUpdate(sql);
            return true;
        }
        return false;
    }

    public String addElement() throws SQLException {
        long id = _generateUniqueId();
        String element = "_" + id;
        _storeElement(element);
        return element;
    }

    public boolean addElement(String name) throws SQLException {
        return _storeElement(name);
    }

    public void removeElement(String id) throws SQLException {
        String sql = "DELETE * FROM Domain WHERE element='" + id + "';";
        executeUpdate(sql);
    }

    public void defineFeature(String feature) throws SQLException {
        if (! isFeature(feature)) {
            String sql =
                "INSERT INTO Feature (feature) SELECT '" + feature + "';";
            executeUpdate(sql);
        }
    }

    public void addFeature(String element, String feature, String value)
        throws SQLException, FeatureLogicException {

        // Ensure consistency of the data model
        checkElement(element);
        checkUsableAsNonPrimitiveElement(element);
        checkElement(value);
        checkFeature(feature);

        // Add feature pair
        String sql =
            "INSERT INTO RelSlot (object, feature, [value]) " +
            "SELECT '" + element + "','" + feature + "','" + value + "';";
        executeUpdate(sql);
    }

    public void changeFeature(String element, String feature, String value)
        throws SQLException, FeatureLogicException {

        // Ensure consistency of the data model
        checkElement(value);

        // Modify feature pair
        String sql =
            "UPDATE RelSlot SET [value]='" + value + "' " +
            "WHERE object='" + element + "' AND feature='" + feature + "';";
        executeUpdate(sql);
    }

    public void removeFeature(String element, String feature)
        throws SQLException, FeatureLogicException {
        throw new UnsupportedOperationException();
    }

    public String findOrCreateAtom(int value)
        throws SQLException, FeatureLogicException {
        return _findOrCreateLiteral(String.valueOf(value));
    }

    public String findOrCreateAtom(long value)
        throws SQLException, FeatureLogicException {
        return _findOrCreateLiteral(String.valueOf(value) + "L");
    }

    public String findOrCreateAtom(double value)
        throws SQLException, FeatureLogicException {
        return _findOrCreateLiteral(String.valueOf(value));
    }

    public String findOrCreateAtom(String value)
        throws SQLException, FeatureLogicException {
        return _findOrCreateLiteral("\"" + value + "\"");
    }

    public String findOrCreateAtom(Date date)
        throws SQLException, FeatureLogicException {
        return _findOrCreateLiteral(String.valueOf(date.getTime()) + "DATE");
    }

    private String _findOrCreateLiteral(String literal) throws SQLException {

        // Does the literal already exist?
        String sql =
            "SELECT element FROM Atom WHERE value = '" + literal + "';";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);
        if (rs.next()) {
            String element = rs.getString(1);
            rs.close();
            stmt.close();
            return element;
        }
        stmt.close();

        // Create a new literal
        String element = addElement();
        sql =
            "INSERT INTO Atom (element, [value]) " +
            "SELECT '" + element + "','" + literal + "';";
        executeUpdate(sql);

        // Return the element
        return element;
    }

//  ------------------------------------------------------------------------
//  Miscellaneous
//  ------------------------------------------------------------------------

    public boolean isElement(String element) throws SQLException {
        String sql =
            "SELECT Count(*) AS n FROM Domain " +
            "WHERE Domain.element='" + element + "';";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        rs.next();
        int n = rs.getInt(1);

        rs.close();
        stmt.close();

        return n != 0;
    }

    public boolean isUsedAsPrimitiveElement(String element)
    throws SQLException {
        String sql =
            "SELECT Count(*) AS n FROM Atom " +
            "WHERE Atom.element='" + element + "';";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        rs.next();
        int n = rs.getInt(1);

        rs.close();
        stmt.close();

        return n != 0;
    }

    public boolean isUsedAsNonPrimitiveElement(String element)
    throws SQLException {
        String sql =
            "SELECT Count(*) AS n FROM RelSlot " +
            "WHERE RelSlot.object='" + element + "';";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        rs.next();
        int n = rs.getInt(1);

        rs.close();
        stmt.close();

        return n != 0;
    }

    public boolean isFeature(String feature) throws SQLException {
        String sql =
            "SELECT Count(*) AS n FROM Feature " +
            "WHERE Feature.feature='" + feature + "';";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        rs.next();
        int n = rs.getInt(1);

        rs.close();
        stmt.close();

        return n != 0;
    }

    public String getFeatureValue(String element, String feature)
    throws SQLException, FeatureLogicException {

        checkElement(element);
        checkFeature(feature);

        String sql =
            "SELECT value " +
            "FROM RelSlot " +
            "WHERE object='" + element + "' AND feature='" + feature + "';";

        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);
        String value = rs.next() ?
            rs.getString(1) :
            null;
        rs.close();
        stmt.close();

        return value;
    }

    public void getFeatureValues(String element, Map<String,String> features)
    throws SQLException, FeatureLogicException {

        checkElement(element);

        String sql =
            "SELECT feature, value " +
            "FROM RelSlot " +
            "WHERE object='" + element + "';";

        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        features.clear();
        while (rs.next()) {
            features.put(rs.getString(1), rs.getString(2));
        }

        rs.close();
        stmt.close();
    }

    public Map<String,Map<String,String>> getFeatureValues(Set<String> elements)
    throws SQLException, FeatureLogicException {
        Map<String,Map<String,String>> map = new HashMap<String,Map<String,String>>();

        for (String element : elements) {
            Map<String,String> features = new HashMap<String,String>();
            getFeatureValues(element, features);

            map.put(element, features);
        }
        return map;
    }

    public Object getAtom(String prmElement)
    throws SQLException, FeatureLogicException {
        checkElement(prmElement);

        // Get value from database
        String sql =
            "SELECT value " +
            "FROM Atom " +
            "WHERE element='" + prmElement + "';";

        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        if (! rs.next())
            throw new
                FeatureLogicException("getAtom(): no primitive element");

        String value = rs.getString(1);
        rs.close();
        stmt.close();

        // Convert string value to typed value
        Object typedValue;
        if (value.startsWith("\""))
            typedValue = value.substring(1, value.length() - 1);
        else if (-1 != value.indexOf('.'))
            typedValue = new Double(value);
        else if (value.endsWith("L")) {
            value = value.substring(0, value.length() - 1);
            typedValue = new Long(value);
        }
        else if (value.endsWith("DATE")) {
            value = value.substring(0, value.length() - 4);
            typedValue = new Date(new Long(value));
        }
        else
            typedValue = new Integer(value);

        return typedValue;
    }

    public Map<String,Object> getAllAtoms() throws SQLException,
            FeatureLogicException {
        throw new UnsupportedOperationException();
    }

//  ------------------------------------------------------------------------
//  Feature Logic algebra
//  ------------------------------------------------------------------------

    private void clearResultSet() throws SQLException {
        final String sql = "DELETE * FROM ResultSet;";
        executeUpdate(sql);
    }

    private int getIndex() {
        return m_iSet++;
    }

    public SetDsc top() {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Domain.element AS object " +
            "FROM Domain;";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc bottom() {
        int outSet = getIndex();
        return new SetDscAccess(outSet);
    }

    public SetDsc element(String x) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Domain.element AS object " +
            "FROM Domain " +
            "WHERE element='" + x + "';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc primitiveElement(int atom) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Atom.element AS object " +
            "FROM Atom WHERE value = '" + atom + "';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc primitiveElement(long atom) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Atom.element AS object " +
            "FROM Atom WHERE value = '" + atom + "';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc primitiveElement(double atom) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Atom.element AS object " +
            "FROM Atom WHERE value = '" + atom + "';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc primitiveElement(String atom) {
        int outSet = getIndex();
        String sql =
            "SELECT element FROM Atom WHERE value = '\"" + atom + "\"';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc extract(String f, SetDsc s) {
        int outSet = getIndex();
        int inSet = ((SetDscAccess)s).getIndex();
        String sql =
            "INSERT INTO ResultSet (object, iSet) " +
            "SELECT DISTINCT RelSlot.value AS object, " + outSet +
                " AS iSet " +
            "FROM RelSlot INNER JOIN ResultSet " +
            "ON RelSlot.object=ResultSet.object " +
            "WHERE ResultSet.iSet=" + inSet + " " +
            "AND RelSlot.feature='" + f + "';";
        ((SetDscAccess)s).add(outSet, sql);
        return s;
    }

    public SetDsc selection(String f, SetDsc s) {
        int outSet = getIndex();
        int inSet = ((SetDscAccess)s).getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, RelSlot.object AS object " +
            "FROM RelSlot, ResultSet " +
            "WHERE RelSlot.feature='" + f + "' " +
            "AND RelSlot.value=ResultSet.object " +
            "AND ResultSet.iSet=" + inSet + ";";
        ((SetDscAccess)s).add(outSet, sql);
        return s;
    }

    public SetDsc existence(String f) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, RelSlot.object AS object " +
            "FROM RelSlot " +
            "WHERE feature='" + f + "';";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc divergence(String f) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, RelSlot.object AS object " +
            "FROM RelSlot " +
            "WHERE object NOT IN (" +
                "SELECT object " +
                "FROM RelSlot " +
                "WHERE feature='" + f + "'" +
            ");";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc agreement(String f, String g) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, RelSlot.object AS object " +
            "FROM RelSlot INNER JOIN RelSlot AS RS ON RelSlot.object = RS.object " +
            "WHERE RelSlot.feature='" + f + "' " +
            "AND RS.feature='" + g + "' " +
            "AND RelSlot.value = RS.value;";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc disagreement(String f, String g) {
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, RelSlot.object AS object " +
            "FROM RelSlot INNER JOIN RelSlot AS RS ON RelSlot.object = RS.object " +
            "WHERE RelSlot.feature='" + f + "' " +
            "AND RS.feature='" + g + "' " +
            "AND RelSlot.value <> RS.value;";
        return new SetDscAccess(outSet, sql);
    }

    public SetDsc union(SetDsc s, SetDsc t) {
        int inSet1 = ((SetDscAccess)s).getIndex();
        int inSet2 = ((SetDscAccess)t).getIndex();
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, ResultSet.object AS object " +
            "FROM ResultSet " +
            "WHERE ResultSet.iSet=" + inSet1 + " OR ResultSet.iSet=" + inSet2 + ";";
        ((SetDscAccess)s).add(outSet, (SetDscAccess)t);
        ((SetDscAccess)s).add(outSet, sql);
        return s;
    }

    public SetDsc intersection(SetDsc s, SetDsc t) {
        int inSet1 = ((SetDscAccess)s).getIndex();
        int inSet2 = ((SetDscAccess)t).getIndex();
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT DISTINCT " + outSet + " AS iSet, ResultSet.object AS object " +
            "FROM ResultSet INNER JOIN ResultSet AS RS ON ResultSet.object=RS.object " +
            "WHERE ResultSet.iSet=" + inSet1 + " " +
            "AND RS.iSet=" + inSet2 + " " +
            "AND ResultSet.object=RS.object;";
        ((SetDscAccess)s).add(outSet, (SetDscAccess)t);
        ((SetDscAccess)s).add(outSet, sql);
        return s;
    }

    public SetDsc difference(SetDsc s, SetDsc t) {
        throw new UnsupportedOperationException("Set difference not yet implemented.");
    }

    public SetDsc complement(SetDsc s) {
        int inSet = ((SetDscAccess)s).getIndex();
        int outSet = getIndex();
        String sql =
            "INSERT INTO ResultSet (iSet, object) " +
            "SELECT " + outSet + " AS iSet, Domain.element AS object " +
            "FROM Domain " +
            "WHERE Domain.element NOT IN (" +
                "SELECT object " +
                "FROM ResultSet " +
                "WHERE ResultSet.iSet=" + inSet +
            ");";
        ((SetDscAccess)s).add(outSet, sql);
        return s;
    }

    public SetDsc implication(SetDsc s, SetDsc t) {
       return union(complement(s), t);
    }

    public SetDsc equivalence(SetDsc s, SetDsc t) {
        return intersection(implication(s, t), implication(t, s));
    }

    public int cardinality(SetDsc s)
    throws SQLException, FeatureLogicException {

        Iterator<String> it = ((SetDscAccess)s).sqlIterator();
        while (it.hasNext()) {
            String sql = it.next();
            executeUpdate(sql);
        }

        int inSet = ((SetDscAccess)s).getIndex();
        String sql =
            "SELECT Count(*) AS n FROM ResultSet " +
            "WHERE ResultSet.iSet=" + inSet + ";";
        Statement stmt = m_con.createStatement();
        ResultSet rs = stmt.executeQuery(sql);

        int n = 0;
        if (rs.next())
            n = rs.getInt(1);

        rs.close();
        stmt.close();

        return n;
    }

//  ------------------------------------------------------------------------
//  Iteration of the result set
//  ------------------------------------------------------------------------

    public ResultSetIterator resultSetIterator(SetDsc s) throws SQLException {

        Iterator<String> it = ((SetDscAccess)s).sqlIterator();
        while (it.hasNext()) {
            String sql = it.next();
            executeUpdate(sql);
        }

        int inSet = ((SetDscAccess)s).getIndex();
        String sql =
            "SELECT * FROM ResultSet WHERE iSet=" + inSet + ";";
        Statement statement = m_con.createStatement();
        ResultSet resultSet = statement.executeQuery(sql);
        return new ResultSetIteratorImp(statement, resultSet);
    }

//  ------------------------------------------------------------------------
//  Helpers
//  ------------------------------------------------------------------------

   /**
    * Excecutes a SQL Update statement
    * @param sql SQL string
    */
    private void executeUpdate(String sql) throws SQLException {
        Statement stmt = m_con.createStatement();
        stmt.executeUpdate(sql);
        stmt.close();
    }

   /**
 * Checks whether an element exists in the Domain. On error an exception
    * is thrown.
    * @param element the element to be checked
    */
    private void checkElement(String element)
        throws SQLException, FeatureLogicException {

        if (! isElement(element))
            throw new FeatureLogicException("Invalid element " + element);
    }

   /**
 * Checks whether an element is usable as non-primitive element. If the
    * element is already used as primitive an exception is thrown.
    * @param element the element to be checked
    */
    private void checkUsableAsNonPrimitiveElement(String element)
        throws SQLException, FeatureLogicException {

        if (isUsedAsPrimitiveElement(element))
            throw new FeatureLogicException(
                "Attempt to use the primitive element " + element +
                " as non-primitive element");
    }

   /**
 * Checks whether an element is usable as non-primitive element. If the
    * element is already used as primitive an exception is thrown.
    * @param element the element to be checked
    */
/**
    private void checkUsableAsPrimitiveElement(String element)
        throws SQLException, FeatureLogicException {

        if (isUsedAsNonPrimitiveElement(element))
            throw new FeatureLogicException(
                "Attempt to use the non-primitive element " + element +
                " as primitive element");
    }
**/

    /**
 * Checks whether a feature is defined. On error an exception is thrown.
    * @param feature the feature name to be checked
    */
    private void checkFeature(String feature)
        throws SQLException, FeatureLogicException {

        if (! isFeature(feature))
            throw new FeatureLogicException("Invalid feature " + feature);
    }
}