/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.geo;

import java.awt.Shape;
import java.awt.geom.Arc2D;
import java.awt.geom.CubicCurve2D;
import java.awt.geom.Ellipse2D;
import java.awt.geom.Line2D;
import java.awt.geom.PathIterator;
import java.awt.geom.Point2D;
import java.awt.geom.QuadCurve2D;
import java.awt.geom.Rectangle2D;
import java.text.DecimalFormat;
import java.text.DecimalFormatSymbols;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Locale;
import java.util.NoSuchElementException;
import java.util.Set;

import cib.util.CoordSpace;

/**
 * Supplies methods for geometric calculations in 2D coordinate space. To reuse
 * the geometric data types from the package java.awt.geom all methods are
 * static.<br/>
 * <br/>
 * The supported geometric types are
 *     {@link Vector2D},
 *     {@link java.awt.geom.Point2D},
 *     {@link java.awt.geom.Line2D} (both bounded segments and unbounded lines),
 *     {@link java.awt.geom.Ellipse2D},
 *     {@link java.awt.geom.Arc2D},
 *     {@link java.awt.geom.QuadCurve2D} and
 *     {@link java.awt.geom.CubicCurve2D}.<br/>
 * <br/>
 * Length and null properties can be calculated. Geometric calculations are
 * length, area, intersection points, equality, parallelism, perpendicularity
 * and containment.<br/>
 * <br/>
 * The accuracy of calculations can be controlled by an epsilon value.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.38 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/06/17 11:25:32 $
 */
public class Geo2D {

    // =========================================================================
    // Accuracy control ========================================================
    // =========================================================================

    private static double m_eps = 1.e-6;
    private static double m_epsSqr = m_eps * m_eps;

   /**
    * Returns the epsilon value. Two points with a distance less than or equal
    * to epsilon are considered coincident.
    * @return the epsilon value
    */
    public static double getEps() {
        return m_eps;
    }

   /**
    * Returns the squared epsilon value.
    * @return the squared epsilon value
    */
    public static double getEpsSqr() {
        return m_epsSqr;
    }

   /**
    * Sets the epsilon value for subsequent calculations.
    * @param eps the epsilon value
    */
    public static void setEps(double eps) {
         m_eps = eps;
         m_epsSqr = eps * eps;
    }

   /**
    * Rounds a double value for output as a text string in a decimal format.
    *
    * Background: Because the accuracy of the floating-point format is limited
    * real values usually cannot be represented accurately. Instead, a given
    * real value is represented by the closest floating-point value available.
    * This strategy, however, causes problems when the floating point value is
    * output as text. For instance, a value of 30.0 is rounded to (and displayed
    * as) 29.999999999999996. The users are confused by this behavior.
    * The method round() rounds a double value to avoid the effects described.
    * The value is rounded to multiples of 10 exp n where n is the minimum
    * integer number possible.
    *
    * @param x the value to round
    * @return the rounded value
    */
    @Deprecated
    protected static double _round(double x) {
        double ulp = Math.ulp(x);
        double exp = Math.log10(ulp);
        double expRounded = Math.ceil(exp) + 1;
        double divisor = Math.pow(10, expRounded);
        System.out.println("x=" + x + " ulp=" + ulp + " exp=" + exp +
            " expRounded=" + expRounded + " divisor=" + divisor + " result=" +
            (x - Math.IEEEremainder(x, divisor)));
        return (x - Math.IEEEremainder(x, divisor));
        //return (x - Math.IEEEremainder(x, 1.e-16));
    }

    /**
     * Returns a rounded double as string (same as formatCoordinate()).
     *
     * @param x the value to round
     * @return the rounded value as String
     */
    @Deprecated
    public static String roundToString(double x) {
        return formatCoordinate(x);
    }

    /**
     * Returns a rounded double as string (same as formatCoordinate()).
     *
     * @param x the value to round
     * @return the rounded value as String
     */
    @Deprecated
    public static String round(double x) {
        return formatCoordinate(x);
    }

    /**
     * Formats a coordinate according to the formatting rules defined in
     * {@link cib.util.CoordSpace}.
     * 
     * @param x the value to round
     * @return the rounded coordinate
     */
    public static String formatCoordinate(double x) {
        if (Double.isNaN(x))
            return "NaN";
        CoordSpace cs = CoordSpace.getCoordSpace();
        return cs.formatCoordinate(x);
    }

    /**
     * Formats an angle given in degrees.
     * 
     * @param degrees the degrees to round
     * @return the rounded degrees
     */
    public static String formatDegrees(double degrees) {
        if (Double.isNaN(degrees))
            return "NaN";
        DecimalFormat nf = new DecimalFormat();
        nf.setDecimalFormatSymbols(new DecimalFormatSymbols(Locale.ENGLISH)); 
        nf.applyLocalizedPattern("0.0##########");
        return nf.format(degrees);
    }
    
    /**
     * Formats an abstract double.
     * 
     * @param value the value to round
     * @return the rounded value
     */
    public static String formatDouble(double value) {
        return Double.toString(value);
    }

    // =========================================================================
    // Length ==================================================================
    // =========================================================================

   /**
    * Calculates the length of a vector.
    * @param vec the vector
    * @return the length
    */
    public static double length(Vector2D vec) {
        return Math.sqrt(vec.getX() * vec.getX() + vec.getY() * vec.getY());
    }

   /**
    * Calculates the length of a line segment.
    * @param seg the line segment
    * @return the length
    */
    public static double length(Line2D seg) {
        double dx = seg.getX2() - seg.getX1();
        double dy = seg.getY2() - seg.getY1();
        return Math.sqrt(dx * dx + dy * dy);
    }

   /**
    * Calculates the length of an ellipse.
    * @param ell the ellipse
    * @return the length
    */
    public static double length(Ellipse2D ell) {
        return _length(ell.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the length of an arc.
    * @param arc the arc
    * @return the length
    */
    public static double length(Arc2D arc) {
        return _length(arc.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the length of a quadratic parametric curve segment.
    * @param qad the quadratic parametric curve segment
    * @return the length
    */
    public static double length(QuadCurve2D qad) {
        return _length(qad.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the length of a cubic parametric curve segment.
    * @param cub the cubic parametric curve segment
    * @return the length
    */
    public static double length(CubicCurve2D cub) {
        return _length(cub.getPathIterator(null, m_eps));
    }

    // Flattening path iterator version
    private static double _length(PathIterator pit) {
        double lenSq = 0;

        Line2D lineSeg = new Line2D.Double();
        double[] mvto = new double[2];
        double[] last = new double[2];
        double[] crds = new double[2];
        for (; ! pit.isDone(); pit.next()) {
            int typ = pit.currentSegment(crds);
            switch (typ) {
            case PathIterator.SEG_MOVETO :
                mvto[0] = crds[0];
                mvto[1] = crds[1];
                last[0] = crds[0];
                last[1] = crds[1];
            break;
            case PathIterator.SEG_LINETO : {
                lineSeg.setLine(last[0], last[1], crds[0], crds[1]);
                last[0] = crds[0];
                last[1] = crds[1];

                double dx = lineSeg.getX2() - lineSeg.getX1();
                double dy = lineSeg.getY2() - lineSeg.getY1();
                lenSq += dx * dx + dy * dy;
            }
            break;
            case PathIterator.SEG_CLOSE : {
                lineSeg.setLine(last[0], last[1], mvto[0], mvto[1]);

                double dx = lineSeg.getX2() - lineSeg.getX1();
                double dy = lineSeg.getY2() - lineSeg.getY1();
                lenSq += dx * dx + dy * dy;
            }
            break;
            default :
                throw new Error("Not a flattening path iterator");
            }
        }

        return Math.sqrt(lenSq);
    }

    // =========================================================================
    // Area ====================================================================
    // =========================================================================

    /**
     * Calculates the area by integrating a line segment.
     * @param seg the line segment
     * @return the signed area
     */
     public static double area(Line2D seg) {
         double dx = seg.getX2() - seg.getX1();
         double dy = .5 * (seg.getY1() + seg.getY2());
         return dx * dy;
     }

    /**
     * Calculates the area by integrating an ellipse.
     * @param ell the ellipse
     * @return the signed area
     */
     public static double area(Ellipse2D ell) {
         return _area(ell.getPathIterator(null, m_eps));
     }

    /**
     * Calculates the area by integrating an arc.
     * @param arc the arc
     * @return the signed area
     */
     public static double area(Arc2D arc) {
         return _area(arc.getPathIterator(null, m_eps));
     }

    /**
     * Calculates the area by integrating a quadratic parametric curve segment.
     * @param qad the quadratic parametric curve segment
     * @return the signed area
     */
     public static double area(QuadCurve2D qad) {
         return _area(qad.getPathIterator(null, m_eps));
     }

    /**
     * Calculates the area by integrating a cubic parametric curve segment.
     * @param cub the cubic parametric curve segment
     * @return the signed area
     */
     public static double area(CubicCurve2D cub) {
         return _area(cub.getPathIterator(null, m_eps));
     }

     // Flattening path iterator version
     private static double _area(PathIterator pit) {
         double area = 0;

         Line2D lineSeg = new Line2D.Double();
         double[] mvto = new double[2];
         double[] last = new double[2];
         double[] crds = new double[2];
         for (; ! pit.isDone(); pit.next()) {
             int typ = pit.currentSegment(crds);
             switch (typ) {
             case PathIterator.SEG_MOVETO :
                 mvto[0] = crds[0];
                 mvto[1] = crds[1];
                 last[0] = crds[0];
                 last[1] = crds[1];
             break;
             case PathIterator.SEG_LINETO : {
                 lineSeg.setLine(last[0], last[1], crds[0], crds[1]);
                 last[0] = crds[0];
                 last[1] = crds[1];

                 double dx = lineSeg.getX2() - lineSeg.getX1();
                 double dy = .5 * (lineSeg.getY1() + lineSeg.getY2());
                 area += dx * dy;
             }
             break;
             case PathIterator.SEG_CLOSE : {
                 lineSeg.setLine(last[0], last[1], mvto[0], mvto[1]);

                 double dx = lineSeg.getX2() - lineSeg.getX1();
                 double dy = .5 * (lineSeg.getY1() + lineSeg.getY2());
                 area += dx * dy;
             }
             break;
             default :
                 throw new Error("Not a flattening path iterator");
             }
         }

         return area;
     }
    
    // =========================================================================
    // Distance ================================================================
    // =========================================================================

    /**
     * Calculates the distance between two points.
     * @param pnt1 1st point
     * @param pnt2 2nd point
     * @return the distance
     */
     public static double distance(Point2D pnt1, Point2D pnt2) {
         return pnt1.distance(pnt2);
     }

     /**
      * Calculates the distance between two points.
      * @param x1 x1
      * @param y1 y1
      * @param x2 x2
      * @param y2 y2
      * @return the distance
      */
      public static double distance(double x1, double y1, double x2,
              double y2) {
          double dx = x2 - x1;
          double dy = y2 - y1;
          return Math.sqrt(dx * dx + dy * dy);
      }

   /**
    * Calculates the distance between a point and a line segment.
    * @param pnt the point
    * @param seg the line segment
    * @return the distance
    */
    public static double distance(Point2D pnt, Line2D seg) {
        // Bug in 1.3.1_04
        // return seg.ptSegDist(pnt);
        return Math.sqrt(Math.abs(seg.ptSegDistSq(pnt)));
    }

   /**
    * Calculates the distance between a point and a line(segment).
    * @param pnt the point
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @return the distance
    */
    public static double distance(Point2D pnt, Line2D any, boolean unbounded) {
        // Bug in 1.3.1_04
        // return unbounded ? any.ptLineDist(pnt) : any.ptSegDist(pnt);
        return unbounded ?
            Math.sqrt(Math.abs(any.ptLineDistSq(pnt))) :
            Math.sqrt(Math.abs(any.ptSegDistSq(pnt)));
    }

   /**
    * Calculates the distance between a point and an ellipse.
    * @param pnt the point
    * @param ell the ellipse
    * @return the distance
    */
    public static double distance(Point2D pnt, Ellipse2D ell) {
        // Distance point circle?
        if (Math.abs(ell.getWidth() - ell.getHeight()) < m_eps) {
            double r = .5 * ell.getWidth();
            return Math.abs(distance(pnt.getX(), pnt.getY(), ell.getCenterX(), 
                    ell.getCenterY()) - r);
        }
        // Distance point ellipse
        return _distance(pnt, ell.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the distance between a point and an arc.
    * @param pnt the point
    * @param arc the arc
    * @return the distance
    */
    public static double distance(Point2D pnt, Arc2D arc) {
        // Distance point circular arc?
        if (Math.abs(arc.getWidth() - arc.getHeight()) < m_eps) {
            // Distance point circle
            double r = .5 * arc.getWidth();
            double distCircle = Math.abs(distance(pnt.getX(), pnt.getY(), 
                    arc.getCenterX(), arc.getCenterY()) - r);

            // Projected point on arc?
            Vector2D cpPnt = new Vector2D(arc.getCenterX(), arc.getCenterY(), 
                    pnt.getX(), pnt.getY());
            if (isNull(cpPnt))
                return distCircle;

            double angle = - Math.toDegrees(Math.atan2(cpPnt.y, cpPnt.x));
            if (arc.containsAngle(angle))
                return distCircle;
            return Math.min(distance(arc.getStartPoint(), pnt), 
                    distance(arc.getEndPoint(), pnt));
        }
        // Distance point ellipse
        return _distance(pnt, arc.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the distance between a point and a quadratic parametric curve
    * segment.
    * @param pnt the point
    * @param qad the quadratic parametric curve segment
    * @return the distance
    */
    public static double distance(Point2D pnt, QuadCurve2D qad) {
        return _distance(pnt, qad.getPathIterator(null, m_eps));
    }

   /**
    * Calculates the distance between a point and a cubic parametric curve
    * segment.
    * @param pnt the point
    * @param cub the cubic parametric curve segment
    * @return the distance
    */
    public static double distance(Point2D pnt, CubicCurve2D cub) {
        return _distance(pnt, cub.getPathIterator(null, m_eps));
    }

    // Flattening path iterator version
    private static double _distance(Point2D pnt, PathIterator pit) {
        if (pit.isDone())
            return Double.NaN;

        double distSq = Double.MAX_VALUE;

        Line2D lineSeg = new Line2D.Double();
        double[] mvto = new double[2];
        double[] last = new double[2];
        double[] crds = new double[2];
        for (; ! pit.isDone(); pit.next()) {
            int typ = pit.currentSegment(crds);
            switch (typ) {
            case PathIterator.SEG_MOVETO :
                mvto[0] = crds[0];
                mvto[1] = crds[1];
                last[0] = crds[0];
                last[1] = crds[1];
            break;
            case PathIterator.SEG_LINETO : {
                lineSeg.setLine(last[0], last[1], crds[0], crds[1]);
                last[0] = crds[0];
                last[1] = crds[1];

                double _distSq = lineSeg.ptSegDistSq(pnt);
                if (_distSq < distSq)
                    distSq = _distSq;
            }
            break;
            case PathIterator.SEG_CLOSE : {
                lineSeg.setLine(last[0], last[1], mvto[0], mvto[1]);
                double _distSq = lineSeg.ptSegDistSq(pnt);
                if (_distSq < distSq)
                    distSq = _distSq;
            }
            break;
            default :
                throw new Error("Not a flattening path iterator");
            }
        }

        return Math.sqrt(distSq);
    }

    // =========================================================================
    // Projection ==============================================================
    // =========================================================================

   /**
    * Projects a point onto a line.
    * @param pnt the point
    * @param lin the line
    * @return the projected point
    */
    public static Point2D projection(Point2D pnt, Line2D lin, Point2D pntDst) {
        if (pntDst == null)
            pntDst = new Point2D.Double();
        final Vector2D linVec = new Vector2D();
        linVec.setLocation(lin.getX2()-lin.getX1(), lin.getY2()-lin.getY1());
        try {
            linVec.normalize();
        }
        catch (NullVectorException e) {
            pntDst.setLocation(lin.getP1());
            return pntDst;
        }
        final Vector2D pntVec = new Vector2D();
        pntVec.setLocation(lin.getP1(), pnt);
        double scale = linVec.getScalarProduct(pntVec);
        linVec.scaleBy(scale).add(lin.getP1());
        pntDst.setLocation(linVec);
        return pntDst;
    }

   /**
    * Projects a line segment onto a line.
    * @param seg the line segment
    * @param lin the line
    * @return the projected point
    */
    public static Line2D projection(Line2D seg, Line2D lin, Line2D linDst) {
        if (linDst == null)
            linDst = new Line2D.Double();
        final Point2D pnt1 = new Point2D.Double();
        projection(seg.getP1(), lin, pnt1);
        final Point2D pnt2 = new Point2D.Double();
        projection(seg.getP2(), lin, pnt2);
        linDst.setLine(pnt1, pnt2);
        return linDst;
    }

    // =========================================================================
    // Intersection ============================================================
    // =========================================================================

    // intersection(line segment, *) >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection point between two line segments.
    * @param seg the 1st line segment
    * @param seg2 the 2nd line segment
    * @param pntDst the intersection point
    * @return the intersection point
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Point2D intersection(Line2D seg, Line2D seg2, Point2D pntDst)
    throws NoIntersectionException {
        return intersection(seg, false, seg2, false, pntDst);
    }

   /**
    * Calculates the intersection point between a line segment and a
    * line(segment).
    * @param seg the 1st line segment
    * @param any the 2nd line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param pntDst the intersection point
    * @return the intersection point
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Point2D intersection(Line2D seg, Line2D any,
    boolean unbounded, Point2D pntDst) throws NoIntersectionException {
        return intersection(seg, false, any, unbounded, pntDst);
    }

   /**
    * Calculates the intersection points between a line segment and an ellipse.
    * @param seg the line segment
    * @param ell the ellipse
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D seg, Ellipse2D ell)
    throws NoIntersectionException {
        return intersection(seg, false, ell);
    }

   /**
    * Calculates the intersection points between a line segment and an arc.
    * @param seg the line segment
    * @param arc the arc
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D seg, Arc2D arc)
    throws NoIntersectionException {
        return intersection(seg, false, arc);
    }

   /**
    * Calculates the intersection points between a line segment and a quadratic
    * parametric curve segment.
    * @param seg the line segment
    * @param qad the quadratic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D seg, QuadCurve2D qad)
    throws NoIntersectionException {
        return intersection(seg, _toCubicCurve(qad, new CubicCurve2D.Double()));
    }

    /**
     * Calculates the intersection points between a line segment and a cubic
     * parametric curve segment.
     * @param seg the line segment
     * @param cub the cubic parametric curve segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(Line2D seg, CubicCurve2D cub)
    throws NoIntersectionException {
        return intersection(_toCubicCurve(seg, new CubicCurve2D.Double()), cub);
    }

    // Intersection(line[segment], *) >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection point between two line(segment)s.
    * @param any the 1st line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param any2 the 2nd line(segment)
    * @param unbounded2 <code>true</code> if <code>any2</code> is a line,
    *     <code>false</code> if <code>any2</code> is a segment
    * @param pntDst the intersection point
    * @return the intersection point
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Point2D intersection(Line2D any, boolean unbounded,
    Line2D any2, boolean unbounded2, Point2D pntDst)
    throws NoIntersectionException {

        // Create point object if necessary
        if (pntDst == null)
            pntDst = new Point2D.Double();

        // Bounds check
        if (! unbounded && ! unbounded2)
            _boundsCheck(any, any2);

        // Degenerated information?
        boolean isNull = isNull(any);
        boolean isNull2 = isNull(any2);
        if (isNull && isNull2) {
            Point2D p = new Point2D.Double(.5 * (any.getX1() + any.getX2()),
                    .5 * (any.getY1() + any.getY2()));
            Point2D p2 = new Point2D.Double(.5 * (any2.getX1() + any2.getX2()),
                    .5 * (any2.getY1() + any2.getY2()));
            if (equality(p, p2)) {
                pntDst.setLocation(.5 * (p.getX() + p2.getX()), 
                        .5 * (p.getY() + p2.getY()));
                return pntDst;
            }
        }
        else if (isNull) {
            Point2D p = new Point2D.Double(.5 * (any.getX1() + any.getX2()),
                    .5 * (any.getY1() + any.getY2()));
            if (containment(any2, unbounded2, p)) {
                pntDst.setLocation(p);
                return pntDst;
            }
            throw new NoIntersectionException();
        }
        else if (isNull2) {
            Point2D p2 = new Point2D.Double(.5 * (any2.getX1() + any2.getX2()),
                    .5 * (any2.getY1() + any2.getY2()));
            if (containment(any, unbounded, p2)) {
                pntDst.setLocation(p2);
                return pntDst;
            }
            throw new NoIntersectionException();
        }
        
        // Linear equation
        double vx1 = any.getX2() - any.getX1();
        double vy1 = any.getY2() - any.getY1();
        double a1  =   vy1;
        double b1  = - vx1;
        double c1  =   vx1 * any.getY1() - vy1 * any.getX1();

        double vx2 = any2.getX2() - any2.getX1();
        double vy2 = any2.getY2() - any2.getY1();
        double a2  =   vy2;
        double b2  = - vx2;
        double c2  =   vx2 * any2.getY1() - vy2 * any2.getX1();

        // Solve equation system
        double d   = a1 * b2 - a2 * b1;
        if (Math.abs(d) <= m_eps) {
            throw new NoIntersectionException("Lines are parallel");
        }
        double dx  = b1 * c2 - b2 * c1;
        double dy  = c1 * a2 - c2 * a1;
        pntDst.setLocation(dx/d, dy/d);

        // Do the line segments contain the intersection point?
        if (! unbounded && ! containment(any, pntDst)) {
            throw new NoIntersectionException(
                "1st line segment doesn't contain intersection point");
        }
        if (! unbounded2 && ! containment(any2, pntDst)) {
            throw new NoIntersectionException(
                "2nd line segment doesn't contain intersection point");
        }
        
        // Return the intersection point
        return pntDst;
    }

   /**
    * Calculates the intersection points between a line(segment) and an ellipse.
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param ell the ellipse
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D any, boolean unbounded,
    Ellipse2D ell) throws NoIntersectionException {
        // Case: arbitrary ellipse
        double df = Math.abs(ell.getWidth() - ell.getHeight());
        if (df > Geo2D.getEps())
            return intersection(((Shape)ell).getPathIterator(null), any,
                    unbounded);
        
        // Case: circle
        try {
            // Circle parameters
            double cx = ell.getCenterX();
            double cy = ell.getCenterY();
            double r = .5 * ell.getWidth();
    
            // Line for circle centered at the origin
            double l1x = any.getX1() - cx;
            double l1y = any.getY1() - cy;
            double l2x = any.getX2() - cx;
            double l2y = any.getY2() - cy;
            
            // Unit vector along line
            Vector2D u = new Vector2D(l2x - l1x, l2y - l1y);
            u.normalize();
            
            // Solve equation
            double r1u = l1x * u.x + l1y * u.y;
            double uu = u.x * u.x + u.y * u.y;
            double r1r1 = l1x * l1x + l1y * l1y;
            
            double radicand = r1u * r1u - r1r1 * uu + r * r * uu;
            if (radicand < - getEps())
                throw new NoIntersectionException();
            double sqrt = radicand <= getEps() ? 0 : Math.sqrt(radicand);
            double t1 = - r1u / uu;
            double t2 = sqrt / uu;
            double lambda1 = t1 + t2;
            double lambda2 = t1 - t2;
            
            // Intersection points
            final Point2D[] pnts = new Point2D[] {
                new Point2D.Double(l1x + lambda1 * u.x + cx,
                    l1y + lambda1 * u.y + cy),
                new Point2D.Double(l1x + lambda2 * u.x + cx,
                    l1y + lambda2 * u.y + cy)
            };
            int nPnts = Geo2D.equality(pnts[0], pnts[1]) ? 1 : 2;

            // All points contained in line segment ?
            if (! unbounded) {
                for (int i = 0; i < nPnts; i++) {
                    if (! Geo2D.containment(any, pnts[i]))
                        pnts[i] = null;
                }
            }
            
            // Return the iterator
            return new _PointIterator(pnts, nPnts);
        }
        catch (NullVectorException e) {
            throw new NoIntersectionException();
        }
    }

   /**
    * Calculates the intersection points between a line(segment) and an arc.
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param arc the arc
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D any, boolean unbounded,
    Arc2D arc) throws NoIntersectionException {

        // Case: arbitrary arc
        double df = Math.abs(arc.getWidth() - arc.getHeight());
        if (df > Geo2D.getEps())
            return intersection(((Shape)arc).getPathIterator(null), any,
                    unbounded);
        
        // Case: circular arc
        Ellipse2D cir = new Ellipse2D.Double();
        double cx = arc.getCenterX();
        double cy = arc.getCenterY();
        double r = .5 * arc.getWidth();
        cir.setFrameFromCenter(cx, cy, cx + r, cy + r);
        Iterator<Point2D> it = intersection(any, unbounded, cir);

        // Accept only points on arc
        Point2D[] pnts = new Point2D[2];
        int nPnts = 0;
        while (it.hasNext()) {
            Point2D p = it.next();
            if (containment(arc, p))
                pnts[nPnts++] = p;
        }
        if (nPnts == 0)
            throw new NoIntersectionException();
        return new _PointIterator(pnts, nPnts);
    }

   /**
    * Calculates the intersection points between a line(segment) and a quadratic
    * parametric curve segment.
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param qad the quadratic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D any, boolean unbounded,
    QuadCurve2D qad) throws NoIntersectionException {
        return intersection(any, unbounded, _toCubicCurve(qad,
            new CubicCurve2D.Double()));
    }

   /**
    * Calculates the intersection points between a line(segment) and a cubic
    * parametric curve segment.
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param cub the cubic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Line2D any, boolean unbounded,
    CubicCurve2D cub) throws NoIntersectionException {

        // Intersect line segment
        if (! unbounded)
            return intersection(_toCubicCurve(any, new CubicCurve2D.Double()),
                _toCubicCurve(cub, new CubicCurve2D.Double()));

        // Intersect line
        final Point2D[] pnts = new Point2D.Double[3];
        final int nPnts = _intersection(any, cub, pnts);

        // Provide results
        if (nPnts == 0)
            throw new NoIntersectionException();
        return new _PointIterator(pnts, nPnts);
    }

    private static int _intersection(Line2D lin, CubicCurve2D cub,
            Point2D[] pnts) {
        return __intersection(lin, cub, pnts, 0);
    }

    private static int __intersection(Line2D lin, CubicCurve2D cub,
        Point2D[] pnts, int nPnts) {

        // Maximum number of intersection points reached?
        if (nPnts >= 3)
            return nPnts;

        // Intersection possible?
        if (_relativePosition(lin, cub) != 0)
            return nPnts;

        // Cubic approximately straight?
        if (cub.getFlatnessSq() < m_epsSqr) {
            // Calculate intersection point
            Line2D lseg = new Line2D.Double(cub.getP1(), cub.getP2());
            Point2D p = new Point2D.Double();
            try {
                intersection(lseg, lin, /*unbounded=*/ true, p);

                // Store point if not already stored?
                for (int i = 0; i < nPnts; i++) {
                    if (equality(pnts[i], p))
                        return nPnts;
                }
                pnts[nPnts++] = p;
            }
            catch (NoIntersectionException e) {
            }
        }

        // Subdivide and intersect again
        else {
            CubicCurve2D left = new CubicCurve2D.Double();
            CubicCurve2D right = new CubicCurve2D.Double();
            cub.subdivide(left, right);

            nPnts = __intersection(lin, left, pnts, nPnts);
            nPnts = __intersection(lin, right, pnts, nPnts);
        }
        return nPnts;
    }

    /**
     * Calculates the relative position between a line and a cubic curve. If
     * the line is degenerated (coincident start and end point) then -1 is
     * returned if the line is outside the bounding box of the cubic curve.
     * @param lin the line
     * @param cub the cubic curve
     * @return the position of the curve relative to the line
     *         -1: curve is left of line
     *         +1: curve is right of line
     *          0: curve (eventually) intersects the line
     */
    private static int _relativePosition(Line2D lin, CubicCurve2D cub) {
        // Is the line degenerated?
        if (isNull(lin)) {
            Rectangle2D rct = cub.getBounds2D();
            double x = lin.getX1() - 0.5 * m_eps;
            double y = lin.getY1() - 0.5 * m_eps;
            return rct.intersects(x, y, m_eps, m_eps) ?
                0 : -1;
        }

        // Create a normalized vector perpendicular to the line
        final Vector2D vec1Left = new Vector2D(lin.getP1(), lin.getP2());
        vec1Left.left();
        vec1Left.normalize();

        // All four points/ control points must be left or right the line
        final Vector2D vec = new Vector2D();

        // P1
        vec.setLocation(lin.getP1(), cub.getP1());
        int ii = _relativePosition(vec1Left, vec);
        if (ii == 0)
            return 0;

        // P2
        vec.setLocation(lin.getP1(), cub.getP2());
        int i = _relativePosition(vec1Left, vec);
        if (i == 0 || i + ii == 0)
            return 0;

        // CtrlP1
        vec.setLocation(lin.getP1(), cub.getCtrlP1());
        i = _relativePosition(vec1Left, vec);
        if (i == 0 || i + ii == 0)
            return 0;

        // CtrlP2
        vec.setLocation(lin.getP1(), cub.getCtrlP2());
        i = _relativePosition(vec1Left, vec);
        if (i == 0 || i + ii == 0)
            return 0;

        return ii;
    }
    private static int _relativePosition(Vector2D vec1Left, Vector2D vec1Pnt) {
        double signedDist = vec1Left.getScalarProduct(vec1Pnt);
        if (signedDist < -m_eps)
            return -1;
        else if (signedDist > +m_eps)
            return +1;
        else
            return 0;
    }

    // Intersection(ellipse, *) >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection points between two ellipses.
    * @param ell the 1st ellipse
    * @param ell2 the 2nd ellipse
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Ellipse2D ell, Ellipse2D ell2)
    throws NoIntersectionException {
        // Intersect two circles?
        if (Math.abs(ell.getWidth() - ell.getHeight()) < m_eps &&
                Math.abs(ell2.getWidth() - ell2.getHeight()) < m_eps) {

            double c1x = ell.getCenterX();
            double c1y = ell.getCenterY();
            double c2x = ell2.getCenterX();
            double c2y = ell2.getCenterY();
            double m = Point2D.distance(c1x, c1y, c2x, c2y);
            double r1 = .5 * ell.getWidth();
            double r2 = .5 * ell2.getWidth();

            if (isNull(m) || isNull(r1) || isNull(r2))
                throw new NoIntersectionException();
            
            double u = .5 * m + (r1 * r1 - r2 * r2) / (m + m);
            double radicand = r1 * r1 - u * u;
            if (radicand < - getEps())
                throw new NoIntersectionException();
            double v = radicand > getEps() ? Math.sqrt(radicand) : 0;
            Vector2D vc1c2 = new Vector2D(c1x, c1y, c2x, c2y);
            vc1c2.normalize();
            Vector2D vPerp = new Vector2D(vc1c2).left();
            vPerp.scaleBy(v);
            vc1c2.scaleBy(u);
            return v < m_eps ?
                    new _PointIterator(new Point2D[] {
                        new Point2D.Double(c1x + vc1c2.x, c1y + vc1c2.y)}, 1)
                    :
                    new _PointIterator(new Point2D[] {
                        new Point2D.Double(c1x + vc1c2.x + vPerp.x, 
                                c1y + vc1c2.y + vPerp.y), 
                        new Point2D.Double(c1x + vc1c2.x - vPerp.x, 
                                c1y + vc1c2.y - vPerp.y)}, 2);
        }
        // General case: no circles
        return intersection(((Shape)ell).getPathIterator(null),
            ((Shape)ell2).getPathIterator(null));
    }

   /**
    * Calculates the intersection points between an ellipse and an arc.
    * @param ell the ellipse
    * @param arc the arc
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Ellipse2D ell, Arc2D arc)
    throws NoIntersectionException {
        // Intersect two circles?
        if (Math.abs(ell.getWidth() - ell.getHeight()) < m_eps &&
                Math.abs(arc.getWidth() - arc.getHeight()) < m_eps) {

            // Exact calculation with a full circle
            Ellipse2D cir = new Ellipse2D.Double(arc.getMinX(), arc.getMinY(), 
                    arc.getWidth(), arc.getHeight());
    
            // Intersection located on arc?
            Point2D[] pnts = new Point2D[2];
            int nPnts = 0;
            Iterator<Point2D> it = intersection(ell, cir);
            while (it.hasNext()) {
                Point2D p = it.next();
                if (containment(arc, p))
                    pnts[nPnts++] = p;
            }
            return new _PointIterator(pnts, nPnts);
        }        

        // General case
        return intersection(((Shape)ell).getPathIterator(null),
            ((Shape)arc).getPathIterator(null));
    }

   /**
    * Calculates the intersection points between an ellipse and a quadratic
    * parametric curve segment.
    * @param ell the ellipse
    * @param qad the quadratic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Ellipse2D ell, QuadCurve2D qad)
    throws NoIntersectionException {
        return intersection(((Shape)ell).getPathIterator(null),
            ((Shape)qad).getPathIterator(null));
    }

    /**
     * Calculates the intersection points between an ellipse and a cubic
     * parametric curve segment.
     * @param ell the ellipse
     * @param cub the cubic parametric curve segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(Ellipse2D ell, 
            CubicCurve2D cub) throws NoIntersectionException {
        return intersection(((Shape)ell).getPathIterator(null),
            ((Shape)cub).getPathIterator(null));
    }

    // Intersection(arc, *) >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection points between two arcs.
    * @param arc the 1st arc
    * @param arc2 the 2nd arc
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Arc2D arc, Arc2D arc2)
    throws NoIntersectionException {
        // Intersect two circles?
        if (Math.abs(arc.getWidth() - arc.getHeight()) < m_eps &&
                Math.abs(arc2.getWidth() - arc2.getHeight()) < m_eps) {

            Ellipse2D cir = new Ellipse2D.Double(arc.getMinX(), arc.getMinY(), 
                    arc.getWidth(), arc.getHeight());
            Ellipse2D cir2 = new Ellipse2D.Double(arc2.getMinX(), arc2.getMinY(),
                    arc2.getWidth(), arc2.getHeight());
    
            // Intersection located on Ellipses?
            Point2D[] pnts = new Point2D[2];
            int nPnts = 0;
            Iterator<Point2D> it = intersection(cir, cir2);
            while (it.hasNext()) {
                Point2D p = it.next();
                if (containment(arc, p) && containment(arc2,p))
                    pnts[nPnts++] = p;
            }
            return new _PointIterator(pnts, nPnts);
        }

        // General case
        return intersection(((Shape)arc).getPathIterator(null),
                ((Shape)arc2).getPathIterator(null));
    }

   /**
    * Calculates the intersection points between an arc and a quadratic
    * parametric curve segment.
    * @param arc the arc
    * @param qad the quadratic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(Arc2D arc, QuadCurve2D qad)
    throws NoIntersectionException {
        return intersection(((Shape)arc).getPathIterator(null),
            ((Shape)qad).getPathIterator(null));
    }

    /**
     * Calculates the intersection points between an arc and a cubic parametric
     * curve segment.
     * @param arc the arc
     * @param cub the cubic parametric curve segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(Arc2D arc, CubicCurve2D cub)
    throws NoIntersectionException {
        return intersection(((Shape)arc).getPathIterator(null),
            ((Shape)cub).getPathIterator(null));
    }

    // Intersection(quadratic parametric curve segment, *) >>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection points between two quadratic parametric curve
    * segments.
    * @param qad the 1st quadratic parametric curve segment
    * @param qad2 the 2nd quadratic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(QuadCurve2D qad, 
            QuadCurve2D qad2) throws NoIntersectionException {
        return intersection(_toCubicCurve(qad, new CubicCurve2D.Double()),
            _toCubicCurve(qad2, new CubicCurve2D.Double()));
    }

    /**
     * Calculates the intersection points between a quadratic parametric curve
     * segment and a cubic parametric curve segment.
     * @param qad the quadratic parametric curve segment
     * @param cub the cubic parametric curve segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(QuadCurve2D qad, 
            CubicCurve2D cub) throws NoIntersectionException {
        return intersection(_toCubicCurve(qad, new CubicCurve2D.Double()), cub);
    }

    // Intersection(cubic parametric curve segment, *) >>>>>>>>>>>>>>>>>>>>>>>>>

   /**
    * Calculates the intersection points between two cubic parametric curve
    * segment.
    * @param cub1 the 1st cubic parametric curve segment
    * @param cub2 the 2nd cubic parametric curve segment
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(CubicCurve2D cub1, 
            CubicCurve2D cub2) throws NoIntersectionException {
        // Intersect
        final Point2D[] pnts = new Point2D.Double[9];
        final int nPnts = _intersection(cub1, cub2, pnts);

        // Provide results
        if (nPnts == 0)
            throw new NoIntersectionException();
        return new _PointIterator(pnts, nPnts);
    }

    private static class _PointIterator implements Iterator<Point2D> {
        private Point2D[] m_pnts;
        private int m_index = 0;
        private _PointIterator(Point2D[] pnts, int _nPnts) {
            int nPnts = 0;
            for (int i = 0; i < _nPnts; i++) {
                if (pnts[i] != null)
                    nPnts++;
            }
            m_pnts = new Point2D.Double[nPnts];
            int j = 0;
            for (int i = 0; i < _nPnts; i++) {
                if (pnts[i] != null)
                    m_pnts[j++] = pnts[i];
            }
            m_index = 0;
        }
        public boolean hasNext() {
            return m_index < m_pnts.length;
        }
        public Point2D next() {
            if (m_index >= m_pnts.length)
                throw new NoSuchElementException();
            return m_pnts[m_index++];
        }
        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    
    private static int _intersection(CubicCurve2D cub1, CubicCurve2D cub2,
            Point2D[] pnts) {
        return __intersection(cub1, cub2, pnts, 0);
    }

    private static int __intersection(CubicCurve2D cub1, CubicCurve2D cub2,
    Point2D[] pnts, int nPnts) {

        // Maximum number of intersection points reached?
        if (nPnts >= 9)
            return nPnts;

        // Intersection possible?
        if (! _collision(cub1, cub2))
            return nPnts;

        // Both curves approximately straight?
        boolean isFlat1 = cub1.getFlatnessSq() < m_epsSqr;
        boolean isFlat2 = cub2.getFlatnessSq() < m_epsSqr;
        if (isFlat1 && isFlat2) {
            // Calculate intersection point
            Line2D l1 = new Line2D.Double(cub1.getP1(), cub1.getP2());
            Line2D l2 = new Line2D.Double(cub2.getP1(), cub2.getP2());
            Point2D p = new Point2D.Double();
            try {
                intersection(l1, l2, p);

                // Store point if not already stored?
                for (int i = 0; i < nPnts; i++) {
                    if (equality(pnts[i], p))
                        return nPnts;
                }
                pnts[nPnts++] = p;
            }
            catch (NoIntersectionException e) {
            }
        }

        // Subdivide and intersect again
        else if (isFlat1) {
            // Curve 2 left or right of line 1?
            Line2D l1 = new Line2D.Double(cub1.getP1(), cub1.getP2());
            if (_relativePosition(l1, cub2) != 0)
                return nPnts;

            // Subdivide curve 2
            CubicCurve2D left2 = new CubicCurve2D.Double();
            CubicCurve2D right2 = new CubicCurve2D.Double();
            cub2.subdivide(left2, right2);

            nPnts = __intersection(cub1, left2, pnts, nPnts);
            nPnts = __intersection(cub1, right2, pnts, nPnts);
        }
        else if (isFlat2) {
            // Curve 1 left or right of line 2?
            Line2D l2 = new Line2D.Double(cub2.getP1(), cub2.getP2());
            if (_relativePosition(l2, cub1) != 0)
                return nPnts;

            // Subdivide curve 1
            CubicCurve2D left1 = new CubicCurve2D.Double();
            CubicCurve2D right1 = new CubicCurve2D.Double();
            cub1.subdivide(left1, right1);

            nPnts = __intersection(left1, cub2, pnts, nPnts);
            nPnts = __intersection(right1, cub2, pnts, nPnts);
        }
        else {
            // Subdivide both curves
            CubicCurve2D left1 = new CubicCurve2D.Double();
            CubicCurve2D right1 = new CubicCurve2D.Double();
            cub1.subdivide(left1, right1);

            CubicCurve2D left2 = new CubicCurve2D.Double();
            CubicCurve2D right2 = new CubicCurve2D.Double();
            cub2.subdivide(left2, right2);

            nPnts = __intersection(left1, left2, pnts, nPnts);
            nPnts = __intersection(left1, right2, pnts, nPnts);
            nPnts = __intersection(right1, left2, pnts, nPnts);
            nPnts = __intersection(right1, right2, pnts, nPnts);
        }
        return nPnts;
    }

    // Intersection(path, *) >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    /**
     * Calculates the intersection points between a path and a line segment.
     * @param pit the path
     * @param seg the line segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(PathIterator pit, Line2D seg)
    throws NoIntersectionException {
        return intersection(pit, ((Shape)seg).getPathIterator(null));
    }

    /**
     * Calculates the intersection points between a path and a line(segment).
     * @param pit the path
     * @param any the line(segment)
     * @param unbounded <code>true</code> if <code>any</code> is a line,
     *     <code>false</code> if <code>any</code> is a segment
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
     public static Iterator<Point2D> intersection(PathIterator pit, Line2D any,
     boolean unbounded) throws NoIntersectionException {

         // Apply this algorithm to unbounded lines only
         if (! unbounded) {
             PathIterator pitl = ((Shape)any).getPathIterator(null);
             return intersection(pit, pitl);
         }

         // Create a cubic curve to save space
         CubicCurve2D cub1 = new CubicCurve2D.Double();

         // Perform intersection algorithm
         final Set<Point2D> pntSet = new HashSet<Point2D>();
         pntSet.clear();
         Shape seg1 = null;
         double[] mvto1 = new double[2];
         double[] last1 = new double[2];
         double[] crds1 = new double[6];
         for (; ! pit.isDone(); pit.next()) {
             int typ1 = pit.currentSegment(crds1);
             switch (typ1) {
             case PathIterator.SEG_MOVETO :
                 mvto1[0] = crds1[0];
                 mvto1[1] = crds1[1];
                 seg1 = null;
                 last1[0] = crds1[0];
                 last1[1] = crds1[1];
                 break;
             case PathIterator.SEG_LINETO :
                 seg1 = new Line2D.Double(last1[0], last1[1], crds1[0],
                     crds1[1]);
                 last1[0] = crds1[0];
                 last1[1] = crds1[1];
                 break;
             case PathIterator.SEG_QUADTO :
                 seg1 = new QuadCurve2D.Double(last1[0], last1[1], crds1[0],
                     crds1[1], crds1[2], crds1[3]);
                 last1[0] = crds1[2];
                 last1[1] = crds1[3];
                 break;
             case PathIterator.SEG_CUBICTO :
                 seg1 = new CubicCurve2D.Double(last1[0], last1[1], crds1[0],
                     crds1[1], crds1[2], crds1[3], crds1[4], crds1[5]);
                 last1[0] = crds1[4];
                 last1[1] = crds1[5];
                 break;
             case PathIterator.SEG_CLOSE :
                 seg1 = new Line2D.Double(last1[0], last1[1], mvto1[0],
                     mvto1[1]);
                 break;
             default :
                 throw new Error("Unexpected segment type");
             }
             if (seg1 == null)
                 continue;

             try {
                 if (seg1 instanceof Line2D) {
                     Point2D p = new Point2D.Double();
                     intersection((Line2D)seg1, /*unbounded=*/false, any,
                         /*unbounded=*/true, p);
                     _addPoint(pntSet, p);
                 }
                 else {
                     Iterator<Point2D> it = intersection(any, 
                             /*unbounded=*/true, _toCubicCurve(seg1, cub1));
                     while (it.hasNext()) {
                         Point2D p = it.next();
                         _addPoint(pntSet, p);
                     }
                 }
             }
             catch (NoIntersectionException e) {
             }
         }
         if (pntSet.isEmpty())
             throw new NoIntersectionException();
         return pntSet.iterator();
     }

    /**
     * Calculates the intersection points between a path and an ellipse.
     * @param pit the path
     * @param ell the ellipse
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(PathIterator pit, 
            Ellipse2D ell) throws NoIntersectionException {
        // Case: arbitrary ellipse
        double df = Math.abs(ell.getWidth() - ell.getHeight());
        if (df > Geo2D.getEps())
            return intersection(pit, ((Shape)ell).getPathIterator(null));

        // Perform intersection algorithm
        final Set<Point2D> pntSet = new HashSet<Point2D>();
        pntSet.clear();
        Shape seg1 = null;
        double[] mvto1 = new double[2];
        double[] last1 = new double[2];
        double[] crds1 = new double[6];
        for (; ! pit.isDone(); pit.next()) {
            int typ1 = pit.currentSegment(crds1);
            switch (typ1) {
            case PathIterator.SEG_MOVETO :
                mvto1[0] = crds1[0];
                mvto1[1] = crds1[1];
                seg1 = null;
                last1[0] = crds1[0];
                last1[1] = crds1[1];
                break;
            case PathIterator.SEG_LINETO :
                seg1 = new Line2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1]);
                last1[0] = crds1[0];
                last1[1] = crds1[1];
                break;
            case PathIterator.SEG_QUADTO :
                seg1 = new QuadCurve2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1], crds1[2], crds1[3]);
                last1[0] = crds1[2];
                last1[1] = crds1[3];
                break;
            case PathIterator.SEG_CUBICTO :
                seg1 = new CubicCurve2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1], crds1[2], crds1[3], crds1[4], crds1[5]);
                last1[0] = crds1[4];
                last1[1] = crds1[5];
                break;
            case PathIterator.SEG_CLOSE :
                seg1 = new Line2D.Double(last1[0], last1[1], mvto1[0],
                    mvto1[1]);
                break;
            default :
                throw new Error("Unexpected segment type");
            }
            if (seg1 == null)
                continue;

            try {
                Iterator<Point2D> it = null;

                if (seg1 instanceof Line2D)
                    it = intersection((Line2D)seg1, /*unbounded=*/false, ell);
                else if (seg1 instanceof QuadCurve2D)
                    it = intersection(ell, (QuadCurve2D)seg1);
                else if (seg1 instanceof CubicCurve2D)
                    it = intersection(ell, (CubicCurve2D)seg1);
                else
                    throw new InternalError("Unexpected segment type");

                while (it.hasNext()) {
                    Point2D p = it.next();
                    _addPoint(pntSet, p);
                }
            }
            catch (NoIntersectionException e) {
            }
        }
        if (pntSet.isEmpty())
            throw new NoIntersectionException();
        return pntSet.iterator();
    }

    /**
     * Calculates the intersection points between a path and an arc.
     * @param pit the path
     * @param arc the arc
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(PathIterator pit, Arc2D arc)
    throws NoIntersectionException {
        return intersection(pit, ((Shape)arc).getPathIterator(null));
    }

    /**
     * Calculates the intersection points between a path and a quadratic curve.
     * @param pit the path
     * @param qad the quadratic curve
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(PathIterator pit, 
            QuadCurve2D qad) throws NoIntersectionException {
        return intersection(pit, ((Shape)qad).getPathIterator(null));
    }

    /**
     * Calculates the intersection points between a path and a cubic curve.
     * @param pit the path
     * @param cub the cubic curve
     * @return the iterator for the intersection points
     * @exception NoIntersectionException if the geometries do not intersect
     */
    public static Iterator<Point2D> intersection(PathIterator pit, 
            CubicCurve2D cub) throws NoIntersectionException {
        return intersection(pit, ((Shape)cub).getPathIterator(null));
    }

   /**
    * Calculates the intersection points between two Paths.
    * segment.
    * @param pit1 the 1st path iterator
    * @param pit2 the 2nd path iterator
    * @return the iterator for the intersection points
    * @exception NoIntersectionException if the geometries do not intersect
    */
    public static Iterator<Point2D> intersection(PathIterator pit1, 
            PathIterator pit2) throws NoIntersectionException {

        // Second PathIterator must be rewindable
        RewindablePathIterator rwpit2 = new RewindablePathIterator(pit2);

        // Create two cubic curves to save space
        CubicCurve2D cub1 = new CubicCurve2D.Double();
        CubicCurve2D cub2 = new CubicCurve2D.Double();

        // Perform intersection algorithm
        final Set<Point2D> pntSet = new HashSet<Point2D>();
        pntSet.clear();
        Shape seg1 = null;
        double[] mvto1 = new double[2];
        double[] last1 = new double[2];
        double[] crds1 = new double[6];
        for (; ! pit1.isDone(); pit1.next()) {
            int typ1 = pit1.currentSegment(crds1);
            switch (typ1) {
            case PathIterator.SEG_MOVETO :
                mvto1[0] = crds1[0];
                mvto1[1] = crds1[1];
                seg1 = null;
                last1[0] = crds1[0];
                last1[1] = crds1[1];
            break;
            case PathIterator.SEG_LINETO :
                seg1 = new Line2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1]);
                last1[0] = crds1[0];
                last1[1] = crds1[1];
            break;
            case PathIterator.SEG_QUADTO :
                seg1 = new QuadCurve2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1], crds1[2], crds1[3]);
                last1[0] = crds1[2];
                last1[1] = crds1[3];
            break;
            case PathIterator.SEG_CUBICTO :
                seg1 = new CubicCurve2D.Double(last1[0], last1[1], crds1[0],
                    crds1[1], crds1[2], crds1[3], crds1[4], crds1[5]);
                last1[0] = crds1[4];
                last1[1] = crds1[5];
            break;
            case PathIterator.SEG_CLOSE :
                seg1 = new Line2D.Double(last1[0], last1[1], mvto1[0],
                    mvto1[1]);
            break;
            default :
                throw new Error("Unexpected segment type");
            }
            if (seg1 == null)
                continue;

            // Second curve
            Shape seg2 = null;
            double[] mvto2 = new double[2];
            double[] last2 = new double[2];
            double[] crds2 = new double[6];
            for (rwpit2.rewind(); ! rwpit2.isDone(); rwpit2.next()) {
                int typ2 = rwpit2.currentSegment(crds2);
                switch (typ2) {
                case PathIterator.SEG_MOVETO :
                    mvto2[0] = crds2[0];
                    mvto2[1] = crds2[1];
                    seg2 = null;
                    last2[0] = crds2[0];
                    last2[1] = crds2[1];
                break;
                case PathIterator.SEG_LINETO :
                    seg2 = new Line2D.Double(last2[0], last2[1], crds2[0],
                        crds2[1]);
                    last2[0] = crds2[0];
                    last2[1] = crds2[1];
                break;
                case PathIterator.SEG_QUADTO :
                    seg2 = new QuadCurve2D.Double(last2[0], last2[1], crds2[0],
                        crds2[1], crds2[2], crds2[3]);
                    last2[0] = crds2[2];
                    last2[1] = crds2[3];
                break;
                case PathIterator.SEG_CUBICTO :
                    seg2 = new CubicCurve2D.Double(last2[0], last2[1], crds2[0],
                        crds2[1], crds2[2], crds2[3], crds2[4], crds2[5]);
                    last2[0] = crds2[4];
                    last2[1] = crds2[5];
                break;
                case PathIterator.SEG_CLOSE :
                    seg2 = new Line2D.Double(last2[0], last2[1], mvto2[0],
                        mvto2[1]);
                break;
                default :
                    throw new Error("Unexpected segment type");
                }
                if (seg2 != null) {
                    try {
                        if (seg1 instanceof Line2D && seg2 instanceof Line2D) {
                            Point2D p = new Point2D.Double();
                            intersection((Line2D)seg1, (Line2D)seg2, p);
                            _addPoint(pntSet, p);
                        }
                        else {
                            Iterator<Point2D> it = intersection(
                                _toCubicCurve(seg1, cub1),
                                _toCubicCurve(seg2, cub2));
                            while (it.hasNext()) {
                                Point2D p = it.next();
                                _addPoint(pntSet, p);
                            }
                        }
                    }
                    catch (NoIntersectionException e) {
                    }
                }
            }
        }
        if (pntSet.isEmpty())
            throw new NoIntersectionException();
        return pntSet.iterator();
    }

    // Helpers >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

    private static void _boundsCheck(Shape shp1, Shape shp2)
    throws NoIntersectionException {
        if (! _collision(shp1, shp2))
            throw new NoIntersectionException(
                "Bounds check: Shapes don't intersect");
    }

    private static boolean _collision(Shape shp1, Shape shp2) {
        // Get shape bounds ...
        Rectangle2D bounds1 = shp1.getBounds2D();
        Rectangle2D bounds2 = shp2.getBounds2D();

        // ... and extend by eps
        bounds1.setFrame(bounds1.getX() - m_eps,
                         bounds1.getY() - m_eps,
                         bounds1.getWidth() + m_eps + m_eps,
                         bounds1.getHeight() + m_eps + m_eps);
        bounds2.setFrame(bounds2.getX() - m_eps,
                         bounds2.getY() - m_eps,
                         bounds2.getWidth() + m_eps + m_eps,
                         bounds2.getHeight() + m_eps + m_eps);
        return bounds1.intersects(bounds2);
    }

    private static CubicCurve2D _toCubicCurve(Shape seg, CubicCurve2D cub) {
        if (cub == null)
            cub = new CubicCurve2D.Double();
        if (seg instanceof Line2D) {
            Line2D src = (Line2D)seg;
            cub.setCurve(src.getX1(), src.getY1(), src.getX1(), src.getY1(),
                src.getX2(), src.getY2(), src.getX2(), src.getY2());
        }
        else if (seg instanceof Ellipse2D) {
            throw new InternalError("Can't convert Ellipse2D to CubicCurve2D");
        }
        else if (seg instanceof Arc2D) {
            throw new InternalError("Can't convert Arc2D to CubicCurve2D");
        }
        else if (seg instanceof QuadCurve2D) {
            QuadCurve2D src = (QuadCurve2D)seg;
            cub.setCurve(src.getX1(), src.getY1(), src.getCtrlX(),
                src.getCtrlY(), src.getCtrlX(), src.getCtrlY(), src.getX2(),
                src.getY2());
        }
        else {
            CubicCurve2D src = (CubicCurve2D)seg;
            cub.setCurve(src.getX1(), src.getY1(), src.getCtrlX1(),
                src.getCtrlY1(), src.getCtrlX2(), src.getCtrlY2(), src.getX2(),
                src.getY2());
        }
        return cub;
    }

    private static boolean _addPoint(Set<Point2D> set, Point2D pnt) {
        Iterator<Point2D> it = set.iterator();
        while (it.hasNext()) {
            Point2D _pnt = it.next();
            if (equality(pnt, _pnt))
                return false;
        }
        set.add(pnt);
        return true;
    }

    // =========================================================================
    // Equality ================================================================
    // =========================================================================

   /**
    * Determines whether or not two double values are equal. Two instances of
    * <code>Point2D</code> are equal if their distance is less or equal to
    * <code>HtGeo.getEps()</code>.
    * @param d1 the 1st double
    * @param d2 the 2nd double
    * @return <code>true</code> if doubles are equal, <code>false</code> if
    *     double are not equal
    */
    public static boolean equality(double d1, double d2) {
        return m_eps > Math.abs(d2 - d1);
    }

    /**
     * Determines whether or not two points are equal. Two instances of
     * <code>Point2D</code> are equal if their distance is less or equal to
     * <code>HtGeo.getEps()</code>.
     * @param pnt1 the 1st point
     * @param pnt2 the 2nd point
     * @return <code>true</code> if points are equal, <code>false</code> if
     *     points are not equal
     */
     public static boolean equality(Point2D pnt1, Point2D pnt2) {
         return equality(pnt1.getX(), pnt1.getY(), pnt2.getX(), pnt2.getY());
     }

     /**
      * Determines whether or not two points are equal. Two instances of
      * <code>Point2D</code> are equal if their distance is less or equal to
      * <code>HtGeo.getEps()</code>.
      * @param x1 x1
      * @param y1 y1
      * @param x2 x2
      * @param y2 y2
      * @return <code>true</code> if points are equal, <code>false</code> if
      *     points are not equal
      */
      public static boolean equality(double x1, double y1, double x2,
              double y2) {
          double dx = x2 - x1;
          double dy = y2 - y1;
          return m_epsSqr > dx * dx + dy * dy;
      }

   /**
    * Determines whether or not two vectors are equal. Two instances of
    * <code>Vector2D</code> are equal if their distance is less or equal to
    * <code>HtGeo.getEps()</code>.
    * @param vec1 the 1st vector
    * @param vec2 the 2nd vector
    * @return <code>true</code> if vectors are equal, <code>false</code> if
    *     vectors are not equal
    */
    public static boolean equality(Vector2D vec1, Vector2D vec2) {
        double dx = vec2.getX() - vec1.getX();
        double dy = vec2.getY() - vec1.getY();
        return m_epsSqr > dx * dx + dy * dy;
    }

   /**
    * Determines whether or not two line segments are equal. Two line segments
    * of type <code>Line2D</code> are equal if their endpoints are coincident
    * and if they have the same direction.
    * @param seg1 the 1st line segment
    * @param seg2 the 2nd line segment
    * @return <code>true</code> if line segments are equal, <code>false</code>
    *     if line segments are not equal
    */
    public static boolean equality(Line2D seg1, Line2D seg2) {
        return
            equality(seg1.getP1(), seg2.getP1()) &&
            equality(seg1.getP2(), seg2.getP2());
    }

   /**
    * Determines whether or not two line(segment)s are equal. Two line segments
    * of type <code>Line2D</code> are equal if their endpoints are coincident,
    * two lines are equal if one line contains start and endpoint of the other
    * line.
    * @param any1 the 1st line(segment)
    * @param any2 the 2nd line(segment)
    * @param unbounded <code>true</code> if <code>any1</code> and
    *     <code>any2</code> are lines, <code>false</code> if <code>any1</code>
    *     and <code>any2</code> are line segments
    * @return <code>true</code> if line segments are equal, <code>false</code>
    *     if line segments are not equal
    */
    public static boolean equality(Line2D any1, Line2D any2, boolean unbounded) {
        if (unbounded) {
            boolean isNull1 = isNull(any1);
            boolean isNull2 = isNull(any2);
            if (isNull1 && isNull2) {
                return equality(any1.getP1(), any2.getP1());
            }
            else if (! isNull1 && ! isNull2) {
                Line2D longLine = any1;
                Line2D shortLine = any2;
                if (length(any2) > length(any1)) {
                    longLine = any2;
                    shortLine = any1;
                }
                if (containment(longLine, unbounded, shortLine.getP1()) &&
                    containment(longLine, unbounded, shortLine.getP2())) {
                    return true;
                }
            }
            return false;
        }
        return equality(any1, any2);
    }

   /**
    * Determines whether or not two ellipses are equal.
    * if their bounding boxes are coincident.
    * @param ell1 the 1st ellipse
    * @param ell2 the 2nd ellipse
    * @return <code>true</code> if ellipses are equal, <code>false</code>
    *     if ellipses are not equal
    */
    public static boolean equality(Ellipse2D ell1, Ellipse2D ell2) {
        final Point2D pnt1 = new Point2D.Double();
        final Point2D pnt2 = new Point2D.Double();
        pnt1.setLocation(ell1.getX(), ell1.getY());
        pnt2.setLocation(ell2.getX(), ell2.getY());
        return
            equality(pnt1, pnt2) &&
            equality(ell1.getWidth(), ell2.getWidth()) &&
            equality(ell1.getHeight(), ell2.getHeight());
    }

   /**
    * Determines whether or not two arcs are equal.
    * @param arc1 the 1st arc
    * @param arc2 the 2nd arc
    * @return <code>true</code> if arcs are equal, <code>false</code> if arcs
    *     are not equal
    */
    public static boolean equality(Arc2D arc1, Arc2D arc2) {
        final Point2D pnt1 = new Point2D.Double();
        final Point2D pnt2 = new Point2D.Double();
        pnt1.setLocation(arc1.getX(), arc1.getY());
        pnt2.setLocation(arc2.getX(), arc2.getY());
        return
            equality(pnt1, pnt2) &&
            equality(arc1.getWidth(), arc2.getWidth()) &&
            equality(arc1.getHeight(), arc2.getHeight()) &&
            equality(arc1.getAngleStart(), arc2.getAngleStart()) &&
            equality(arc1.getAngleExtent(), arc2.getAngleExtent());
    }

   /**
    * Determines whether or not two quadratic parametric curve segments are
    * equal.
    * @param qad1 the 1st quadratic parametric curve segment
    * @param qad2 the 2nd quadratic parametric curve segment
    * @return <code>true</code> if segments are equal, <code>false</code>
    *     if segments are not equal
    */
    public static boolean equality(QuadCurve2D qad1, QuadCurve2D qad2) {
        return
            equality(qad1.getP1(), qad2.getP1()) &&
            equality(qad1.getP2(), qad2.getP2()) &&
            equality(qad1.getCtrlPt(), qad2.getCtrlPt());
    }

   /**
    * Determines whether or not two cubic parametric curve segments are equal.
    * @param cub1 the 1st cubic parametric curve segment
    * @param cub2 the 2nd cubic parametric curve segment
    * @return <code>true</code> if segments are equal, <code>false</code>
    *     if segments are not equal
    */
    public static boolean equality(CubicCurve2D cub1, CubicCurve2D cub2) {
        return
            equality(cub1.getP1(), cub2.getP1()) &&
            equality(cub1.getP2(), cub2.getP2()) &&
            equality(cub1.getCtrlP1(), cub2.getCtrlP1()) &&
            equality(cub1.getCtrlP2(), cub2.getCtrlP2());
    }

    // =========================================================================
    // Parallelism =============================================================
    // =========================================================================

    /**
    * Determines whether or not two line(segment)s are parallel or not.
    * @param any1 the 1st line (segment)
    * @param any2 the 2nd line (segment)
    * @return <code>true</code> if lines are parallel, <code>false</code>
    *     if lines are not parallel
    */
    public static boolean parallelism(Line2D any1, Line2D any2) {
        // Compare translated line any2
        Line2D _any2 = new Line2D.Double(
            any1.getX1(),
            any1.getY1(),
            any1.getX1() + any2.getX2() - any2.getX1(),
            any1.getY1() + any2.getY2() - any2.getY1());
        return equality(any1, _any2, /*unbounded=*/ true);
    }

    // =========================================================================
    // Perpendicularity ========================================================
    // =========================================================================

    /**
    * Determines whether or not two vectors are perpendicular or not.
    * @param vec1 the 1st vector
    * @param vec2 the 2nd vector
    * @return <code>true</code> if vectors are perpendicular,
    *     <code>false</code> if vectors are not perpendicular
    */
    public static boolean perpendicularity(Vector2D vec1, Vector2D vec2) {
        return Math.abs(vec1.getAngleSmallest(vec2) - Math.PI) < m_eps;
    }

    /**
    * Determines whether or not two line(segment)s are perpendicular or not.
    * @param any1 the 1st line (segment)
    * @param any2 the 2nd line (segment)
    * @return <code>true</code> if lines are perpendicular,
    *     <code>false</code> if lines are not perpendicular
    */
    public static boolean perpendicularity(Line2D any1, Line2D any2) {
        Vector2D vec1 = new Vector2D(any1.getX2() - any1.getX1(),
            any1.getY2() - any1.getY1());
        Vector2D vec2 = new Vector2D(any2.getX2() - any2.getX1(),
            any2.getY2() - any2.getY1());
        return perpendicularity(vec1, vec2);
    }

    // =========================================================================
    // Containment =============================================================
    // =========================================================================

   /**
    * Determines whether a line segment contains a point.
    * @param seg the line segment
    * @param pnt the point
    * @return <code>true</code> if the segment contains the point,
    *     <code>false</code> if not
    */
    public static boolean containment(Line2D seg, Point2D pnt) {
        return seg.ptSegDistSq(pnt) <= m_epsSqr;
    }

   /**
    * Determines whether a line(segment) contains a point.
    * @param any the line(segment)
    * @param unbounded <code>true</code> if <code>any</code> is a line,
    *     <code>false</code> if <code>any</code> is a segment
    * @param pnt the point
    * @return <code>true</code> if the line(segment) contains the point,
    *     <code>false</code> if not
    */
    public static boolean containment(Line2D any, boolean unbounded,
    Point2D pnt) {
        return unbounded ?
            any.ptLineDistSq(pnt) <= m_epsSqr:
            any.ptSegDistSq(pnt) <= m_epsSqr;
    }

   /**
    * Determines whether an ellipse contains a point.
    * @param ell the ellipse
    * @param pnt the point
    * @return <code>true</code> if the ellipse contains the point,
    *     <code>false</code> if not
    */
    public static boolean containment(Ellipse2D ell, Point2D pnt) {
        return distance(pnt, ell) <= m_eps;
    }

   /**
    * Determines whether an arc contains a point.
    * @param arc the arc
    * @param pnt the point
    * @return <code>true</code> if the arc contains the point,
    *     <code>false</code> if not
    */
    public static boolean containment(Arc2D arc, Point2D pnt) {
        return distance(pnt, arc) <= m_eps;
    }

   /**
    * Determines whether a quadratic parametric curve segment contains a point.
    * @param qad the quadratic parametric curve segment
    * @param pnt the point
    * @return <code>true</code> if the quadratic parametric curve segment
    *     contains the point, <code>false</code> if not
    */
    public static boolean containment(QuadCurve2D qad, Point2D pnt) {
        return distance(pnt, qad) <= m_eps;
    }

   /**
    * Determines whether a cubic parametric curve segment contains a point.
    * @param cub the cubic parametric curve segment
    * @param pnt the point
    * @return <code>true</code> if the cubic parametric curve segment
    *     contains the point, <code>false</code> if not
    */
    public static boolean containment(CubicCurve2D cub, Point2D pnt) {
        return distance(pnt, cub) <= m_eps;
    }

    // =========================================================================
    // Null property ===========================================================
    // =========================================================================

    /**
     * Determines whether a distance is null.
     * 
     * @param dist the distance
     * @return <code>true</code> if the vector null, <code>false</code> if not
     */
    public static boolean isNull(double dist) {
        return dist * dist <= m_epsSqr;
    }

    /**
     * Determines whether a vector is null. A vector with coordinates (0, 0) is
     * null.
     * @param vec the vector
     * @return <code>true</code> if the vector null, <code>false</code> if not
     */
    public static boolean isNull(Vector2D vec) {
        return m_epsSqr > vec.getX() * vec.getX() + vec.getY() * vec.getY();
    }

    /**
     * Sets a vector to null.
     * @param vec the vector
     */
    public static void setNull(Vector2D vec) {
        vec.x = 0;
        vec.y = 0;
    }

   /**
    * Determines whether a line segment is null. A line segment with length 0
    * is null,
    * @param seg the line segment
    * @return <code>true</code> if the line segment is null, <code>false</code>
    * if not
    */
    public static boolean isNull(Line2D seg) {
        double dx = seg.getX2() - seg.getX1();
        double dy = seg.getY2() - seg.getY1();
        return m_epsSqr > dx * dx + dy * dy;
    }

    /**
     * Sets a line segment to null
     * @param seg the line segment
     */
    public static void setNull(Line2D seg) {
        double x = seg.getX1();
        double y = seg.getY1();
        seg.setLine(x, y, x, y);
    }

   /**
    * Determines whether an ellipse is null. An ellipse with radius 0 is null.
    * @param ell the ellipse
    * @return <code>true</code> if the ellipse is null, <code>false</code>
    * if not
    */
    public static boolean isNull(Ellipse2D ell) {
        return m_eps > ell.getWidth() || m_eps > ell.getHeight();
    }

    /**
     * Sets an ellipse to null.
     * @param ell the ellipse
     */
    public static void setNull(Ellipse2D ell) {
        double x = ell.getCenterX();
        double y = ell.getCenterY();
        ell.setFrameFromCenter(x, y, x, y);
    }

   /**
    * Determines whether an arc is null. An arc with radius 0 is null.
    * @param arc the arc
    * @return <code>true</code> if the arc is null, <code>false</code> if not
    */
    public static boolean isNull(Arc2D arc) {
        return m_eps > arc.getWidth() || m_eps > arc.getHeight();
    }

    /**
     * Sets an arc to null.
     * @param arc the arc
     */
    public static void setNull(Arc2D arc) {
        double x = arc.getCenterX();
        double y = arc.getCenterY();
        arc.setFrameFromCenter(x, y, x, y);
    }

   /**
    * Determines whether a quadratic parametric curve is null. A
    * quadratic parametric curve is null if start and end point are
    * coincident.
    * @param qad the quadratic parametric curve segment
    * @return <code>true</code> if the segment is null, <code>false</code> if
    * not
    */
    public static boolean isNull(QuadCurve2D qad) {
        return equality(qad.getP1(), qad.getP2());
    }

    /**
     * Sets a quadratic parametric curve to null.
     * @param qad the quadratic parametric curve
     */
    public static void setNull(QuadCurve2D qad) {
        double x = qad.getX1();
        double y = qad.getX1();
        qad.setCurve(x, y, x, y, x, y);
    }

   /**
    * Determines whether a cubic parametric curve is null. A cubic
    * parametric curve is null if start and end point are coincident.
    * @param cub the cubic parametric curve segment
    * @return <code>true</code> if the segment is null, <code>false</code> if
    * not
    */
    public static boolean isNull(CubicCurve2D cub) {
        return equality(cub.getP1(), cub.getP2());
    }

    /**
     * Sets a cubic parametric curve to null.
     * @param cub the cubic parametric curve
     */
    public static void setNull(CubicCurve2D cub) {
        double x = cub.getX1();
        double y = cub.getX1();
        cub.setCurve(x, y, x, y, x, y, x, y);
    }

    // =========================================================================
    // Tangent calculations ====================================================
    // =========================================================================
    
    /**
     * Calculates the 2 tangents through a point and a circle. If the point is
     * located inside the circle then a tangent cannot be calculated. If the
     * point is located outside the circle then two tangents are calculated.
     * If the point is located on the circle then a single tangent is returned.
     * 
     * @param pnt a point on the tangent
     * @param ell the circle
     * @param tangentPoints an array at least of size 2 to store the calculated
     *      tangent points
     * @return the number of tangent points stored in tangentPoints. If pnt is
     *      located on the circle then the tangent point is coincident.
     */
    public static int calculateTangentLines(Point2D pnt, Ellipse2D ell, 
            Point2D[] tangentPoints) {

        // Is it a circle?
        if (Math.abs(ell.getWidth() - ell.getHeight()) > m_eps)
            throw new IllegalArgumentException("Not a circle");

        // Point must be located outside or on the circle
        double dst = Point2D.distance(ell.getCenterX(), ell.getCenterY(),
                pnt.getX(), pnt.getY());
        double rad = .5 * ell.getWidth(); 

        
        if (dst - rad < -m_eps) { // Point located inside the circle
            return 0;
        }
        else if (Math.abs(dst - rad) < m_eps) { // Point located on the circle
            if (tangentPoints[0] == null)
                tangentPoints[0] = new Point2D.Double();
            tangentPoints[0].setLocation(pnt);
            return 1;
        }
            
        // Point located outside the circle: apply Thales' Theorem
        double cx = .5 * (pnt.getX() + ell.getCenterX());
        double cy = .5 * (pnt.getY() + ell.getCenterY());
        double r = .5 * Point2D.distance(pnt.getX(), pnt.getY(), 
                ell.getCenterX(), ell.getCenterY());
        Ellipse2D thales = new Ellipse2D.Double();
        thales.setFrameFromCenter(cx, cy, cx + r, cy + r);
        try {
            Iterator<Point2D> it = Geo2D.intersection(ell, thales);
            
            if (tangentPoints[0] == null)
                tangentPoints[0] = new Point2D.Double();
            tangentPoints[0].setLocation(it.next());

            if (tangentPoints[1] == null)
                tangentPoints[1] = new Point2D.Double();
            tangentPoints[1].setLocation(it.next());
            return 2;
        }
        catch (NoIntersectionException e) {
            return 0;
        }
        catch (NoSuchElementException e) {
            return 0;
        }
    }

    /**
     * Calculates up to 4 tangents of 2 given circles. The tangent start and end
     * points are located on the circles. If the circles are tangent to each 
     * other then the tangent start and end point are coincident.
     * 
     * @param ell1 the first circle
     * @param ell2 the second circle
     * @param tangents an array at least of size 4 to store the calculated
     *      tangent lines
     * @return the number of tangents stored in parameter tangents (0 ... 4)
     */
    public static int calculateTangentLines(Ellipse2D ell1, Ellipse2D ell2, 
            Line2D[] tangents) {
        
        // Circle parameters?
        if (Math.abs(ell1.getWidth() - ell1.getHeight()) > m_eps ||
                Math.abs(ell2.getWidth() - ell2.getHeight()) > m_eps)
            throw new IllegalArgumentException("Not a circle");

        // Preset parameters
        double r1 = .5 * ell1.getWidth();
        double c1x = ell1.getCenterX();
        double c1y = ell1.getCenterY();

        double r2 = .5 * ell2.getWidth();
        double c2x = ell2.getCenterX();
        double c2y = ell2.getCenterY();

        if (r2 > r1) {
            r2 = .5 * ell1.getWidth();
            c2x = ell1.getCenterX();
            c2y = ell1.getCenterY();

            r1 = .5 * ell2.getWidth();
            c1x = ell2.getCenterX();
            c1y = ell2.getCenterY();
        }

        Point2D c1 = new Point2D.Double(c1x, c1y);
        Point2D c2 = new Point2D.Double(c2x, c2y);

        // Outer tangents
        double r = r1 - r2;
        Ellipse2D ell = new Ellipse2D.Double();
        ell.setFrameFromCenter(c1x, c1y, c1x + r, c1y + r);
        Point2D[] tangentPoints = new Point2D[2];
        int n = calculateTangentLines(c2, ell, tangentPoints);

        Vector2D v = new Vector2D();
        if (n > 0) {
            v.setLocation(c1, tangentPoints[0]);
            v.normalize();
            v.scaleBy(r2);
            if (tangents[0] == null)
                tangents[0] = new Line2D.Double();
            tangents[0].setLine(tangentPoints[0].getX() + v.x,
                    tangentPoints[0].getY() + v.y, c2x + v.x, c2y + v.y);
        }
        if (n > 1) {
            v.setLocation(c1, tangentPoints[1]);
            v.normalize();
            v.scaleBy(r2);
            if (tangents[1] == null)
                tangents[1] = new Line2D.Double();
            tangents[1].setLine(tangentPoints[1].getX() + v.x,
                    tangentPoints[1].getY() + v.y, c2x + v.x, c2y + v.y);
        }
        
        // Inner tangents
        r = r1 + r2;
        ell.setFrameFromCenter(c1x, c1y, c1x + r, c1y + r);
        int m = calculateTangentLines(c2, ell, tangentPoints);

        if (m > 0) {
            v.setLocation(c1, tangentPoints[0]);
            v.normalize();
            v.scaleBy(r2);
            if (tangents[n] == null)
                tangents[n] = new Line2D.Double();
            tangents[n].setLine(tangentPoints[0].getX() - v.x,
                    tangentPoints[0].getY() - v.y, c2x - v.x, c2y - v.y);
        }
        if (m > 1) {
            v.setLocation(c1, tangentPoints[1]);
            v.normalize();
            v.scaleBy(r2);
            if (tangents[n+1] == null)
                tangents[n+1] = new Line2D.Double();
            tangents[n+1].setLine(tangentPoints[1].getX() - v.x,
                    tangentPoints[1].getY() - v.y, c2x - v.x, c2y - v.y);
        }
        
        // Return number of tangents
        return n + m;
    }

    /**
     * Calculates up to 4 tangent circles of 2 given lines.
     * 
     * @param line1 the first line
     * @param line2 the second line
     * @param radius the radius of the tangent circles
     * @param tangentCircleCenters an array at least of size 4 to store the
     *      tangent circle centers
     * @return the number of tangent circles (0 ... 4)
     */
    public static int calculateTangentCircles(Line2D line1, Line2D line2, 
            double radius, Point2D[] tangentCircleCenters) {

        // Check argument values
        if (radius < getEps())
            throw new IllegalArgumentException();
        
        try {
            Point2D p = intersection(line1, true, line2, true, 
                    new Point2D.Double());
            Vector2D u = new Vector2D(line1.getP1(), line1.getP2()).normalize();
            Vector2D v = new Vector2D(line2.getP1(), line2.getP2()).normalize();
            Vector2D u_ = new Vector2D(-u.x, -u.y);
            Vector2D v_ = new Vector2D(-v.x, -v.y);
            
            if (tangentCircleCenters[0] == null)
                tangentCircleCenters[0] = new Point2D.Double();
            _calculateCenter(p, u, v, radius, tangentCircleCenters[0]);
            
            if (tangentCircleCenters[1] == null)
                tangentCircleCenters[1] = new Point2D.Double();
            _calculateCenter(p, u, v_, radius, tangentCircleCenters[1]);
            
            if (tangentCircleCenters[2] == null)
                tangentCircleCenters[2] = new Point2D.Double();
            _calculateCenter(p, u_, v, radius, tangentCircleCenters[2]);
            
            if (tangentCircleCenters[3] == null)
                tangentCircleCenters[3] = new Point2D.Double();
            _calculateCenter(p, u_, v_, radius, tangentCircleCenters[3]);
            
            return 4;
        }
        catch (NoIntersectionException e) {
            return 0;
        }
        catch (NullVectorException e) {
            return 0;
        }
    }
    private static Point2D _calculateCenter(Point2D p, Vector2D u, Vector2D v,
            double radius, Point2D ctr) {

        Vector2D c = new Vector2D(u.x + v.x, u.y + v.y).normalize();
        double angle = u.getAngleSmallest(v);
        double t = radius / Math.sin(.5 * angle);
        c.scaleBy(t);
        
        if (ctr == null)
            ctr = new Point2D.Double();
        ctr.setLocation(p.getX() + c.x, p.getY() + c.y);
        return ctr;
    }
    
    /**
     * Calculates up to 8 tangent circles of a given line and a given circle.
     * 
     * @param line the line
     * @param cir the circle
     * @param r the radius of the tangent circles
     * @param tangentCircleCenters an array at least of size 8 to store the
     *      tangent circle centers
     * @return the number of tangent circles (0 ... 8)
     */
    public static int calculateTangentCircles(Line2D line, Ellipse2D cir, 
            double r, Point2D[] tangentCircleCenters) {

        // Check argument values
        if (isNull(r))
            throw new IllegalArgumentException();

        // Null geometries?
        if (isNull(line) || isNull(cir))
            return 0;

        // Line parameters
        Vector2D nrm = new Vector2D(line.getP1(), line.getP2());
        nrm.left();
        nrm.normalize();
        nrm.scaleBy(r);
        
        // Circle parameters
        double r1 = .5 * cir.getWidth();
        double x1 = cir.getCenterX();
        double y1 = cir.getCenterY();
        
        // Intersection problem
        int nCircles = 0;
        Line2D l = new Line2D.Double();
        Ellipse2D c = new Ellipse2D.Double();
        for (int i = 0; i < 2; i++) {
            l.setLine(line.getX1() + nrm.x, line.getY1() + nrm.y, 
                    line.getX2() + nrm.x, line.getY2() + nrm.y);

            // Large circle
            try {
                c.setFrameFromCenter(x1, y1, x1 + r1 + r, y1 + r1 + r);
                Iterator<Point2D> it = intersection(l, true, c);
                while (it.hasNext())
                    tangentCircleCenters[nCircles++] = it.next();
            }
            catch (NoIntersectionException e) {
            }

            // Small circle
            try {
                c.setFrameFromCenter(x1, y1, x1 + r1 - r, y1 + r1 - r);
                Iterator<Point2D> it = intersection(l, true, c);
                while (it.hasNext())
                    tangentCircleCenters[nCircles++] = it.next();
            }
            catch (NoIntersectionException e) {
            }

            // Invert normal vector
            nrm.scaleBy(-1);
        }
            
        // Return the number of circles
        return nCircles;
    }    
    
    /**
     * Calculates up to 8 tangent circles of a given line and a given circle.
     * 
     * @param cir1 the first circle
     * @param cir2 the second circle
     * @param r the radius of the tangent circles
     * @param tangentCircleCenters an array at least of size 8 to store the
     *      tangent circle centers
     * @return the number of tangent circles (0 ... 8)
     */
    public static int calculateTangentCircles(Ellipse2D cir1, Ellipse2D cir2, 
            double r, Point2D[] tangentCircleCenters) {

        // Check argument values
        if (r < getEps())
            throw new IllegalArgumentException();

        // Calculate
        int nCircles = 0;
        
        double r1 = .5 * cir1.getWidth();
        double x1 = cir1.getCenterX();
        double y1 = cir1.getCenterY();
        
        double r2 = .5 * cir2.getWidth();
        double x2 = cir2.getCenterX();
        double y2 = cir2.getCenterY();

        // Swap coordinates if necessary
        boolean swapped = false;
        if (Math.abs(x2 - x1) < Geo2D.getEps()) {
            swapped = true;
            double t = x1;
            x1 = y1;
            y1 = t;
            t = x2;
            x2 = y2;
            y2 = t;
        }
        if (Math.abs(x2 - x1) < Geo2D.getEps())
            return 0;
        
        double _a = (y2 - y1) / (x2 - x1);
        double[] _b = new double[4];
        _b[0] = .5 * ((r1 + r) * (r1 + r) - (r2 + r) * (r2 + r) - x1 * x1 + 
                x2 * x2 - y1 * y1 + y2 * y2) / (x2 - x1); 
        _b[1] = .5 * ((r1 + r) * (r1 + r) - (r2 - r) * (r2 - r) - x1 * x1 + 
                x2 * x2 - y1 * y1 + y2 * y2) / (x2 - x1); 
        _b[2] = .5 * ((r1 - r) * (r1 - r) - (r2 + r) * (r2 + r) - x1 * x1 + 
                x2 * x2 - y1 * y1 + y2 * y2) / (x2 - x1); 
        _b[3] = .5 * ((r1 - r) * (r1 - r) - (r2 - r) * (r2 - r) - x1 * x1 + 
                x2 * x2 - y1 * y1 + y2 * y2) / (x2 - x1); 

        double a = 1. + _a * _a;
        double[] b = new double[4];
        for (int i = 0; i < 4; i++)
            b[i] = 2. * (_a * x1 - _a * _b[i] - y1);
        double[] c = new double[4];
        c[0] = _b[0] * _b[0] - 2. * _b[0] * x1 + x1 * x1 + y1 * y1 - 
                (r1 + r) * (r1 + r);
        c[1] = _b[1] * _b[1] - 2. * _b[1] * x1 + x1 * x1 + y1 * y1 -
                (r1 + r) * (r1 + r);
        c[2] = _b[2] * _b[2] - 2. * _b[2] * x1 + x1 * x1 + y1 * y1 -
                (r1 - r) * (r1 - r);
        c[3] = _b[3] * _b[3] - 2. * _b[3] * x1 + x1 * x1 + y1 * y1 -
                (r1 - r) * (r1 - r);

        for (int i = 0; i < 4; i++) {
            double discriminant = .25 * b[i] * b[i] - a * c[i];
            if (discriminant > - Geo2D.getEps()) {
                double sqrt = discriminant < Geo2D.getEps() ?
                        0 : Math.sqrt(discriminant);
                if (! Double.isNaN(sqrt)) {
                    double y = (- .5 * b[i] + sqrt) / a;
                    double x = _b[i] - _a * y;
                    tangentCircleCenters[nCircles++] = swapped ? 
                            new Point2D.Double(y, x) :
                            new Point2D.Double(x, y);
                    
                    if (discriminant > Geo2D.getEps()) {
                        y = (- .5 * b[i] - sqrt) / a;
                        x = _b[i] - _a * y;
                        tangentCircleCenters[nCircles++] = swapped ?
                                new Point2D.Double(y, x) :
                                new Point2D.Double(x, y);
                    }
                }
            }
        }
        
        return nCircles;
    }
    
    // =========================================================================
    // Utilities ===============================================================
    // =========================================================================
    
    public static final int OUTSIDE = 0;
    public static final int INSIDE = -1;
    public static final int CLIPPED = 1;

    /**
     * Clips a line segment against a rectangle.
     * @param rect the clip rectangle
     * @param l the line segment to be clipped
     * @return OUTSIDE if the line is outside the rectangle, INSIDE if the line
     *      is inside the rectangle and CLIPPED if the line was CLIPPED 
     */
    public static int clip(Rectangle2D rect, Line2D l) {
        // Calculate outcodes
        int outcode1 = rect.outcode(l.getP1());
        int outcode2 = rect.outcode(l.getP2());

        // Line outside?
        if ((outcode1 & outcode2) != 0)
            return OUTSIDE;

        // Both points inside?
        if (outcode1 == 0 && outcode2 == 0)
            return INSIDE;

        // Clipping
        double x1 = l.getX1();
        double y1 = l.getY1();
        double x2 = l.getX2();
        double y2 = l.getY2();

        // Clip near start point
        if ((outcode1 & Rectangle2D.OUT_LEFT) != 0) {
            y1 = y1 + (y2 - y1) * (rect.getMinX() - x1) / (x2 - x1);
            x1 = rect.getMinX();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_RIGHT) != 0) {
            y1 = y1 + (y2 - y1) * (rect.getMaxX() - x1) / (x2 - x1);
            x1 = rect.getMaxX();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_TOP) != 0) {
            x1 = x1 + (x2 - x1) * (rect.getMinY() - y1) / (y2 - y1);
            y1 = rect.getMinY();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_BOTTOM) != 0) {
            x1 = x1 + (x2 - x1) * (rect.getMaxY() - y1) / (y2 - y1);
            y1 = rect.getMaxY();
            outcode1 = rect.outcode(x1, y1);
        }

        // Clip near end point
        if ((outcode2 & Rectangle2D.OUT_LEFT) != 0) {
            y2 = y1 + (y2 - y1) * (rect.getMinX() - x1) / (x2 - x1);
            x2 = rect.getMinX();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_RIGHT) != 0) {
            y2 = y1 + (y2 - y1) * (rect.getMaxX() - x1) / (x2 - x1);
            x2 = rect.getMaxX();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_TOP) != 0) {
            x2 = x1 + (x2 - x1) * (rect.getMinY() - y1) / (y2 - y1);
            y2 = rect.getMinY();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_BOTTOM) != 0) {
            x2 = x1 + (x2 - x1) * (rect.getMaxY() - y1) / (y2 - y1);
            y2 = rect.getMaxY();
            outcode2 = rect.outcode(x2, y2);
        }

        // Set values and return
        l.setLine(x1, y1, x2, y2);
        return CLIPPED;
    }
    
    public static boolean _clip(Rectangle2D rect, Line2D l) {
        // Calculate outcodes
        int outcode1 = rect.outcode(l.getP1());
        int outcode2 = rect.outcode(l.getP2());

        // Line outside?
        if ((outcode1 & outcode2) != 0)
            return false;

        // Both points inside?
        if (outcode1 == 0 && outcode2 == 0)
            return true;

        // Clipping
        double x1 = l.getX1();
        double y1 = l.getY1();
        double x2 = l.getX2();
        double y2 = l.getY2();

        // Clip near start point
        if ((outcode1 & Rectangle2D.OUT_LEFT) != 0) {
            y1 = y1 + (y2 - y1) * (rect.getMinX() - x1) / (x2 - x1);
            x1 = rect.getMinX();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_RIGHT) != 0) {
            y1 = y1 + (y2 - y1) * (rect.getMaxX() - x1) / (x2 - x1);
            x1 = rect.getMaxX();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_TOP) != 0) {
            x1 = x1 + (x2 - x1) * (rect.getMinY() - y1) / (y2 - y1);
            y1 = rect.getMinY();
            outcode1 = rect.outcode(x1, y1);
        }
        if ((outcode1 & Rectangle2D.OUT_BOTTOM) != 0) {
            x1 = x1 + (x2 - x1) * (rect.getMaxY() - y1) / (y2 - y1);
            y1 = rect.getMaxY();
            outcode1 = rect.outcode(x1, y1);
        }

        // Clip near end point
        if ((outcode2 & Rectangle2D.OUT_LEFT) != 0) {
            y2 = y1 + (y2 - y1) * (rect.getMinX() - x1) / (x2 - x1);
            x2 = rect.getMinX();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_RIGHT) != 0) {
            y2 = y1 + (y2 - y1) * (rect.getMaxX() - x1) / (x2 - x1);
            x2 = rect.getMaxX();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_TOP) != 0) {
            x2 = x1 + (x2 - x1) * (rect.getMinY() - y1) / (y2 - y1);
            y2 = rect.getMinY();
            outcode2 = rect.outcode(x2, y2);
        }
        if ((outcode2 & Rectangle2D.OUT_BOTTOM) != 0) {
            x2 = x1 + (x2 - x1) * (rect.getMaxY() - y1) / (y2 - y1);
            y2 = rect.getMaxY();
            outcode2 = rect.outcode(x2, y2);
        }

        // Set values and return
        l.setLine(x1, y1, x2, y2);
        return true;
    }

   /**
    * Determines whether a cubic parametric curve segment can be transformed
    * into a circular arc.
    * 
    * Hint: A cubic parametric curve is just an approximation of a circular arc.
    * The maximum radial error is about 0.000273 * radius. A radial difference
    * of this magnitude is accepted when performing the algorithm.
    * 
    * @param cub the cubic parametric curve segment
    * @param arc the circular arc segment
    * @return <code>true</code> if cub is an arc, <code>false</code> if not
    */
    public static boolean getCircularArc(CubicCurve2D cub, Arc2D arc) {

        // Start and end point must not be coincident
        Point2D p1 = cub.getP1();
        Point2D p2 = cub.getP2();
        if (Geo2D.equality(p1, p2))
            return false;
        
        // Subdivide to find a point on the arc
        CubicCurve2D left = new CubicCurve2D.Double();
        CubicCurve2D right = new CubicCurve2D.Double();
        cub.subdivide(left, right);
        Point2D pc = left.getP2();
        
        // Create arc from three points
        arc = calculateArc2D(p1, pc, p2, arc);
        if (arc == null)
            return false;
        double radius = 0.5 * arc.getWidth();
        
        // Reduce the eps value because the arc is only approximated
        double eps = 0.0003 * radius;
        
        // Subdivide left and check if center is on arc
        left.subdivide(left, null);
        double dist = Geo2D.distance(left.getX2(), left.getY2(),
                arc.getCenterX(), arc.getCenterY());
        double delta = Math.abs(radius - dist);
        if (delta > eps)
            return false;
        
        // Subdivide right and check if center is on arc
        right.subdivide(null, right);
        dist = Geo2D.distance(right.getX1(), right.getY1(),
                arc.getCenterX(), arc.getCenterY());
        delta = Math.abs(radius - dist);
        if (delta > eps)
            return false;
        
        // Is an arc
        return true;
    }

    /**
     * Calculates the center point of an arc segment that is defined by two
     * points and a bulge value.  The bulge A is defined as<br/>
     *   A :=  2 * H / L <br/>
     * where H and L are the height and length of the chord.<br/>
     *
     * @param p1 the start point of the arc
     * @param p2 the end point of the arc
     * @param bulge the bulge value in the range [-1, +1].
     * @param cp reference to a Point2D instance to be returned. If the
     * reference is null then a new Point2D instance will be returned.
     * @return the center point or null for a straight line segment
     * @throws IllegalArgumentException
     */
    public static Point2D getCenterPoint(Point2D p1, Point2D p2, double bulge,
    Point2D cp) throws IllegalArgumentException {
        // Case: Straight line segment
        if (Math.abs(bulge) < getEps())
            return null;

        // Calculate the center point
        if (bulge < -1 - getEps() || bulge > 1 + getEps())
            throw new IllegalArgumentException("Illegal bulge value");

        double l = distance(p1, p2);
        if (l < getEps())
            return null;
        double h = 0.5 * bulge * l;
        double r = 0.5 * l / Math.sin(0.5 * angleFromBulge(bulge));

        Vector2D vec12 = new Vector2D(p2).subtract(p1);
        Vector2D vecPerp = new Vector2D(vec12);
        vecPerp.left();
        vecPerp.normalize();
        vecPerp.scaleBy(r-h);

        Vector2D veccp = new Vector2D(p1);
        vec12.scaleBy(0.5);
        veccp.add(vec12);
        veccp.add(vecPerp);

        if (cp == null)
            cp = new Point2D.Double();
        cp.setLocation(veccp);
        return cp;
    }

    /**
     * Converts a bulge value to an angle.<br/>
     *     phi = 4 * atan(bulge)
     *
     * @param bulge the bulge value
     * @return the angle in radians
     */
    public static double angleFromBulge(double bulge) {
        return 4. * Math.atan(bulge);
    }

    /**
     * Converts an angle to a bulge value.<br/>
     *     phi = 4 * atan(bulge)
     *
     * @param angle the angle in radians
     * @return the bulge value
     */
    public static double bulgeFromAngle(double angle) {
        return Math.tan(0.25 * angle);
    }

    /**
     * Calculates a circle from 3 given points.<br/>
     * 
     * If two points are coincident then the circle is centered at the center 
     * of the connecting line between the two coincident points and the third
     * point. This is also true if all three points are coincident.<br/>
     *
     * If three non-coincident points are located on a line then the ellipse
     * with infinite radius is indicated by a null Ellipse2D returned.
     * 
     * @param _p1 the first point
     * @param _p2 the second point
     * @param _p3 the third point
     * @param dst the ellipse to be set
     * @return the ellipse dst or null if the radius is infinite
     */
    public static Ellipse2D calculateEllipse2D(Point2D _p1, Point2D _p2,
            Point2D _p3, Ellipse2D dst) {

        // Shift coordinates to avoid numerical problems
        double xoff = (_p1.getX() + _p2.getX() + _p3.getX()) / 3.;
        double yoff = (_p1.getY() + _p2.getY() + _p3.getY()) / 3.;
        
        Point2D p1 = new Point2D.Double(_p1.getX() - xoff, _p1.getY() - yoff);
        Point2D p2 = new Point2D.Double(_p2.getX() - xoff, _p2.getY() - yoff);
        Point2D p3 = new Point2D.Double(_p3.getX() - xoff, _p3.getY() - yoff);
        
        // Degenerated circle?
        boolean equalityP1P2 = equality(p1, p2);
        boolean equalityP2P3 = equality(p2, p3);
        boolean equalityP1P3 = equality(p1, p3);
        if (equalityP1P2 && equalityP2P3 && equalityP1P3) {
            dst.setFrameFromCenter(p1.getX() + xoff, p1.getY() + yoff, 
                    p1.getX() + xoff, p1.getY() + yoff);
            return dst;
        }

        // Coincident start and end point?
        Point2D p = null;  // Location of two coincident points
        Point2D q = null;  // Location of third point
        if (equalityP1P2) {
            p = p1;
            q = p3;
        }
        else if (equalityP1P3) {
            p = p1;
            q = p2;
        }
        else if (equalityP2P3) {
            p = p2;
            q = p1;
        }
        if (p != null) {
            double radius = 0.5 * Geo2D.distance(p, q);
            double centerX = .5 * (p.getX() + q.getX());
            double centerY = .5 * (p.getY() + q.getY());
            double cornerX = centerX + radius;
            double cornerY = centerY + radius;
            dst.setFrameFromCenter(centerX + xoff, centerY + yoff, 
                    cornerX + xoff, cornerY + yoff);
            return dst;
        }
        
        // All points on a line?
        Line2D l = new Line2D.Double(p1, p3);
        if (Geo2D.containment(l, /*unbounded*/ true, p2))
            return null;
        l.setLine(p1, p2);
        if (Geo2D.containment(l, /*unbounded*/ true, p3))
            return null;
        l.setLine(p2, p3);
        if (Geo2D.containment(l, /*unbounded*/ true, p1))
            return null;
        
        /*
         * Let
         *    A := xm * xm + ym * ym - r * r
         *    B := 2 * xm
         *    C := 2 * ym
         *    with the center (xm, ym) and the radius r
         *
         * Equation system to be solved
         *    A - x1 * B - y1 * C = -x1 * x1 - y1 * y1
         *    A - x2 * B - y2 * C = -x2 * x2 - y2 * y2
         *    A - x3 * B - y3 * C = -x3 * x3 - y3 * y3
         *
         * Solution
         *    xm = B / 2
         *    ym = C / 2
         *    r * r = xm * xm + ym * ym - A
         */


        // Set coefficients
        double a1 = 1;
        double b1 = -p1.getX();
        double c1 = -p1.getY();
        double d1 = -p1.getX() * p1.getX() - p1.getY() * p1.getY();

        double a2 = 1;
        double b2 = -p2.getX();
        double c2 = -p2.getY();
        double d2 = -p2.getX() * p2.getX() - p2.getY() * p2.getY();

        double a3 = 1;
        double b3 = -p3.getX();
        double c3 = -p3.getY();
        double d3 = -p3.getX() * p3.getX() - p3.getY() * p3.getY();

        // Solve via Cramer's rule
        double detD = a1 * (b2 * c3 - b3 * c2) - b1 * (a2 * c3 - a3 * c2) +
        c1 * (a2 * b3 - a3 * b2);
        double detA = d1 * (b2 * c3 - b3 * c2) - b1 * (d2 * c3 - d3 * c2) +
        c1 * (d2 * b3 - d3 * b2);
        double detB = a1 * (d2 * c3 - d3 * c2) - d1 * (a2 * c3 - a3 * c2) +
        c1 * (a2 * d3 - a3 * d2);
        double detC = a1 * (b2 * d3 - b3 * d2) - b1 * (a2 * d3 - a3 * d2) +
        d1 * (a2 * b3 - a3 * b2);

        double a = detA / detD;
        double b = detB / detD;
        double c = detC / detD;
        if (Double.isNaN(a) || Double.isNaN(b) || Double.isNaN(c))
            return null;

        double xm = 0.5 * b;
        double ym = 0.5 * c;
        double r = Math.sqrt(xm * xm + ym * ym - a);

        dst.setFrameFromDiagonal(xm - r + xoff, ym - r + yoff, xm + r + xoff, 
                ym + r + yoff);
        return dst;
    }

    /**
     * Calculates a circular arc from 3 given points.<br/>
     * 
     * If start point and end point are coincident then the arc is centered
     * at the center of the connecting line between the two coincident points
     * and the point on the arc. This is also true if all three points are
     * coincident.<br/>
     *
     * If three non-coincident points are located on a line then the arc
     * with infinite radius is indicated by a null Arc2D returned.
     *
     * @param _sp the start point
     * @param _p a point on the arc
     * @param _ep the end point
     * @param dst the arc to be set
     * @return the arc dst or null if the radius is infinite
     */
    public static Arc2D calculateArc2D(Point2D _sp, Point2D _p, Point2D _ep,
    Arc2D dst) {

        // Shift coordinates to avoid numerical problems
        double xoff = (_sp.getX() + _p.getX() + _ep.getX()) / 3.;
        double yoff = (_sp.getY() + _p.getY() + _ep.getY()) / 3.;
        
        Point2D sp = new Point2D.Double(_sp.getX() - xoff, _sp.getY() - yoff);
        Point2D p = new Point2D.Double(_p.getX() - xoff, _p.getY() - yoff);
        Point2D ep = new Point2D.Double(_ep.getX() - xoff, _ep.getY() - yoff);
        
        // Calculate the ellipse to get the center and radius
        Ellipse2D ell = calculateEllipse2D(sp, p, ep, new Ellipse2D.Double());

        // Radius infinite?
        if (ell == null)
            return null;

        // Extract circle parameters
        double xm = ell.getCenterX();
        double ym = ell.getCenterY();
        double r = 0.5 * ell.getHeight();

        // Radius = 0?
        if (r < Geo2D.getEps()) {
            dst.setArcByCenter(xm + xoff, ym + yoff, r, 0, 0, Arc2D.OPEN);
            return dst;
        }
        
        // Calculate the arc
        Vector2D vsp = new Vector2D(xm, ym, sp.getX(), sp.getY());
        Vector2D vp = new Vector2D(xm, ym, p.getX(), p.getY());
        Vector2D vep = new Vector2D(xm, ym, ep.getX(), ep.getY());

        try {
            double startAngle = Vector2D.X_UNIT.getAngleCCW(vsp);
            double sweepAngle = vsp.getAngleCCW(vep);
            double angleVspVp = vsp.getAngleCCW(vp);
            if (angleVspVp > sweepAngle)
                sweepAngle = - (Math.PI + Math.PI - sweepAngle);
            dst.setArcByCenter(xm + xoff, ym + yoff, r,
                    - Math.toDegrees(startAngle),
                    - Math.toDegrees(sweepAngle), Arc2D.OPEN);
            
            // Check bounding box
            Rectangle2D rct = dst.getBounds2D();
            if (Double.isNaN(rct.getWidth()))
                return null;
        }
        catch (NullVectorException e) {
            dst.setArcByCenter(xm, ym, r, 0, 0, Arc2D.OPEN);
        }
        
        // Arc calculated
        return dst;
    }
}