/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.msg;

import java.io.DataInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.InetAddress;
import java.net.Socket;

import cib.util.Dbg;

/** 
 * The ClientMsgHandler class provides - based upon the functionality of
 * the {@link MsgHandler MsgHandler} class - a high level protocol for
 * communication in TCP/IP networks (e.g. the Internet).
 * <p>
 * Currently the ClientMsgHandler class allows for only one Connection to be
 * established (which should be obviously the server represented by a
 * {@link ServerMsgHandler ServerMsgHandler} object).
 * <p>
 * Connections are actually created by the class function
 * {@link ClientMsgHandler#getServerConnection(String serverHostName,
 * int port) getServerConnection()} which returns a ClientMsgHandler object
 * on success.
 * <p>
 * In addition to the MsgHandler class unique client IDs and (non-unique)
 * client names (currently the DNS name) are assigned to each ClientMsgHandler
 * object.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.4 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 * 
 * @see Msg
 * @see MsgReactor
 * @see ServerMsgHandler
 * @see MsgHandler
 * @see MsgBrokenConnection
 */
public class ClientMsgHandler extends MsgHandler {

    private static int s_nHandlers = 0;
    private int m_clientId = 0;
    private static Socket s_socket = null;

   /** 
    * Establishes a connection to a server. The message listener thread is
    * started automatically.
    * @param serverHostName the DNS name of the server
    * @param port the port number of the server process
    * @return the ClientMsgHandler object or null
    */
    public static synchronized ClientMsgHandler getServerConnection(
        String serverHostName, int port) {

        if (s_nHandlers != 0)
            return null;    // One Handler only
        s_nHandlers++;

        try {
            InetAddress ipAddress = InetAddress.getByName(serverHostName);

            // Connect to Chat Server
            System.out.println(
                "ClientMsgHandler: Try to connect to server at ipAddress " +
                ipAddress + " and port " + port + " ...");
            s_socket = new Socket(ipAddress, port);
            System.out.println(
                "ClientMsgHandler: ... Connection to server established" +
                s_socket);

            // Create Message handler
            InputStream in = s_socket.getInputStream();
            OutputStream out = s_socket.getOutputStream();
            ClientMsgHandler cmh = new ClientMsgHandler(in, out);

            // Get Id from server
            DataInputStream din = new DataInputStream(in);
            int id = din.readInt();
            cmh.setClientId(id);

            // Start the message handler thread
            cmh.start();
            return cmh;
        }
        catch (java.net.UnknownHostException e) {
            Dbg.println("ServerMsgHandler.getServerConnection(): " + e);
        }
        catch (java.io.IOException e) {
            Dbg.println("ServerMsgHandler.getServerConnection(): " + e);
        }
        return null;
    }

    // Construction and destruction
    private ClientMsgHandler(InputStream in, OutputStream out) {
        super(in, out);
    }


   /** 
    * Closes this mesage handler. You don't have to inform the Message
    * Handler on the other side of the channel.
    */
    public void close() {
        super.close();
        if (s_socket != null) {
            try {
                s_socket.close();
                s_socket = null;
            }
            catch (IOException e) {
            }
        }
        s_nHandlers = 0;
    }

   /** 
    * Returns the client ID that was automatically generated by the server.
    * @return the unique client ID
    */
    public int getClientId() {
        return m_clientId;
    }

    private void setClientId(int id) {
        m_clientId = id;
    }

   /** 
    * Returns the client name (currently as DNS name)
    * @return the non-unique client name
    */
    public String getClientName() {
        InetAddress addr = s_socket.getLocalAddress();
        return addr.getHostName();
    }
}