/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.msg;

import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.util.Collections;
import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Set;

import cib.util.Dbg;

/**
 * The MsgHandler class provides a commmunication infrastructure for TCP/IP
 * networks (e.g. the Internet).
 * <p>
 * The user of the MsgHandler class must establish a network connection by
 * the help of the two Java classes {@link java.net.Socket Socket} and
 * {@link java.net.ServerSocket ServerSocket}. Once the
 * connection is established a MsgHandler object must be created on each
 * peer passing the respective input and output streams to the constructors.
 * Messages between the two MsgHandler objects are exchanged by
 * serializing {@link Msg Message} objects. The process of serialization is
 * completely hidden to the user of this class.
 * <p>
 * An arbitrary number of {@link MsgReactor Message Reactors} can be
 * registered to the class of each message. The MsgHandler continuously
 * reads incoming messages in a separate thread (without blocking). For each
 * incoming message the Message Reactors are notified by the method {@link
 * MsgReactor#notify(Msg m, Object context) MsgReactor.notify()}.
 * <p>
 * If the connection to the other peer breaks down then a {@link
 * MsgBrokenConnection MsgBrokenConnection} object will be sent to all
 * registered reactors. It is recommended for applications to register a
 * reactor for this event.
 * <p>
 * The MsgHandler represents a low level interface for network communication.
 * The MsgHandler keeps track about all MsgHandler objects created in this
 * application. These objects can be queried.
 * <p>
 * Normally you will build your collaborating applications based upon the
 * classes {@link ServerMsgHandler ServerMsgHandler} and {@link
 * ClientMsgHandler ClientMsgHandler}.
 *
 * @author Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 *
 * @see Msg
 * @see MsgReactor
 * @see ServerMsgHandler
 * @see ClientMsgHandler
 * @see MsgBrokenConnection
 */
public class MsgHandler extends Thread {

    // Data representation
    private static Set<MsgHandler> s_allHandlers =
        Collections.synchronizedSet(new HashSet<MsgHandler>());

    private Boolean m_connectionBroken = Boolean.TRUE;  // needed for sync
    private OutputStream m_out = null;
    private InputStream m_in = null;
    private Hashtable<Class,HashSet<_MessageData>> m_reactors =
        new Hashtable<Class,HashSet<_MessageData>>();

   /**
    * Creates a MsgHandler. The MsgHandler thread must be started manually.
    * @param in input stream
    * @param out output stream
    */
    public MsgHandler(InputStream in, OutputStream out) {
        synchronized (s_allHandlers) {
            s_allHandlers.add(this);
        }
        m_connectionBroken = Boolean.FALSE;
        m_out = out;
        m_in = in;
    }

   /**
    * Closes this mesage handler and removes all Message Reactors.
    */
    public void close() {
        synchronized (s_allHandlers) {
            s_allHandlers.remove(this);
        }
        synchronized (m_reactors) {
            m_reactors.clear();
        }
        try {
            if (m_in != null) {
                m_in.close();
                m_in = null;
            }
            if (m_out != null) {
                m_out.close();
                m_out = null;
            }
        }
        catch (IOException e) {
        }
    }

   /**
    * Closes the streams.
    */
    protected void finalize() {
        close();
    }

   /**
    * The message handler keeps track about all MsgHandler objects created
    * in this application.
    * @return the set of all MsgHandler objects created in this application
    */
    public static Set<MsgHandler> getAllMsgHandlers() {
        synchronized (s_allHandlers) {
            return s_allHandlers;
//            HashSet s = (HashSet) s_allHandlers;
//            return (Set) s.clone();
        }
    }

   /**
    * Sends (narrowcasts) a Message object to another MsgHandler
    * Messages of type {@link MsgBrokenConnection MsgBrokenConnection} will
    * never be sent.
    * @param m the message object
    */
    public void send(Msg m) {

        if (connectionBroken() || m_out == null)
            return; // Why send messages

        if (m instanceof MsgBrokenConnection)
            return; // Is a local message only

        Dbg.println(this, "Send Msg of type " + m.getClass().getName());

        try {
              ObjectOutputStream oout = new ObjectOutputStream(m_out);
               oout.writeObject(m);
               oout.flush();
        }
        catch (IOException e) {
            Dbg.println(e);
            handleBrokenConnection();
        }
    }

   /**
    * Broadcasts a Message object to all known MsgHandlers except this one.
    * @param m the message object
    */
    public static void broadcast(Msg m) {
        synchronized (s_allHandlers) {
            for (MsgHandler mh : s_allHandlers) {
                if (mh != null)
                    mh.send(m);
            }
        }
    }

   /**
    * Registers a Reactor and a Context Object to a messages class. For each
    * incoming message of type class the Reactor will be notified and the
    * context object will be passed.
    * @param msgClass the class object of the message
    * @param r the Reactor object to notify
    * @param context the context object to be passed to the Reactor on notification
    */
    public void registerReactor(Class msgClass, MsgReactor r, Object context) {

        // Create the message data object
        _MessageData data = new _MessageData(r, context);

        // Create set
        synchronized (m_reactors) {
            HashSet<_MessageData> set = m_reactors.get(msgClass);
            if (set == null)
                set = new HashSet<_MessageData>();

            set.add(data);
                m_reactors.put(msgClass, set);
        }
    }

   /**
    * Registers a Reactor to a messages class. For each
    * incoming message of type class the Reactor will be notified.
    * @param msgClass the class object of the message
    * @param r the Reactor object to notify
    */
    public void registerReactor(Class msgClass, MsgReactor r) {
        this.registerReactor(msgClass, r, null);
    }

   /**
    * Registers a Reactor to the Class of a message object. For each
    * incoming message of type class the Reactor will be notified and the
    * context object will be passed.
    * @param m the message object
    * @param r the Reactor object to notify
    * @param context the context object to be passed to the Reactor on notification
    */
    public void registerReactor(Msg m, MsgReactor r, Object context) {
        this.registerReactor(m.getClass(), r, context);
    }

   /**
    * Registers a Reactor to the Class of a message object. For each
    * incoming message of type class the Reactor will be notified.
    * @param m the message object
    * @param r the Reactor object to notify
    */
    public void registerReactor(Msg m, MsgReactor r) {
        this.registerReactor(m.getClass(), r, null);
    }

   /**
    * Unregisters a reactor
    * @param r the Reactor object to unregister
    */
    public void unregisterReactor(MsgReactor r) {
        synchronized (m_reactors) {
            for (Enumeration<Class> e = m_reactors.keys(); e.hasMoreElements(); ) {
                Class cls = e.nextElement();
                HashSet<_MessageData> set = m_reactors.get(cls);
                Iterator<_MessageData> it = set.iterator();
                while (it.hasNext()) {
                    _MessageData data = it.next();
                    if (r == data.getReactor()) {
                        Dbg.println(this, "unregisterReactor " + cls);
                        it.remove();
                    }
                }
                if (set.isEmpty()) {
                    m_reactors.remove(cls);
                }
            }
        }
    }

   /**
    * Unregisters all reactor objects
    */
    public void unregisterAllReactors() {
        synchronized (m_reactors) {
            m_reactors.clear();
        }
    }

   /**
    * Continuously listens for incoming messages. For each arriving message
    * the registered reactors are notified.
    */
    public void run() {
        try {
//            while (true) {
            while (m_in != null) {
                // Read object from stream ...
                ObjectInputStream oin = new ObjectInputStream(m_in);
                Msg msg = (Msg) oin.readObject();   // Blocks

                // ... and notify the reactors
                notifyReactors(msg);
            }
        }
        catch (IOException e) {
            handleBrokenConnection();
        }
        catch (ClassNotFoundException e) {
            handleBrokenConnection();
        }
    }

    // Helpers

    protected boolean connectionBroken() {
        synchronized (m_connectionBroken) {
            return m_connectionBroken.booleanValue();
        }
    }

    private void setConnectionBroken() {
        synchronized (m_connectionBroken) {
            m_connectionBroken = Boolean.TRUE;
        }
    }

    private void notifyReactors(Msg msg) {

        // During iteration numerous reactors are notified. The reactors
        // may change the collection during iteration. This is the reason
        // why we do it in two steps.

        // Find the reactor objects ...
        synchronized (m_reactors) {
            HashSet<_MessageData> dataObjects = new HashSet<_MessageData>();

            Class _class = msg.getClass();
            while (_class != null) {
                HashSet<_MessageData> set = m_reactors.get(_class);
                if (set != null) {
                    for (_MessageData messageData : set) {
                        dataObjects.add(messageData);
                    }
                }
                _class = _class.getSuperclass();
            }

            // ... and notify the reactor objects
            for (_MessageData messageData2 : dataObjects) {
                MsgReactor r = messageData2.getReactor();
                r.notify(msg, messageData2.getContext());
            }
        }
    }

    protected void handleBrokenConnection() {
        synchronized (s_allHandlers) {
            s_allHandlers.remove(this);
        }
        setConnectionBroken();
        this.notifyReactors(new MsgBrokenConnection());
        this.close();
    }
}
class _MessageData {   // Helper to store reactor and context object

    private MsgReactor reactor = null;
    private Object context = null;

    _MessageData(MsgReactor r, Object c) {
        this.reactor = r;
        this.context = c;
    }

    MsgReactor getReactor() {
        return this.reactor;
    }

    Object getContext() {
        return this.context;
    }
}