/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.update;

import java.io.Serializable;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import cib.util.coll.CollUtilities;
import cib.util.coll.ReverseAccessMap;

/**
 * This class represents an update interface between objects and graph nodes.
 *
 * @author Christian Koch
 * @version $Revision: 1.14 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 * @see cib.util.update.UpdateInterface
 * @see cib.util.update.UpdatableGraph
 * @see cib.util.update.Node
 * @see cib.util.update.Update
 * @see cib.util.update.GraphException
 * @see cib.util.coll.ReverseAccessMap
 * @see cib.util.coll.CollUtilities
 */
@Deprecated
public class UpdateInterfaceInstance implements UpdateInterface, Serializable {

    private static final long serialVersionUID = 0L;

    /**
     * The object-node map.
     */
    private ReverseAccessMap<Updatable,Object> m_pairs = null;

    /**
     * The updatable link graph.
     */
    private UpdatableGraph m_updatableGraph = null;

    /**
     * The current update.
     */
    private transient Update m_currentUpdate = null;

    /**
     * Constructs an update interface.
     */
    public UpdateInterfaceInstance() {
        this.m_pairs =
            CollUtilities.bijectiveReverseAccessMap(new HashMap<Updatable,Object>());
        this.m_updatableGraph = new _UpdatableGraph(this);
    }

    /**
     * Creates a node for a specified, updatable object, adds it to the
     * updatable link graph and updates the object-node map.
     * @param u the specified, updatable object
     * @return <code>true</code> if the node was added
     */
    public boolean addNode(Updatable u) {
        if (!m_pairs.containsKey(u)) {
            Node n = new Node();
            this.m_pairs.put(u, n);
            this.m_updatableGraph.addNode(n);
            return true;
        }
        return false;
    }

    /**
     * Removes a node for a specified, updatable object from the updatable link
     * graph and
     * updates the object-node map.
     * @param u the specified updatable object
     * @return <code>true</code> if the node was removed
     */
    public boolean removeNode(Updatable u) {
        if (this.m_pairs.containsKey(u)) {
            //System.out.println("removeNode in _UpdateInterface called...");
            Node n = (Node)this.m_pairs.get(u);
            this.m_updatableGraph.removeNode(n);
            this.m_pairs.remove(u);
            return true;
        }
        return false;
    }

    /**
     * Adds an edge to the updatable link graph between source node and
     * destination node for a specified source and destination objects.
     * @param src the specified source object
     * @param dest the specified destination object
     * @return <code>true</code> if the edge was added
     * @throws GraphException a graph exception
     */
    public boolean addEdge(Updatable src, Updatable dest) throws GraphException {
        if (this.m_pairs.containsKey(src) && this.m_pairs.containsKey(dest)) {
            Node _src = (Node)this.m_pairs.get(src);
            Node _dest = (Node)this.m_pairs.get(dest);
            return this.m_updatableGraph.addEdge(_src, _dest);
        }
        return false;
    }

    /**
     * Removes an edge from the updatable link graph between source node and
     * destination node for a specified source and destination objects.
     * @param src the specified source object
     * @param dest the specified destination object
     * @return <code>true</code> if the edge was removed
     */
    public boolean removeEdge(Updatable src, Updatable dest) {
        if (this.m_pairs.containsKey(src) && this.m_pairs.containsKey(dest)) {
            Node _src = (Node)this.m_pairs.get(src);
            Node _dest = (Node)this.m_pairs.get(dest);
            return this.m_updatableGraph.removeEdge(_src, _dest);
        }
        return false;
    }

    /**
     * Adds a node for a specified updatable object to the set of changed nodes in the
     * updatable link graph.
     * @param u the specified updatable object
     */
    public void wasChanged(Updatable u) {
        if (this.m_pairs.containsKey(u)) {
            Node n = (Node)this.m_pairs.get(u);
            this.m_updatableGraph.wasChanged(n);
            //System.out.println("changed node added: "+n.toString());
        }
    }

    /**
     * Gets the updatable object for a specified node object.
     * @param node the specified node object
     * @return the corrensponding updatable object
     */
    public Updatable getUpdatable(Node node) {
        return this.m_pairs.getKey(node);
    }

    /**
     * Gets the node object for a specified updatable object.
     * @param u the specified updatable object
     * @return the corrensponding node object
     */
    public Node getNode(Updatable u) {
        return (Node)this.m_pairs.get(u);
    }

    /**
     * Gets the set of updatable objects that correspond to nodes in
     * the updateable graph.
     * @return the set of <code>Updatable</code> objects
     */
    public Set<Updatable> getGraphUpdatables() {
        Set nodes = m_updatableGraph.getNodeSet();
        Set<Updatable> updatables = new HashSet<Updatable>();
        Iterator it = nodes.iterator();
        while (it.hasNext()) {
            updatables.add(this.m_pairs.getKey(it.next()));
        }
        return updatables;
    }

    /**
     * Gets the set of updatable objects that are successors of the
     * specified updatable object in the updatable graph.
     * @param u the specified <code>Updatable</code> object
     * @return the set of <code>Updatable</code> objects
     */
    public Set<Updatable> getSuccessors(Updatable u) {
        Set<Updatable> edges = new HashSet<Updatable>();
        if (m_pairs.containsKey(u)) {
            Node node = (Node)m_pairs.get(u);
            Node[] nodes = m_updatableGraph.getEdges(node);
            for (int i=0; i<nodes.length; i++) {
                edges.add(m_pairs.getKey(nodes[i]));
            }
        }
        return edges;
    }

    /**
     * Starts an update concerning all node objects as destination objects.
     */
    public void update(Object context) {
        Update u = new _Update(this.m_updatableGraph);
        this.m_currentUpdate = u;
        this.m_updatableGraph.setUpdate(u);
        Iterator it = this.m_updatableGraph.getNodeSet().iterator();
        while (it.hasNext()) {
            u.addDestNode((Node)it.next());
        }
        this.m_updatableGraph.update(context);
        this.m_currentUpdate = null;
    }

    /**
     * Starts an update concerning the objects described by the specified iterator
     * object.
     * @param it the iterator that describes the destination objects
     */
    public void update(Iterator it, Object context) {
        Update u = new _Update(this.m_updatableGraph);
        this.m_currentUpdate = u;
        this.m_updatableGraph.setUpdate(u);
        while (it.hasNext()) {
            u.addDestNode((Node)this.m_pairs.get(it.next()));
        }
        this.m_updatableGraph.update(context);
        this.m_currentUpdate = null;
    }

    /**
     * Determines if a specified updatable object has still to be updated
     * concerning current update process.
     * @param u the specified object
     * @return <code>true</code> if the object has still to be updated
     */
    protected boolean stillToUpdate(Updatable u) {
        if (this.m_currentUpdate != null) {
            Node n = (Node)this.m_pairs.get(u);
            return this.m_currentUpdate.stillToUpdate(n);
        }
        throw new IllegalStateException("No update in progress!");
    }

    /**
     * Determines if a specified updatable object has already been updated
     * concerning current update process.
     * @param u the specified object
     * @return <code>true</code> if the node has already been updated
     */
    protected boolean alreadyUpdated(Updatable u) {
        if (this.m_currentUpdate != null) {
            Node n = (Node)this.m_pairs.get(u);
            return this.m_currentUpdate.alreadyUpdated(n);
        }
        throw new IllegalStateException("No update in progress!");
    }

    /**
     * Gets all predecessors of a specified updatable object concerning current
     * update process.
     * @param u the specified object
     * @return the predecessors
     */
    public Set<Updatable> getPredecessors(Updatable u) {
        Set<Updatable> p = new HashSet<Updatable>();
        if (this.m_pairs.containsKey(u)) {
            boolean[][] l = this.m_updatableGraph.getLink();
            List<Node> nodes = Arrays.asList(this.m_updatableGraph.getNodes());
            Node n = (Node)this.m_pairs.get(u);
            int index = nodes.indexOf(n);
            for (int i=0; i<l.length; i++) {
                if (l[i][index]) {
                    Object node = nodes.get(i);
                    p.add(this.m_pairs.getKey(node));
                }
            }
        }
        return p;
    }

    /**
     * ONLY FOR TESTING
     */
    public void printPairs() {
        String s = "---Pairs---\n";
        Set<Updatable> keys = this.m_pairs.keySet();
        Iterator<Updatable> it_keys = keys.iterator();
        Object key;
        while (it_keys.hasNext()) {
            key = it_keys.next();
            s += key.toString() + " <--> " + (this.m_pairs.get(key)).toString()+"\n";
        }
        System.out.println(s + "---Pairs---\n");
    }

    /**
     * ONLY FOR TESTING
     */
    public void printGraph() {
        String s = "---Graph---\n";
        Graph g = this.m_updatableGraph.getLinkGraph();
        Node[] n = g.getNodes();
        for (int i=0; i<n.length; i++) {
            s += "" + n[i] + ": ";
            Node[] e = g.getEdges(n[i]);
            for (int j=0; j<e.length; j++) {
                s += e[j] + ", ";
            }
            s += "\n";
        }
        System.out.println(s + "---Graph---\n");
    }
}