/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.update;

import java.io.Serializable;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import cib.util.coll.CollUtilities;

/**
 * This class represents an update.
 *
 * @author Christian Koch
 * @version $Revision: 1.16 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 * @see cib.util.update.Update
 * @see cib.util.update.UpdatableGraph
 * @see cib.util.update.Graph
 * @see cib.util.update.Node
 * @see cib.util.coll.CollUtilities
 */
@Deprecated
public class _Update implements Update, Serializable {

    private static final long serialVersionUID = 0L;

    /**
     * The set of destination nodes.
     */
    private Set<Node> m_destinationSet = null;

    /**
     * The set of domain nodes.
     */
    private Set<Node> m_domainSet = null;

    /**
     * The updatable link graph that has to be updated.
     */
    private UpdatableGraph m_linkGraph = null;

    /**
     * The list of domain nodes.
     */
    private ArrayList<Node> m_domainNodes = null;

    /**
     * The adjacency matrix for the link graph of domain nodes.
     */
    private transient boolean [][] m_domainLink = null;

    /**
     * The chrono matrix for the link graph of domain nodes.
     */
    private transient int[][] m_domainChronos = null;

    /**
     * The age vector of the domain nodes.
     */
    private transient int[] m_domainAge = null;

    private transient Node m_currentUpdatedNode = null;

    /**
     * Standard constructor.
     *
     */
    public _Update() {
    }
    /**
     * Constructs an update with a specified updatable graph object.
     * @param ug the specified updatable graph object
     */
    public _Update(UpdatableGraph ug) {
        this.m_linkGraph = ug;
        this.m_destinationSet = CollUtilities.subSet(new HashSet<Node>(),
                CollUtilities.observableSet(m_linkGraph.getNodeSet()));
        this.m_domainSet = CollUtilities.subSet(new HashSet<Node>(),
                CollUtilities.observableSet(m_linkGraph.getNodeSet()));
    }

    /**
     * Sets the upatable graph object for this update.
     * @see cib.util.update.Update#setUpdatableGraph(cib.util.update.UpdatableGraph)
     */
    public void setUpdatableGraph(UpdatableGraph up) {
        this.m_linkGraph = up;
        this.m_destinationSet = CollUtilities.subSet(new HashSet<Node>(),
                CollUtilities.observableSet(m_linkGraph.getNodeSet()));
        this.m_domainSet = CollUtilities.subSet(new HashSet<Node>(),
                CollUtilities.observableSet(m_linkGraph.getNodeSet()));
    }
    /**
     * Determines the node sequence in the update process and updates nodes
     * sequentially.<br>
     * 1. Determining the set of domain nodes<br>
     * 2. Computing the adjacency matrix for the link graph of domain nodes<br>
     * 3. Computing the chrono matrix for the link graph of domain nodes<br>
     * 4. Computing the age vector of domain nodes<br>
     * 5. Calling the update method for domain nodes sequentially
     */
    public void update(Object context) {
        this.setDomainSet();
        this.computeDomainLink();
        //this.printDomainLink();
        //this.printUpdate();
        this.computeDomainChronos();
        //this.printDomainChronos();
        this.computeDomainAge();
        //this.printDomainAge();
        this.callDomainUpdates(context);
        this.m_currentUpdatedNode = null;
    }

    /**
     * Adds a specified node to the set of destination nodes of this update.
     * @param node the specified node
     * @return <code>true</code> if the node was added
     */
    public boolean addDestNode(Node node) {
        return this.m_destinationSet.add(node);
    }

    /**
     * Removes a specified node from the set of destination nodes of this update.
     * @param node the specified node
     * @return <code>true</code> if the node was removed
     */
    public boolean removeDestNode(Node node) {
        return this.m_destinationSet.remove(node);
    }

    /**
     * Gets the set of domain nodes of this update.
     * @return the set of domain nodes
     */
    public Set<Node> getDomainSet() {
        return this.m_domainSet;
    }

    /**
     * Gets the set of destination nodes of this update.
     * @return the set of destination nodes
     */
    public Set<Node> getDestSet() {
        return this.m_destinationSet;
    }

    /*
     private void printUpdate() {
     String s = new String("---Update:---\nsemantics: \""+this.m_semantics+"\"\n");
     s += "changedSet: (";
     Iterator chan = this.m_linkGraph.getChangedSet().iterator();
     while (chan.hasNext())
     s += ((TestObject)((Node)chan.next()).getKey()).getID()+", ";
     s += ")\n";
     s += "destinationSet: (";
     Iterator dest = this.m_destinationSet.iterator();
     while (dest.hasNext())
     s += ((TestObject)((Node)dest.next()).getKey()).getID()+", ";
     s += ")\n";
     s += "domainSet: (";
     Iterator dom = this.m_domainSet.iterator();
     while (dom.hasNext())
     s += ((TestObject)((Node)dom.next()).getKey()).getID()+", ";
     s += ")";
     System.out.println(s);
     }
     */

    /**
     * Determining the set of domain nodes.
     */
    private void setDomainSet() {
        boolean[][] hull = this.m_linkGraph.getHull();
        Set changed = this.m_linkGraph.getChangedNodeSet();

        ///*
        //elements of changed set
        Iterator it_aa = changed.iterator();
        Iterator<Node> it_az;
        Node node_aa;
        Node node_az;
        while (it_aa.hasNext()) {
            node_aa = (Node)it_aa.next();
            int index_aa = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_aa);
            //System.out.println("a: "+index_a);
            it_az = this.m_destinationSet.iterator();
            while (it_az.hasNext()) {
                node_az = it_az.next();
                int index_az = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_az);
                if (hull[index_aa][index_az] == true) this.m_domainSet.add(node_aa);
                //System.out.println("z: "+index_z);
            }
        }
        //*/

        ///*
        //elements of destination set
        Iterator<Node> it_zz = this.m_destinationSet.iterator();
        Iterator it_za;
        Node node_zz;
        Node node_za;
        while (it_zz.hasNext()) {
            node_zz = it_zz.next();
            int index_zz = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_zz);
            //System.out.println("a: "+index_a);
            it_za = changed.iterator();
            while (it_za.hasNext()) {
                node_za = (Node)it_za.next();
                int index_za = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_za);
                if (hull[index_za][index_zz] == true) this.m_domainSet.add(node_zz);
                //System.out.println("z: "+index_z);
            }
        }
        //*/

        ///*
        //elements of chains
        ArrayList<Node> e_z;
        Iterator<Node> it_ez = this.m_destinationSet.iterator();
        Iterator it_ea;
        Node node_ez;
        Node node_ea;
        Node node_ee;
        while (it_ez.hasNext()) {
            e_z = new ArrayList<Node>();
            node_ez = it_ez.next();
            int index_ez = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_ez);
            //all elements ez depends on
            for (int i=0; i < hull.length; i++) {
                if (hull[i][index_ez] == true) {
                    node_ee = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().get(i);
                    e_z.add(node_ee);
                }
            }
            it_ea = changed.iterator();
            while (it_ea.hasNext()) {
                //elements ee that depend on ea
                node_ea = (Node)it_ea.next();
                int index_ea = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(node_ea);
                for (int j=0; j < e_z.size(); j++) {
                    int index_ee = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(e_z.get(j));
                    if (hull[index_ea][index_ee] == true) this.m_domainSet.add(e_z.get(j));
                }
            }
        }
        //*/

        ///*
        // the intersection of a and z
        Iterator it_a = changed.iterator();
        Node node_a;
        while (it_a.hasNext()) {
            node_a = (Node)it_a.next();
            if (this.m_destinationSet.contains(node_a))
                this.m_domainSet.add(node_a);
        }
        //*/
    }

    /**
     * Computes the adjacency matrix for the link graph of domain nodes.
     */
    private void computeDomainLink() {
        boolean[][] link = this.m_linkGraph.getLink();
        this.m_domainNodes = new ArrayList<Node>();

        Iterator<Node> it = this.m_domainSet.iterator();
        while (it.hasNext()) this.m_domainNodes.add(it.next());

        int dim = this.m_domainNodes.size();
        this.m_domainLink = new boolean[dim][dim];
        for (int i=0; i < dim; i++) {
            for (int j=0; j < dim; j++) {
                int i_link = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(
                        this.m_domainNodes.get(i));
                int j_link = ((_Graph)this.m_linkGraph.getLinkGraph()).getNodeList().indexOf(
                        this.m_domainNodes.get(j));
                this.m_domainLink[i][j] = link[i_link][j_link];
            }
        }
    }

    /*
     private void printDomainLink() {
     String s = new String("---DomainLinks:--- \n");
     for (int i=0; i < this.m_domainLink.length; i++) {
     s += "( ";
     for (int j=0; j < this.m_domainLink.length; j++) {
     s += this.m_domainLink[i][j]+" ";
     }
     s += ")\n";
     }
     System.out.println(s);
     }
     */

    /**
     * Computes the chrono matrix for the link graph of domain nodes.
     */
    private void computeDomainChronos() {
        final int NONE = -1;
        int dim = this.m_domainLink.length;
        int test;
        this.m_domainChronos = new int[dim][dim];
        // setting start values
        for (int i=0; i<dim; i++) {
            for (int j=0; j<dim; j++) {
                if (i == j) this.m_domainChronos[i][j] = 0;
                else if (this.m_domainLink[i][j]) this.m_domainChronos[i][j] = 1;
                else this.m_domainChronos[i][j] = NONE;
            }
        }
        // compute maximum length of all paths
        for (int k=0; k<dim; k++) {
            for (int i=0; i<dim; i++) {
                if (this.m_domainChronos[i][k] == NONE) continue;
                for (int j=0; j<dim; j++) {
                    if (this.m_domainChronos[k][j] == NONE) continue;
                    test = this.m_domainChronos[i][k] + this.m_domainChronos[k][j];
                    if (test > this.m_domainChronos[i][j])
                        this.m_domainChronos[i][j] = test;
                }
            }
        }
    }

    /*
     private void printDomainChronos() {
     String s = new String("---DomainChronos:--- \n");
     for (int i=0; i < this.m_domainChronos.length; i++) {
     s += "( ";
     for (int j=0; j < this.m_domainChronos.length; j++) {
     s += this.m_domainChronos[i][j]+" ";
     }
     s += ")\n";
     }
     System.out.println(s);
     }
     */

    /**
     * Computes the age vector domain nodes.
     */
    private void computeDomainAge() {
        int dim = this.m_domainChronos.length;
        int test, m;
        this.m_domainAge = new int[dim];
        //compute age of domain nodes
        for (int j=0; j<dim; j++) {
            test = 0;
            for (int i=0; i<dim; i++) {
                if (test < (m = this.m_domainChronos[i][j])) test = m;
            }
            this.m_domainAge[j] = test;
        }
    }

    /*
     private void printDomainAge() {
     String s = new String("---DomainAge:--- \n");
     s += "( ";
     for (int i=0; i < this.m_domainAge.length; i++) {
     s += this.m_domainAge[i]+" ";
     }
     s += ")\n";
     System.out.println(s);
     }
     */

    /**
     * Updates domain nodes sequentially concerning their age.
     */
    private void callDomainUpdates(Object context) {
        for (int i=0; i < this.m_domainAge.length; i++) {
            for (int j=0; j < this.m_domainAge.length; j++) {
                if (this.m_domainAge[j] == i) {
                    //call interface update method wirh parameter node
                    Node n = this.m_domainNodes.get(j);
                    this.m_currentUpdatedNode = n;
                    this.m_linkGraph.updateNode(n, context);
                }
            }
        }
    }

    /**
     * Determines if a specified node has still to be updated.
     * @param n the specified node
     * @return <code>true</code> if the node has still to be updated
     */
    public boolean stillToUpdate(Node n) {
        if (this.m_currentUpdatedNode != null) {
            // test if domain set contains n
            if (this.m_domainNodes.contains(n)) {
                int ni = this.m_domainNodes.indexOf(n);
                int ci = this.m_domainNodes.indexOf(this.m_currentUpdatedNode);
                int na = this.m_domainAge[ni];
                int ca = this.m_domainAge[ci];
                //System.out.println("na="+na+", ca="+ca);
                // checks domain age
                if (na >= ca)
                    return true;
                return false;
            }
            return false;
        }
        throw new IllegalStateException("No update in progress!");
    }

    /**
     * Determines if a specified node has already been updated.
     * @param n the specified node
     * @return <code>true</code> if the node has already been updated
     */
    public boolean alreadyUpdated(Node n) {
        if (this.m_currentUpdatedNode != null) {
            // test if domain set contains n
            if (this.m_domainNodes.contains(n)) {
                int ni = this.m_domainNodes.indexOf(n);
                int ci = this.m_domainNodes.indexOf(this.m_currentUpdatedNode);
                int na = this.m_domainAge[ni];
                int ca = this.m_domainAge[ci];
                //System.out.println("na="+na+", ca="+ca);
                // checks domain age
                if (na <= ca)
                    return true;
                return false;
            }
            return false;
        }
        throw new IllegalStateException("No update in progress!");
    }

}//end of class