/* 
 * Copyright (C) 2002-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.view;

import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Point2D;
import java.awt.geom.Rectangle2D;
import java.io.Serializable;

/** 
 * The Projection model.
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.5 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/04/29 08:45:34 $
 */
public class Projection2D implements Cloneable, Serializable {

    private static final long serialVersionUID = 0L;

    public static final int BORDER = 10;

    private Point2D m_vrp = new Point2D.Double(0, 0);
    private Point2D m_vup = new Point2D.Double(0, 1);
    private Rectangle2D m_viewPort = new Rectangle2D.Double(-1, -1, 10, 10);

    private double m_deviceWidth = 100;
    private double m_deviceHeight = 100;
    private transient double m_deviceDX = 0.0;
    private transient double m_deviceDY = 0.0;

    private AffineTransform m_modelToView = new AffineTransform();
    private AffineTransform m_viewToModel = new AffineTransform();

    // Constructors and cloning
    private Projection2D() {
    }
    public Projection2D(Rectangle2D viewPort, double deviceWidth,
    double deviceHeight) throws NoninvertibleTransformException {
        setViewPort(viewPort);
        setDeviceDimensions(deviceWidth, deviceHeight);
        setProjection();
    }
    public Object clone() {
        // Make a shallow copy
        Projection2D clone = new Projection2D();

        // Set attributes (make a deep copy)
        clone.m_vrp = new Point2D.Double(m_vrp.getX(), m_vrp.getY());
        clone.m_vup = new Point2D.Double(m_vup.getX(), m_vup.getY());
        clone.m_viewPort = new Rectangle2D.Double(m_viewPort.getX(),
            m_viewPort.getY(), m_viewPort.getWidth(), m_viewPort.getHeight());
        clone.m_deviceWidth = m_deviceWidth;
        clone.m_deviceHeight = m_deviceHeight;
        clone.m_deviceDX = m_deviceDX;
        clone.m_deviceDY = m_deviceDY;
        clone.m_modelToView = new AffineTransform(m_modelToView);
        clone.m_viewToModel = new AffineTransform(m_viewToModel);

        // Return the clone
        return clone;
    }

    // Set transformation
    public void setProjection() throws NoninvertibleTransformException {
        // Step 0: Define border in model units
        double sb = Math.min((m_deviceWidth) / m_viewPort.getWidth(),
            (m_deviceHeight) / m_viewPort.getHeight());
        double delta = BORDER / sb;
        final Rectangle2D viewPort = new Rectangle2D.Double(
            m_viewPort.getX() - delta, m_viewPort.getY() - delta,
            m_viewPort.getWidth() + delta + delta,
            m_viewPort.getHeight() + delta + delta);
        
        // Step 1: Translate (-vrpx, -vrpy)
        AffineTransform at = AffineTransform.getTranslateInstance(
            - m_vrp.getX(), - m_vrp.getY());

        // Step 2: Rotate so that VUP points in positive y-direction
        double angle = Math.asin(m_vup.getX() / m_vup.distance(0, 0));
        at.preConcatenate(AffineTransform.getRotateInstance(angle));

        // Step 3: Translate (-umin, vmax)
        double umin = /*m_*/viewPort.getX();
        double vmax = /*m_*/viewPort.getY() + /*m_*/viewPort.getHeight();
        at.preConcatenate(AffineTransform.getTranslateInstance(-umin, -vmax));

        // Step 4: Mirror about x-axis
        final AffineTransform MIRROR_X_TRF = new AffineTransform(
            /*m00*/  1, /*m10*/  0,
            /*m01*/  0, /*m11*/ -1,
            /*m02*/  0, /*m12*/  0
        );
        at.preConcatenate(MIRROR_X_TRF);

        // Step 5: Scale viewport to device
        double s = Math.min(m_deviceWidth / viewPort.getWidth(),
            m_deviceHeight / viewPort.getHeight());
        at.preConcatenate(AffineTransform.getScaleInstance(s, s));

        // Step 6: Center the view 
        double deviceRatio = m_deviceWidth / m_deviceHeight;
        double viewPortRatio = viewPort.getWidth() / viewPort.getHeight();
        if (deviceRatio < viewPortRatio) {
            double dx = 0;
            double ih = m_deviceWidth / viewPortRatio;
            double dy = 0.5 * (m_deviceHeight - ih);
            dy += 0.5;    // Raster device
            m_deviceDX = dx;
            m_deviceDY = dy;
            at.preConcatenate(AffineTransform.getTranslateInstance(dx, dy));
        }
        else {
            double dy = 0;
            double iw = m_deviceHeight * viewPortRatio;
            double dx = 0.5 * (m_deviceWidth - iw);
            dx += 0.5;    // Raster device
            m_deviceDX = dx;
            m_deviceDY = dy;
            at.preConcatenate(AffineTransform.getTranslateInstance(dx, dy));
        }
        
        // Store the transformation and the inverse transformation
        m_modelToView = at;
        m_viewToModel = m_modelToView.createInverse();
    }

    // Transformation matrix
    public AffineTransform getModelToView() {
        return (AffineTransform)m_modelToView.clone();
    }
    public AffineTransform getViewToModel() {
        return (AffineTransform)m_viewToModel.clone();
    }

    // Attributes
    public void setVrp(double x, double y) {
        m_vrp.setLocation(x, y);
    }
    public void setVrp(Point2D vrp) {
        m_vrp.setLocation(vrp);
    }
    public Point2D getVrp() {
        return (Point2D)m_vrp.clone();
    }
    public double getVrpX() {
        return m_vrp.getX();
    }
    public double getVrpY() {
        return m_vrp.getY();
    }

    public void setVup(double x, double y) {
        m_vup.setLocation(x, y);
    }
    public void setVup(Point2D vup) {
        m_vup.setLocation(vup);
    }
    public Point2D getVup() {
        return (Point2D)m_vup.clone();
    }
    public double getVupX() {
        return m_vup.getX();
    }
    public double getVupY() {
        return m_vup.getY();
    }

    public boolean setViewPort(double x, double y, double width,
            double height) {
        if (width < Double.MIN_VALUE || height < Double.MIN_VALUE)
            return false;
        m_viewPort.setRect(x, y, width, height);
        return true;
    }
    public boolean setViewPort(Rectangle2D viewPort) {
        return setViewPort(viewPort.getX(), viewPort.getY(),
            viewPort.getWidth(), viewPort.getHeight());
    }
    public Rectangle2D getViewPort() {
        return (Rectangle2D)m_viewPort.clone();
    }
    public double getMinU() {
        return m_viewPort.getX();
    }
    public double getMinV() {
        return m_viewPort.getY();
    }
    public double getMaxU() {
        return m_viewPort.getX() + m_viewPort.getWidth();
    }
    public double getMaxV() {
        return m_viewPort.getY() + m_viewPort.getHeight();
    }

    public void setDeviceDimensions(double width, double height) {
        if (width < 1 || height < 1)
            throw new IllegalArgumentException();
        m_deviceWidth = width;
        m_deviceHeight = height;
    }
    public double getDeviceWidth() {
        return m_deviceWidth;
    }
    public double getDeviceHeight() {
        return m_deviceHeight;
    }
    public double getDeviceDX() {
        return m_deviceDX;
    }
    public double getDeviceDY() {
        return m_deviceDY;
    }
}