/* 
 * Copyright (C) 2009-2010 Berthold Firmenich
 * 
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 * 
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 * 
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.view;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Insets;
import java.awt.LayoutManager;
import java.awt.Point;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.util.HashMap;
import java.util.Map;

import javax.swing.AbstractButton;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JMenuItem;
import javax.swing.JPanel;

/** 
 * Layout Manager for a single row of Components. If there is not enough space
 * available then a button is displayed at the right side of the row. After
 * pressing the button the missing components can be selected from a popup menu.
 * The names presented in the popup menu are queried via getName(). 
 * 
 * @author Berthold Firmenich
 * @version $Revision: 1.4 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/02/23 13:49:13 $
 */
public class SingleRowLayout implements LayoutManager {
    private boolean m_sizeUnknown = true;
    private int m_gap;
    private int m_minWidth = 0;
    private int m_minHeight = 0;
    private int m_preferredWidth = 0;
    private int m_preferredHeight = 0;
    
    private JButton m_continuationButton;
    private ScrollableJPopupMenu m_popupMenu = new ScrollableJPopupMenu();
    private ActionListener m_popupMenuItemListener = null;
    private Map<JMenuItem,Component> m_menuItemToComponentMap = 
        new HashMap<JMenuItem,Component>();

    public SingleRowLayout() {
        this(new JButton("\u00bb"), 5);
    }
    public SingleRowLayout(JButton continuationButton, int gap) {
        m_continuationButton = continuationButton;
        m_gap = gap;
        
        // Popup menu
        m_continuationButton.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                if (! m_popupMenu.isVisible())
                    m_popupMenu.show(m_continuationButton, 0,
                            m_continuationButton.getHeight());
                else
                    m_popupMenu.setVisible(false);
            } 
        });
        m_popupMenuItemListener = new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JMenuItem menuItem = (JMenuItem)e.getSource();
                Component c = m_menuItemToComponentMap.get(menuItem);

                // In case of a panel: refer to panel's first Component
                if (c instanceof JPanel)
                    c = ((JPanel)c).getComponent(0);

                if (c instanceof JComboBox) {
                    JComboBox cb = (JComboBox)c;
                    Point originalLocation = cb.getLocation();
                    
                    // System crashes if the combobox is outside the screen!
                    Dimension screenSize =
                        Toolkit.getDefaultToolkit().getScreenSize();
                    Point p = cb.getLocationOnScreen();
                    if (p.x >= screenSize.width) {
                        int dx = p.x - screenSize.width;
                        cb.setLocation(cb.getX() - dx - 1, cb.getY());
                    }
                    if (p.y >= screenSize.height) {
                        int dy = p.y - screenSize.height;
                        cb.setLocation(cb.getX(), cb.getY() - dy - 1);
                    }
                    
                    // Show the popup menu
                    cb.showPopup();
                    
                    // Move the component to the old position
                    cb.setLocation(originalLocation);
                }
                else {
                    Object[] ls = c.getListeners(ActionListener.class);
                    for (int i = 0; i < ls.length; i++) {
                        ActionListener al = (ActionListener)ls[i];
                        al.actionPerformed(new ActionEvent(c,
                                ActionEvent.ACTION_PERFORMED, ""));
                    }
                }
                m_popupMenu.setVisible(false);
            }
        };
    }

    /* Required by LayoutManager. */
    public void addLayoutComponent(String name, Component comp) {
    }

    /* Required by LayoutManager. */
    public void removeLayoutComponent(Component comp) {
    }

    private void setSizes(Container parent) {
        // Init
        int nComps = parent.getComponentCount();

        // Set minimum sizes
        m_minWidth = nComps == 0 ?
                0 : m_continuationButton.getWidth();
        m_minHeight = nComps == 0 ?
                0 : m_continuationButton.getHeight();

        // Set preferred sizes 
        m_preferredWidth = nComps * m_gap;
        m_preferredHeight = 0;
        for (int i = 0; i < nComps; i++) {
            Component c = parent.getComponent(i);
            if (c.isVisible()) {
                Dimension d = c.getPreferredSize();
                m_preferredWidth += d.width;
                if (d.height > m_preferredHeight)
                    m_preferredHeight = d.height;
            }
        }
    }

    /* Required by LayoutManager. */
    public Dimension preferredLayoutSize(Container parent) {
        // Init
        Dimension dim = new Dimension(0, 0);
        setSizes(parent);

        //Always add the container's insets!
        Insets insets = parent.getInsets();
        dim.width = m_preferredWidth + insets.left + insets.right;
        dim.height = m_preferredHeight + insets.top + insets.bottom;

        m_sizeUnknown = false;

        return dim;
    }

    /* Required by LayoutManager. */
    public Dimension minimumLayoutSize(Container parent) {
        Dimension dim = new Dimension(0, 0);
        setSizes(parent);

        //Always add the container's insets!
        Insets insets = parent.getInsets();
        dim.width = m_minWidth + insets.left + insets.right;
        dim.height = m_minHeight + insets.top + insets.bottom;

        m_sizeUnknown = false;

        return dim;
    }

    /* Required by LayoutManager. */
    /*
     * This is called when the panel is first displayed,
     * and every time its size changes.
     * Note: You CAN'T assume preferredLayoutSize or
     * minimumLayoutSize will be called -- in the case
     * of applets, at least, they probably won't be.
     */
    public void layoutContainer(Container parent) {
        Insets insets = parent.getInsets();
        int maxWidth = parent.getWidth() - (insets.left + insets.right);
        int maxHeight = parent.getHeight() - (insets.top + insets.bottom);
        int x = insets.left + m_gap, y = insets.top;

        // Go through the components' sizes, if neither
        // preferredLayoutSize nor minimumLayoutSize has
        // been called.
        if (m_sizeUnknown)
            setSizes(parent);

        // Remove the continuation button from the container
        parent.remove(m_continuationButton);

        // Enough space for all components?
        if (m_preferredWidth < maxWidth) {
            for (int i = 0; i < parent.getComponentCount(); i++) {
                Component c = parent.getComponent(i);
                if (c.isVisible()) {
                    Dimension d = c.getPreferredSize();
                    c.setBounds(x, y, d.width, d.height);
                    x += d.width + m_gap;
                }
            }
        }
        else {
            m_popupMenu.removeAll();
            m_menuItemToComponentMap.clear();
            int _maxWidth = maxWidth - m_continuationButton.getWidth() - m_gap;
            for (int i = 0; i < parent.getComponentCount(); i++) {
                Component c = parent.getComponent(i);
                if (c.isVisible()) {
                    Dimension d = c.getPreferredSize();
    
                    if ((x + d.width) > _maxWidth) {
                        // Component outside: Set right of the visible area
                        //   but not outside of the screen (crashes otherwise)
                        c.setBounds(maxWidth, maxHeight, d.width, d.height);

                        // Add a new item to the continuation popup menu
                        if (c instanceof JLabel) {
                            m_popupMenu.addSeparator();
                        }
                        else {
                            String text = c instanceof AbstractButton ?
                                    ((AbstractButton)c).getText() : c.getName();
                            JMenuItem item = new JMenuItem(text);
                            item.addActionListener(m_popupMenuItemListener);
                            m_popupMenu.add(item);
                            m_menuItemToComponentMap.put(item, c);
                        }
                    }
                    else {
                        // Component inside
                        c.setBounds(x, y, d.width, d.height);
                    }

                    // Next position
                    x += d.width + m_gap;
                }
            }

            // Set the continuation symbol
            Dimension d = m_continuationButton.getPreferredSize();
            m_continuationButton.setBounds(maxWidth - d.width, y, d.width,
                    d.height);
            parent.add(m_continuationButton);
        }
    }
}
