/*
 * Copyright (C) 2002-2010 Berthold Firmenich
 *
 * This program is free software; you can redistribute it and/or modify it under
 * the terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2 of the License, or (at your option) any later
 * version.
 *
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more
 * details.
 *
 * You should have received a copy of the GNU General Public License along with
 * this program; if not, look at "http://www.gnu.org/copyleft/gpl.html" or
 * write to the Free Software Foundation, Inc., 59 Temple Place, Suite 330,
 * Boston, MA  02111-1307  USA
 *
 * Contact address: berthold.firmenich@cademia.org
 * Homepage       : http://www.cademia.org/
 */

package cib.util.view;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.Paint;
import java.awt.Shape;
import java.awt.Toolkit;
import java.awt.Window;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.FocusEvent;
import java.awt.event.FocusListener;
import java.awt.geom.AffineTransform;
import java.awt.geom.NoninvertibleTransformException;
import java.awt.geom.Rectangle2D;
import java.awt.image.BufferedImage;
import java.awt.image.MemoryImageSource;
import java.net.URL;

import javax.swing.ImageIcon;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuItem;
import javax.swing.JToggleButton;
import javax.swing.KeyStroke;
import javax.swing.ListCellRenderer;
import javax.swing.plaf.basic.BasicComboPopup;


/**
 * View utilities.
 *
 * @author Daniel G. Beer, Berthold Firmenich, Christian Koch
 * @version $Revision: 1.25 $, $Author: cvs-cad-firmenic $
 * @version $Date: 2010/01/13 15:05:11 $
 */
public class ViewUtilities {

    public static boolean MAC_OS = false;
    public static boolean WINDOWS = false;
    public static boolean LINUX = false;
    static {
        String os = System.getProperty("os.name").toLowerCase();
        if (os.startsWith("mac os x"))
            MAC_OS = true;
        else if (os.startsWith("windows"))
            WINDOWS = true;
        else if (os.startsWith("linux"))
            LINUX = true;
    }
    
    private static boolean s_menuIconsEnabled = true;

    public static final int BUTTON_WIDTH = 21;
    public static final int BUTTON_HEIGHT = 21;

    /**
     * The CTRL key on Windows and META key on the Mac.
     */
    public static final int SHORTCUT =
        Toolkit.getDefaultToolkit().getMenuShortcutKeyMask();
    public static final int SHIFT = ActionEvent.SHIFT_MASK;
    public static final int ALT = ActionEvent.ALT_MASK;


    /**
     * Creates an image button based on URL.
     *
     * @param name button name
     * @param url image URL
     * @param help tooltip text
     * @return created button
     */
    public static JButton createImageButton(String name,
        URL url, String help) {
        
        ImageIcon icon;
        if (url != null)
            icon = new ImageIcon(url);
        else {
            int w = BUTTON_WIDTH;// * 2 / 4;
            int h = BUTTON_HEIGHT;// * 2 / 4;
            int pix[] = new int[w * h];
            for (int i = 4; i < w - 4; i++)
                for (int j = 4; j < h - 4; j++)
                    if (i == j || w - i == j)
                        pix[j * w + i] = Color.RED.getRGB();
                    else
                        pix[j * w + i] = 0;
            icon = new ImageIcon(Toolkit.getDefaultToolkit().createImage(
                    new MemoryImageSource(w, h, pix, 0, w)));
        }
        int iw = icon.getIconWidth();
        int ih = icon.getIconHeight();
        final JButton button = new JButton(icon);
        button.putClientProperty("JButton.buttonType", "toolbar");
        button.setName(name);
        Dimension dim = new Dimension(iw+2, ih+2);
        button.setMinimumSize(dim);
        button.setMaximumSize(dim);
        button.setPreferredSize(dim);
        button.setToolTipText(help);
        return button;
    }

    /**
     * Creates a toggle button based in URLs
     *
     * @param name button name
     * @param onUrl image url for selected button
     * @param offUrl image url for unselected button
     * @param help tooltip text
     * @param selected if <code>true</code> button's selected
     * @return the toggle button
     */
    public static JToggleButton createToggleButton(String name,
        URL onUrl, URL offUrl, String help, boolean selected) {
        ImageIcon onIcon, offIcon;
        if (onUrl != null) onIcon = new ImageIcon(onUrl);
        else {
            int w = BUTTON_WIDTH;// * 2 / 4;
            int h = BUTTON_HEIGHT;// * 2 / 4;
            int pix[] = new int[w * h];
            for (int i = 4; i < w-4; i++)
                for(int j = 4; j < h-4; j++)
                    if (i==j || w-i==j)
                        pix[j*w+i] = Color.RED.getRGB();
                    else
                        pix[j*w+i] = 0;
            onIcon = new ImageIcon(Toolkit.getDefaultToolkit().createImage(
                new MemoryImageSource(w, h, pix, 0, w)));
        }
        if (offUrl != null) offIcon = new ImageIcon(onUrl);
        else {
            int w = BUTTON_WIDTH;// * 2 / 4;
            int h = BUTTON_HEIGHT;// * 2 / 4;
            int pix[] = new int[w * h];
            for (int i = 4; i < w-4; i++)
                for(int j = 4; j < h-4; j++)
                    if (i==j || w-i==j)
                        pix[j*w+i] = Color.RED.getRGB();
                    else
                        pix[j*w+i] = 0;
            offIcon = new ImageIcon(Toolkit.getDefaultToolkit().createImage(
                new MemoryImageSource(w, h, pix, 0, w)));
        }

        int iw = Math.max(onIcon.getIconWidth(), offIcon.getIconWidth());
        int ih = Math.max(onIcon.getIconHeight(), offIcon.getIconHeight());

        final JToggleButton button = new JToggleButton();
        button.setName(name);
        button.setSelected(selected);
        button.setIcon(onIcon);
        button.setSelectedIcon(offIcon);

        Dimension dim = new Dimension(iw+2, ih+2);
        button.setMinimumSize(dim);
        button.setMaximumSize(dim);
        button.setPreferredSize(dim);
        button.setToolTipText(help);
        return button;
    }

    /**
     * Creates an image icon based on URL.
     *
     * @param url image URL
     * @param showProxy if <code>true</code> proxy is shown
     * @return created image icon
     */
    public static ImageIcon createImageByName(URL url, boolean showProxy) {
        if (url == null) {  //"missing" icon (x) if showProxy
            if (! showProxy) {
                int w = BUTTON_WIDTH - 2;
                int h = BUTTON_HEIGHT - 2;
                int pix[] = new int[w * h];
                ImageIcon icon = new ImageIcon(
                    Toolkit.getDefaultToolkit().createImage(
                        new MemoryImageSource(w, h, pix, 0, w)));
                return icon;
            }
            int w = BUTTON_WIDTH;// * 2 / 4;
            int h = BUTTON_HEIGHT;// * 2 / 4;
            int pix[] = new int[w * h];
            
            // draw three dots (...)
            pix[BUTTON_HEIGHT*11+8] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*11+9] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*11+12] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*11+13] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*11+16] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*11+17] = Color.BLACK.getRGB();
            
            pix[BUTTON_HEIGHT*12+8] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*12+9] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*12+12] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*12+13] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*12+16] = Color.BLACK.getRGB();
            pix[BUTTON_HEIGHT*12+17] = Color.BLACK.getRGB();
            

            ImageIcon icon = new ImageIcon(
                Toolkit.getDefaultToolkit().createImage(
                    new MemoryImageSource(w, h, pix, 0, w)));
            return icon;
        }
        return new ImageIcon(url);
    }

    /**
     * Creates a button showing a combo popup when pressed.
     *
     * @param b button
     * @param c combo box
     * @param selectLower <code>true</code> if selecting lower is allowed
     * @return modified button
     */
    public static JButton createPopupButton(JButton b,
        final JComboBox c, final boolean selectLower) {
        b.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                JButton _b = (JButton)e.getSource();
                c.setRenderer(new _LabelComboBoxRenderer(selectLower));
                final BasicComboPopup popup = new BasicComboPopup(c);
                popup.show(_b, 0, _b.getBounds().height);
                c.addActionListener(new ActionListener() {
                    public void actionPerformed(ActionEvent e) {
                        popup.hide();
                    }
                });
            }
        });
        return b;
    }

    // line width icon
    public static ImageIcon createImageIconByLineWidth(double d) {
        int b = 3;
        int w = 3 * BUTTON_WIDTH / 2;
        int h = BUTTON_HEIGHT - 2*b;
        int t = (int)Math.round(11*d);
        if (t <= 1) t = 2;
        if (t >= h) t = h - 1;
        int pix[] = new int[w * h];
        for (int i = b; i <= w-b; i++)
            for(int j = (h/2-t/2); j < (h/2+t/2); j++) {
                pix[j*w+i+1] = Color.BLACK.getRGB();
            }
        ImageIcon icon = new ImageIcon(Toolkit.getDefaultToolkit().createImage(
            new MemoryImageSource(w, h, pix, 0, w)));
        return icon;
    }

    // line pattern icon
    public static ImageIcon createImageIconByLinePattern(double[] d) {
        if (d == null || d.length == 0 || d.length == 1)
            d = new double[]{5.0, 0.0};
        int b = 3;
        int w = 5 * BUTTON_WIDTH / 2;
        int h = BUTTON_HEIGHT - 2*b;
        int t = 1;
        int pix[] = new int[w * h];
        int col[] = new int[w - 2*b]; //
        boolean gap = false;
        boolean end = false;
        int k = 0;
        double sum = 0.0;
        for (int i = 0; i < d.length; i++) {
            sum += d[i];
        }
        for (h = 0; h < w/sum+1; h++) {
            for (int i = 0; i < d.length; i++) {
                for (int j = 0; j < 2*d[i]; j++) {
                    if (k == col.length-1) {
                        end = true;
                        break;
                    }
                    if (!gap) col[k] = Color.BLACK.getRGB();
                    k++;
                }
                if (end) break;
                gap = !gap;
            }
            if (end) break;
        }
        for (int i = b; i < w-b; i++)
            for(int j = ((h-t)/2); j < ((h+t)/2); j++) {
                int index = j * w + i + 1;
                if (index >= 0 && index < pix.length)
                    pix[index] = col[i-b];
            }
        ImageIcon icon = new ImageIcon(Toolkit.getDefaultToolkit().createImage(
            new MemoryImageSource(w, h, pix, 0, w)));
        return icon;
    }

    // draw paint icon
    public static ImageIcon createImageIconByDrawPaint(Paint p) {
        int b = 3;
        int w = BUTTON_WIDTH - 2*b;
        int h = BUTTON_HEIGHT - 2*b;
        BufferedImage bi = new BufferedImage(w, h,
            BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D big = (Graphics2D)bi.getGraphics();
        big.setTransform(AffineTransform.getTranslateInstance(b+1, b+1));
        big.setBackground(new Color(0,0,0,0));
        big.clearRect(-1, -1, bi.getWidth()+1 , bi.getHeight()+1);
        Rectangle2D rect = new Rectangle2D.Double(
            0, 0, 8, 8);
        BasicStroke stroke = new BasicStroke(2.0f);
        Shape shape = stroke.createStrokedShape(rect);
        big.setPaint(Color.BLACK);
        big.draw(shape);
        big.setPaint(p);
        big.fill(shape);
        big.dispose();
        bi.flush();
        return new ImageIcon(bi);
    }

    // fill paint icon
    public static ImageIcon createImageIconByFillPaint(Paint p) {
        int b = 3;
        int w = BUTTON_WIDTH - 2*b;
        int h = BUTTON_HEIGHT - 2*b;
        BufferedImage bi = new BufferedImage(w, h,
            BufferedImage.TYPE_4BYTE_ABGR);
        Graphics2D big = (Graphics2D)bi.getGraphics();

        if (p instanceof FillPaint) {
            FillPaint fp = (FillPaint)p;
            double sv_p2m = fp.getPaperToModel();
            AffineTransform sv_m2v = fp.getModelToView();
            double p2m = 1.0;//((double)w)/10.0;
            fp.setPaperToModel(p2m);
            Projection2D p2d = null;
            try {
                p2d = new Projection2D(
                    new Rectangle2D.Double(0.,0.,10.,10.),
                    bi.getWidth(), bi.getHeight());
            }
            catch (NoninvertibleTransformException e) {
                // FIXME: Handle Projection2D = null
                e.printStackTrace();
                big.setPaint(p);
                big.setTransform(AffineTransform.getTranslateInstance(b, b));
                big.setBackground(new Color(0,0,0,0));
                big.clearRect(-1, -1, bi.getWidth()+1 , bi.getHeight()+1);
                Rectangle2D rect = new Rectangle2D.Double(
                    0,0, 10, 10);
                big.fill(rect);
                big.setColor(Color.BLACK);
                big.setStroke(new BasicStroke(0.0f));
                big.draw(rect);
                return new ImageIcon(bi);
            }
            AffineTransform at = p2d.getModelToView();
            at.preConcatenate(AffineTransform.getTranslateInstance(
                -b/p2m, -b/p2m));
            fp.setModelToView(at);

            big.setPaint(fp);
            big.setTransform(AffineTransform.getTranslateInstance(b, b));
            big.setBackground(new Color(0,0,0,0));
            big.clearRect(-1, -1, bi.getWidth()+1 , bi.getHeight()+1);
            Rectangle2D rect = new Rectangle2D.Double(
                0,0, 10, 10);
            big.fill(rect);
            big.setColor(Color.BLACK);
            big.setStroke(new BasicStroke(0.0f));
            big.draw(rect);
            fp.setPaperToModel(sv_p2m);
            fp.setModelToView(sv_m2v);
        }
        else {
            big.setPaint(p);
            big.setTransform(AffineTransform.getTranslateInstance(b, b));
            big.setBackground(new Color(0,0,0,0));
            big.clearRect(-1, -1, bi.getWidth()+1 , bi.getHeight()+1);
            Rectangle2D rect = new Rectangle2D.Double(
                0,0, 10, 10);
            big.fill(rect);
            big.setColor(Color.BLACK);
            big.setStroke(new BasicStroke(0.0f));
            big.draw(rect);
        }
        return new ImageIcon(bi);
    }

    // font family icon
    public static ImageIcon createImageIconByFontFamily(String name) {
        int b = 3;
        int w = (int)(BUTTON_WIDTH * 1.5);
        int h = BUTTON_HEIGHT - 2*b;
        Image image = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        Graphics g = image.getGraphics();
        g.setColor(Color.BLACK);
        g.setFont(new Font(name, Font.PLAIN, 14));
        g.drawString("abc", b, h-b);
        return new ImageIcon(image);
    }

    // text size icon
    public static ImageIcon createImageIconByTextSize(double d) {
        int b = 3;
        int w = (int)(BUTTON_WIDTH * 1.5);
        int h = BUTTON_HEIGHT - 2*b;
        Image image = new BufferedImage(w, h, BufferedImage.TYPE_INT_ARGB);
        Graphics g = image.getGraphics();
        g.setColor(Color.BLACK);
        int t = (int)(2*Math.round(d));
        if (t > h) t = h;
        g.setFont(new Font("arial", Font.PLAIN, t));
        g.drawString(String.valueOf(d), b, h-b);
        return new ImageIcon(image);
    }

    /**
     * Creates a menu.
     *
     * @param name
     * @param mnemonic (KeyEvent.VK_[A-Z])
     * @return the created menu
     */
    public static JMenu createMenu(String name, int mnemonic){
        JMenu menu = new JMenu(name);
        menu.setName(name);
        menu.setMnemonic(mnemonic);
        return menu;
    }

    public static void enableMenuIcons(boolean enable) {
        s_menuIconsEnabled = enable;
    }

    public static boolean menuIconsEnabled() {
        return s_menuIconsEnabled;
    }
    
    /**
     * Creates an image icon based on URL if menuIconsEnabled().
     *
     * @param url image URL
     * @param showProxy if <code>true</code> proxy is shown
     * @return created image icon
     */
    public static ImageIcon createMenuImageByName(URL url, boolean showProxy) {
        if (! s_menuIconsEnabled)
            return null;
        return createImageByName(url, showProxy);
    }
    
    /**
     * Creates a menu item. An icon is only created if menuIconsEnabled().
     *
     * @param name
     * @param url
     * @param mnemonic
     * @param keyCode
     * @param modifiers
     * @param help
     * @return the created menu item
     */
    public static JMenuItem createMenuItem(String name, URL url,
        char mnemonic, int keyCode, int modifiers, String help) {
        JMenuItem item;
        if (s_menuIconsEnabled) {
            ImageIcon icon = createImageByName(url, false);
            item = new JMenuItem(name, icon);
        }
        else {
            item = new JMenuItem(name);
        }
        item.setName(name);
        if (keyCode != 0)
            item.setAccelerator(KeyStroke.getKeyStroke(keyCode, modifiers));
        if (mnemonic != ' ' && name.indexOf(mnemonic) != -1)
            item.setMnemonic(mnemonic);
        item.setToolTipText(help);
        return item;
    }

    /**
     * Combobox renderer with label items that allows selecting all items
     * below that current one.
     */
    public static class _LabelComboBoxRenderer extends JLabel implements
        ListCellRenderer {
        private static final long serialVersionUID = 0L;

        private boolean m_selectLower = false;

        public _LabelComboBoxRenderer(boolean selectLower) {
            setOpaque(true);
            setFocusable(false);
            setHorizontalAlignment(LEFT);
            setVerticalAlignment(CENTER);
            m_selectLower = selectLower;
        }
        public Component getListCellRendererComponent(JList list, Object value,
            int index, boolean isSelected, boolean cellHasFocus) {

            if (value != null) {
                if (value instanceof JLabel) {
                    JLabel l = (JLabel) value;
                    setText(" " + l.getText() + " ");
                    setToolTipText(l.getText());
                    setIcon(l.getIcon());
                }
                else setText(" " + value.toString() + " ");
            }
            else {
                setText("");
                setIcon(null);
            }
            if (index == list.getSelectedIndex()
                || (index < list.getSelectedIndex() && m_selectLower)) {
                setBackground(list.getSelectionBackground());
                setForeground(list.getSelectionForeground());
            }
            else {
                setBackground(list.getBackground());
                setForeground(list.getForeground());
            }

            int n = (list.getSelectedIndex() + 1);
            if (m_selectLower) {
                if (!(value instanceof JLabel)) {
                    switch (n) {
                    case 0:
                        list.setToolTipText("No elements.");
                        break;
                    case 1:
                        list.setToolTipText("1 element.");
                        break;
                    default:
                        list.setToolTipText(n + " elements.");
                        break;
                    }
                }
                else {
                    if (!(value instanceof JLabel)) {
                        switch (n) {
                        case 0:
                            list.setToolTipText("No elements.");
                            break;
                        default:
                            String text = (list.getSelectedValue() == null) ? ""
                                : list.getSelectedValue().toString();
                            list.setToolTipText(text);
                            break;
                        }
                    }
                }
            }
            return this;
        }
    }

    /**
     * Requests that a Window gets the input focus. In contrast to
     * Window.requestFocus() this method doesn't return before the focus has
     * been gained.
     *
     * @param comp The Window to get the input focus
     */
    public static void requestFocusAsynchronously(final Window comp) {
        if (comp.isFocusOwner())
            return;
        final _GuardAsynchronous guard = new _GuardAsynchronous();
        FocusListener fl = new FocusListener() {
            public void focusGained(FocusEvent e) {
                comp.removeFocusListener(this);
                guard.resume();
            }
            public void focusLost(FocusEvent e) {
            }
        };
        comp.addFocusListener(fl);
        comp.requestFocus();    // Don't use requestFocusInWindow() in Mac OS X
        guard.suspend();
        return;
    }

    /*
     * Guards of type _GuardAsynchronous support an asynchronous order of
     * suspend and resume. Both sequences (suspend, resume) and 
     * (resume, suspend) are considered to be correct. In both cases the thread
     * will be resumed at the end of the sequence.
     *
     * This behavior was needed to asynchronously request the Focus.
     */
    private static class _GuardAsynchronous { // Needed for focus events only
        private int m_suspended = 0;

        public synchronized void suspend() {
            if (m_suspended > 0)
                throw new InternalError("_GuardAsynchronous: already suspended");
            m_suspended++;
            while (m_suspended > 0) {
                try {
                    wait();
                }
                catch (InterruptedException e) {
                }
            }
        }

        public synchronized void resume() {
            if (--m_suspended == 0)
                notifyAll();
        }
    }
}